@@ -10,6 +10,8 @@
 
 #include <limits>
 
+#include "rr/rr.h"
+
 #include "preload/preload_interface.h"
 
 #include "AutoRemoteSyscalls.h"
@@ -49,6 +51,8 @@ const char* FileId::special_name() const {
       return "(syscallbuf)";
     case PSEUDODEVICE_VDSO:
       return "(vdso)";
+    case PSEUDODEVICE_SYSV_SHM:
+      return "";
   }
   FATAL() << "Not reached";
   return nullptr;
@@ -74,8 +78,11 @@ ostream& operator<<(ostream& o, const Mapping& m) {
 
 /*static*/ MappableResource MappableResource::shared_mmap_file(
     const TraceMappedRegion& file) {
-  return MappableResource(FileId(file.stat(), PSEUDODEVICE_SHARED_MMAP_FILE),
-                          file.file_name().c_str());
+  return MappableResource(
+      FileId(file.stat(), file.type() == TraceMappedRegion::MMAP
+                              ? PSEUDODEVICE_SHARED_MMAP_FILE
+                              : PSEUDODEVICE_SYSV_SHM),
+      file.file_name().c_str());
 }
 
 /*static*/ MappableResource MappableResource::syscallbuf(pid_t tid, int fd,
@@ -87,145 +94,6 @@ ostream& operator<<(ostream& o, const Mapping& m) {
   return MappableResource(FileId(st), path);
 }
 
-/**
- * Represents a refcount set on a particular address.  Because there
- * can be multiple refcounts of multiple types set on a single
- * address, Breakpoint stores explicit USER and INTERNAL breakpoint
- * refcounts.  Clients adding/removing breakpoints at this addr must
- * call ref()/unref() as appropropiate.
- */
-struct Breakpoint {
-  typedef shared_ptr<Breakpoint> shr_ptr;
-
-  // AddressSpace::destroy_all_breakpoints() can cause this
-  // destructor to be invoked while we have nonzero total
-  // refcount, so the most we can assert is that the refcounts
-  // are valid.
-  ~Breakpoint() { assert(internal_count >= 0 && user_count >= 0); }
-
-  shr_ptr clone() { return shr_ptr(new Breakpoint(*this)); }
-
-  void ref(TrapType which) {
-    assert(internal_count >= 0 && user_count >= 0);
-    ++*counter(which);
-  }
-  int unref(TrapType which) {
-    assert(internal_count > 0 || user_count > 0);
-    --*counter(which);
-    assert(internal_count >= 0 && user_count >= 0);
-    return internal_count + user_count;
-  }
-
-  TrapType type() const {
-    // NB: USER breakpoints need to be processed before
-    // INTERNAL ones.  We want to give the debugger a
-    // chance to dispatch commands before we attend to the
-    // internal rr business.  So if there's a USER "ref"
-    // on the breakpoint, treat it as a USER breakpoint.
-    return user_count > 0 ? TRAP_BKPT_USER : TRAP_BKPT_INTERNAL;
-  }
-
-  static shr_ptr create() { return shr_ptr(new Breakpoint()); }
-
-  // "Refcounts" of breakpoints set at |addr|.  The breakpoint
-  // object must be unique since we have to save the overwritten
-  // data, and we can't enforce the order in which breakpoints
-  // are set/removed.
-  int internal_count, user_count;
-  uint8_t overwritten_data;
-  static_assert(sizeof(overwritten_data) ==
-                    sizeof(AddressSpace::breakpoint_insn),
-                "Must have the same size.");
-
-private:
-  Breakpoint() : internal_count(), user_count() {}
-  Breakpoint(const Breakpoint& o) = default;
-
-  int* counter(TrapType which) {
-    assert(TRAP_BKPT_INTERNAL == which || TRAP_BKPT_USER == which);
-    int* p = TRAP_BKPT_USER == which ? &user_count : &internal_count;
-    assert(*p >= 0);
-    return p;
-  }
-};
-
-enum {
-  EXEC_BIT = 1 << 0,
-  READ_BIT = 1 << 1,
-  WRITE_BIT = 1 << 2
-};
-
-/** Return the access bits above needed to watch |type|. */
-static int access_bits_of(WatchType type) {
-  switch (type) {
-    case WATCH_EXEC:
-      return EXEC_BIT;
-    case WATCH_WRITE:
-      return WRITE_BIT;
-    case WATCH_READWRITE:
-      return READ_BIT | WRITE_BIT;
-    default:
-      FATAL() << "Unknown watchpoint type " << type;
-      return 0; // not reached
-  }
-}
-
-/**
- * Track the watched accesses of a contiguous range of memory
- * addresses.
- */
-class Watchpoint {
-public:
-  typedef shared_ptr<Watchpoint> shr_ptr;
-
-  ~Watchpoint() { assert_valid(); }
-
-  shr_ptr clone() { return shr_ptr(new Watchpoint(*this)); }
-
-  void watch(int which) {
-    assert_valid();
-    exec_count += (EXEC_BIT & which) != 0;
-    read_count += (READ_BIT & which) != 0;
-    write_count += (WRITE_BIT & which) != 0;
-  }
-  int unwatch(int which) {
-    assert_valid();
-    if (EXEC_BIT & which) {
-      assert(exec_count > 0);
-      --exec_count;
-    }
-    if (READ_BIT & which) {
-      assert(read_count > 0);
-      --read_count;
-    }
-    if (WRITE_BIT & which) {
-      assert(write_count > 0);
-      --write_count;
-    }
-    return exec_count + read_count + write_count;
-  }
-
-  int watched_bits() const {
-    return (exec_count > 0 ? EXEC_BIT : 0) | (read_count > 0 ? READ_BIT : 0) |
-           (write_count > 0 ? WRITE_BIT : 0);
-  }
-
-  static shr_ptr create() { return shr_ptr(new Watchpoint()); }
-
-private:
-  Watchpoint() : exec_count(), read_count(), write_count() {}
-  Watchpoint(const Watchpoint&) = default;
-
-  void assert_valid() const {
-    assert(exec_count >= 0 && read_count >= 0 && write_count >= 0);
-  }
-
-  // Watchpoints stay alive until all watched access typed have
-  // been cleared.  We track refcounts of each watchable access
-  // separately.
-  int exec_count, read_count, write_count;
-};
-
 /**
  * Advance *str to skip leading blank characters.
  */
@@ -354,6 +222,7 @@ static void iterate_memory_map(Task* t, memory_map_iterator_t it,
 static void print_process_mmap_iterator(void* unused, Task* t,
                                         const struct map_iterator_data* data) {
   fputs(data->raw_map_line, stderr);
+  fputc('\n', stderr);
 }
 
 /**
@@ -363,7 +232,7 @@ static void print_process_mmap(Task* t) {
   iterate_memory_map(t, print_process_mmap_iterator, nullptr);
 }
 
-AddressSpace::~AddressSpace() { session->on_destroy(this); }
+AddressSpace::~AddressSpace() { session_->on_destroy(this); }
 
 void AddressSpace::after_clone() { allocate_watchpoints(); }
 
@@ -420,22 +289,25 @@ remote_ptr<uint8_t> AddressSpace::find_syscall_instruction(Task* t) {
 static void write_rr_page(Task* t, ScopedFd& fd) {
   switch (t->arch()) {
     case x86: {
-      static const uint8_t x86_data[] = { 0x90, 0x90, // padding
-                                          // rr_page_untraced_syscall_ip:
-                                          0xcd, 0x80, // int 0x80
-                                          // rr_page_ip_in_untraced_syscall:
-                                          0xc3, // ret
-                                          0x90, 0x90, 0x90, 0x90, 0x90,
-                                          0x90, 0x90, 0x90, 0x90, // padding
-                                          // rr_page_traced_syscall_ip:
-                                          0xcd, 0x80, // int 0x80
-                                          // rr_page_ip_in_traced_syscall:
-                                          0xc3 // ret
+      static const uint8_t x86_data[] = {
+        0x90, 0x90, // padding
+        // rr_page_untraced_syscall_ip:
+        0xcd, 0x80, // int 0x80
+        // rr_page_ip_in_untraced_syscall:
+        0xc3, // ret
+        0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90,
+        0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, // padding
+        // rr_page_traced_syscall_ip:
+        0xcd, 0x80, // int 0x80
+        // rr_page_ip_in_traced_syscall:
+        0xc3 // ret
       };
       ASSERT(t, sizeof(x86_data) == write(fd, x86_data, sizeof(x86_data)));
       break;
     }
     case x86_64:
+      // See Task::did_waitpid for an explanation of why we have to
+      // modify R11 and RCX here.
       static const uint8_t x86_64_data[] = {
         0x90, 0x90, // padding
         // rr_page_untraced_syscall_ip:
@@ -443,8 +315,10 @@ static void write_rr_page(Task* t, ScopedFd& fd) {
         // rr_page_ip_in_untraced_syscall:
         0x49, 0x81, 0xe3, 0xff,
         0xfe, 0xff, 0xff, // and $0xfffffffffffffeff,%r11
+        0x48, 0xc7, 0xc1, 0xff,
+        0xff, 0xff, 0xff, // mov $-1,%rcx
         0xc3,             // ret
-        0x90, 0x90,       // padding
+        0x90, 0x90, 0x90, // padding
         // rr_page_traced_syscall_ip:
         0x0f, 0x05, // syscall
         // rr_page_ip_in_traced_syscall:
@@ -474,8 +348,10 @@ void AddressSpace::map_rr_page(Task* t) {
 
     AutoRestoreMem child_path(remote, reinterpret_cast<uint8_t*>(path),
                               sizeof(path));
-    int child_fd = remote.syscall(syscall_number_for_open(arch),
-                                  child_path.get(), O_RDONLY);
+    // skip leading '/' since we want the path to be relative to the root fd
+    int child_fd =
+        remote.syscall(syscall_number_for_openat(arch), RR_RESERVED_ROOT_DIR_FD,
+                       child_path.get() + 1, O_RDONLY);
     ASSERT(t, child_fd >= 0);
 
     auto result = remote.mmap_syscall(rr_page_start(), rr_page_size(), prot,
@@ -511,11 +387,13 @@ void AddressSpace::brk(remote_ptr<void> addr) {
   LOG(debug) << "brk(" << addr << ")";
 
   assert(heap.start <= addr);
-  if (addr == heap.end) {
+
+  remote_ptr<void> vm_addr = ceil_page_size(addr);
+  if (vm_addr == heap.end) {
     return;
   }
 
-  update_heap(heap.start, addr);
+  update_heap(heap.start, vm_addr);
   map(heap.start, heap.num_bytes(), heap.prot, heap.flags, heap.offset,
       MappableResource::heap());
 }
@@ -538,7 +416,20 @@ TrapType AddressSpace::get_breakpoint_type_for_retired_insn(
 
 TrapType AddressSpace::get_breakpoint_type_at_addr(remote_ptr<uint8_t> addr) {
   auto it = breakpoints.find(addr);
-  return it == breakpoints.end() ? TRAP_NONE : it->second->type();
+  return it == breakpoints.end() ? TRAP_NONE : it->second.type();
+}
+
+void AddressSpace::replace_breakpoints_with_original_values(
+    uint8_t* dest, size_t length, remote_ptr<uint8_t> addr) {
+  for (auto& it : breakpoints) {
+    remote_ptr<uint8_t> start = max(addr, it.first);
+    remote_ptr<uint8_t> end =
+        min(addr + length, it.first + it.second.data_length());
+    if (start < end) {
+      memcpy(dest + (start - addr),
+             it.second.original_data() + (start - it.first), end - start);
+    }
+  }
 }
 
 void AddressSpace::map(remote_ptr<void> addr, size_t num_bytes, int prot,
@@ -550,6 +441,15 @@ void AddressSpace::map(remote_ptr<void> addr, size_t num_bytes, int prot,
   num_bytes = ceil_page_size(num_bytes);
 
   Mapping m(addr, num_bytes, prot, flags, offset_bytes);
+
+  bool insert_guard_page = false;
+  if (has_mapping(m.end) && mapping_of(m.end).second.is_stack()) {
+    // When inserting a mapping immediately before a grow-down VMA,
+    // the kernel unmaps an extra page to form a guard page. We need to
+    // emulate that.
+    insert_guard_page = true;
+  }
+
   if (mem.end() != mem.find(m)) {
     // The mmap() man page doesn't specifically describe
     // what should happen if an existing map is
@@ -557,7 +457,7 @@ void AddressSpace::map(remote_ptr<void> addr, size_t num_bytes, int prot,
     // In testing, the behavior seems to be as if the
     // overlapping region is unmapped and then remapped
     // per the arguments to the second call.
-    unmap(addr, num_bytes);
+    unmap(addr, num_bytes + (insert_guard_page ? page_size() : 0));
   }
 
   map_and_coalesce(m, res);
@@ -597,16 +497,24 @@ void AddressSpace::at_preload_init(Task* t) {
 }
 
 typedef AddressSpace::MemoryMap::value_type MappingResourcePair;
-MappingResourcePair AddressSpace::mapping_of(remote_ptr<void> addr,
-                                             size_t num_bytes) const {
-  auto it = mem.find(Mapping(addr, num_bytes));
+MappingResourcePair AddressSpace::mapping_of(remote_ptr<void> addr) const {
+  Mapping m(floor_page_size(addr), page_size());
+  auto it = mem.find(m);
   assert(it != mem.end());
-  // TODO callers assume [addr, addr + num_bytes] doesn't cross
-  // resource boundaries
-  assert(it->first.has_subset(Mapping(addr, num_bytes)));
+  assert(it->first.has_subset(m));
   return *it;
 }
 
+bool AddressSpace::has_mapping(remote_ptr<void> addr) const {
+  if (addr + page_size() < addr) {
+    // Assume the last byte in the address space is never mapped; avoid overflow
+    return false;
+  }
+  Mapping m(floor_page_size(addr), page_size());
+  auto it = mem.find(m);
+  return it != mem.end() && it->first.has_subset(m);
+}
+
 /**
  * Return the offset of |m| into |r| updated by |delta|, unless |r| is
  * a pseudo-device that doesn't have offsets, in which case the
@@ -663,7 +571,7 @@ void AddressSpace::remap(remote_ptr<void> old_addr, size_t old_num_bytes,
   LOG(debug) << "mremap(" << old_addr << ", " << old_num_bytes << ", "
              << new_addr << ", " << new_num_bytes << ")";
 
-  auto mr = mapping_of(old_addr, old_num_bytes);
+  auto mr = mapping_of(old_addr);
   const Mapping& m = mr.first;
   const MappableResource& r = mr.second;
 
@@ -679,71 +587,175 @@ void AddressSpace::remap(remote_ptr<void> old_addr, size_t old_num_bytes,
 
 void AddressSpace::remove_breakpoint(remote_ptr<uint8_t> addr, TrapType type) {
   auto it = breakpoints.find(addr);
-  if (it == breakpoints.end() || !it->second || it->second->unref(type) > 0) {
+  if (it == breakpoints.end() || it->second.unref(type) > 0) {
     return;
   }
   destroy_breakpoint(it);
 }
 
-bool AddressSpace::set_breakpoint(remote_ptr<uint8_t> addr, TrapType type) {
+bool AddressSpace::add_breakpoint(remote_ptr<uint8_t> addr, TrapType type) {
   auto it = breakpoints.find(addr);
   if (it == breakpoints.end()) {
-    auto bp = Breakpoint::create();
+    uint8_t overwritten_data;
     // Grab a random task from the VM so we can use its
     // read/write_mem() helpers.
     Task* t = *task_set().begin();
-    if (sizeof(bp->overwritten_data) !=
-        t->read_bytes_fallible(addr, sizeof(bp->overwritten_data),
-                               &bp->overwritten_data)) {
+    if (sizeof(overwritten_data) !=
+        t->read_bytes_fallible(addr, sizeof(overwritten_data),
+                               &overwritten_data)) {
       return false;
     }
     t->write_mem(addr, breakpoint_insn);
 
-    auto it_and_is_new = breakpoints.insert(make_pair(addr, bp));
+    auto it_and_is_new = breakpoints.insert(make_pair(addr, Breakpoint()));
     assert(it_and_is_new.second);
+    it_and_is_new.first->second.overwritten_data = overwritten_data;
     it = it_and_is_new.first;
   }
-  it->second->ref(type);
+  it->second.ref(type);
   return true;
 }
 
-void AddressSpace::destroy_all_breakpoints() {
+void AddressSpace::remove_all_breakpoints() {
   while (!breakpoints.empty()) {
     destroy_breakpoint(breakpoints.begin());
   }
 }
 
+int AddressSpace::access_bits_of(WatchType type) {
+  switch (type) {
+    case WATCH_EXEC:
+      return EXEC_BIT;
+    case WATCH_WRITE:
+      return WRITE_BIT;
+    case WATCH_READWRITE:
+      return READ_BIT | WRITE_BIT;
+    default:
+      FATAL() << "Unknown watchpoint type " << type;
+      return 0; // not reached
+  }
+}
+
 void AddressSpace::remove_watchpoint(remote_ptr<void> addr, size_t num_bytes,
                                      WatchType type) {
   auto it = watchpoints.find(MemoryRange(addr, num_bytes));
   if (it != watchpoints.end() &&
-      0 == it->second->unwatch(access_bits_of(type))) {
+      0 == it->second.unwatch(access_bits_of(type))) {
     watchpoints.erase(it);
   }
   allocate_watchpoints();
 }
 
-bool AddressSpace::set_watchpoint(remote_ptr<void> addr, size_t num_bytes,
+bool AddressSpace::add_watchpoint(remote_ptr<void> addr, size_t num_bytes,
                                   WatchType type) {
   MemoryRange key(addr, num_bytes);
   auto it = watchpoints.find(key);
   if (it == watchpoints.end()) {
     auto it_and_is_new =
-        watchpoints.insert(make_pair(key, Watchpoint::create()));
+        watchpoints.insert(make_pair(key, Watchpoint(num_bytes)));
     assert(it_and_is_new.second);
     it = it_and_is_new.first;
+    update_watchpoint_value(it->first, it->second);
   }
-  it->second->watch(access_bits_of(type));
+  it->second.watch(access_bits_of(type));
   return allocate_watchpoints();
 }
 
-void AddressSpace::destroy_all_watchpoints() {
+void AddressSpace::save_watchpoints() {
+  saved_watchpoints.push_back(watchpoints);
+}
+
+bool AddressSpace::restore_watchpoints() {
+  assert(!saved_watchpoints.empty());
+  watchpoints = saved_watchpoints[saved_watchpoints.size() - 1];
+  saved_watchpoints.pop_back();
+  return allocate_watchpoints();
+}
+
+bool AddressSpace::update_watchpoint_value(const MemoryRange& range,
+                                           Watchpoint& watchpoint) {
+  Task* t = *task_set().begin();
+  bool valid = true;
+  vector<uint8_t> value_bytes = watchpoint.value_bytes;
+  for (size_t i = 0; i < value_bytes.size(); ++i) {
+    value_bytes[i] = 0xFF;
+  }
+  remote_ptr<void> addr = range.addr;
+  size_t num_bytes = range.num_bytes;
+  while (num_bytes > 0) {
+    ssize_t bytes_read = t->read_bytes_fallible(
+        addr, num_bytes, value_bytes.data() + (addr - range.addr));
+    if (bytes_read <= 0) {
+      valid = false;
+      // advance to next page and try to read more. We want to know
+      // when the valid part of a partially invalid watchpoint changes.
+      bytes_read =
+          min<size_t>(num_bytes, (floor_page_size(addr) + page_size()) - addr);
+    }
+    addr += bytes_read;
+    num_bytes -= bytes_read;
+  }
+
+  bool changed = valid != watchpoint.valid ||
+                 memcmp(value_bytes.data(), watchpoint.value_bytes.data(),
+                        value_bytes.size()) != 0;
+  watchpoint.valid = valid;
+  watchpoint.value_bytes = value_bytes;
+  return changed;
+}
+
+void AddressSpace::update_watchpoint_values(remote_ptr<void> start,
+                                            remote_ptr<void> end) {
+  MemoryRange r(start, end);
+  for (auto& it : watchpoints) {
+    if (it.first.intersects(r) &&
+        update_watchpoint_value(it.first, it.second)) {
+      it.second.changed = true;
+      // We do nothing to track kernel reads of read-write watchpoints...
+    }
+  }
+}
+
+static int DR_WATCHPOINT(int n) { return 1 << n; }
+
+static bool watchpoint_triggered(uintptr_t debug_status,
+                                 const vector<int8_t>& regs) {
+  for (auto reg : regs) {
+    if (debug_status & DR_WATCHPOINT(reg)) {
+      return true;
+    }
+  }
+  return false;
+}
+
+bool AddressSpace::notify_watchpoint_fired(uintptr_t debug_status) {
+  bool triggered = false;
+  for (auto& it : watchpoints) {
+    if (((it.second.watched_bits() & WRITE_BIT) &&
+         update_watchpoint_value(it.first, it.second)) ||
+        ((it.second.watched_bits() & (READ_BIT | EXEC_BIT)) &&
+         watchpoint_triggered(debug_status,
+                              it.second.debug_regs_for_exec_read))) {
+      it.second.changed = true;
+      triggered = true;
+    }
+  }
+  return triggered;
+}
+
+void AddressSpace::notify_written(remote_ptr<void> addr, size_t num_bytes) {
+  update_watchpoint_values(addr, addr + num_bytes);
+  session()->accumulate_bytes_written(num_bytes);
+}
+
+void AddressSpace::remove_all_watchpoints() {
   watchpoints.clear();
   allocate_watchpoints();
 }
 
 void AddressSpace::unmap(remote_ptr<void> addr, ssize_t num_bytes) {
   LOG(debug) << "munmap(" << addr << ", " << num_bytes << ")";
+  num_bytes = ceil_page_size(num_bytes);
 
   auto unmapper = [this](const Mapping& m, const MappableResource& r,
                          const Mapping& rem) {
@@ -756,7 +768,11 @@ void AddressSpace::unmap(remote_ptr<void> addr, ssize_t num_bytes) {
     // region, remap the underflow region.
     if (m.start < rem.start) {
       Mapping underflow(m.start, rem.start, m.prot, m.flags, m.offset);
-      mem[underflow] = r;
+      MappableResource new_r = r;
+      // When splitting a stack mapping, the bottom part of the split is no
+      // longer treated as stack by the kernel.
+      new_r.remove_stackness();
+      mem[underflow] = new_r;
     }
     // If the last segment we unmap overflows the unmap
     // region, remap the overflow region.
@@ -767,6 +783,7 @@ void AddressSpace::unmap(remote_ptr<void> addr, ssize_t num_bytes) {
     }
   };
   for_each_in_range(addr, num_bytes, unmapper);
+  update_watchpoint_values(addr, addr + num_bytes);
 }
 
 /**
@@ -803,6 +820,10 @@ static bool is_adjacent_mapping(const MappingResourcePair& left,
                << ": offsets into real device aren't adjacent)";
     return false;
   }
+  if (rleft.id.psuedodevice() == PSEUDODEVICE_SYSV_SHM) {
+    LOG(debug) << "    (SysV shm not coalescable)";
+    return false;
+  }
   LOG(debug) << "    adjacent!";
   return true;
 }
@@ -824,6 +845,25 @@ static bool try_merge_adjacent(Mapping* left_m, const MappableResource& left_r,
   return false;
 }
 
+static PseudoDevice pseudodevice_for_name(const string& name) {
+  if ("[heap]" == name) {
+    return PSEUDODEVICE_HEAP;
+  }
+  if ("[stack" == name.substr(0, 6)) {
+    return PSEUDODEVICE_STACK;
+  }
+  if ("[vdso]" == name) {
+    return PSEUDODEVICE_VDSO;
+  }
+  if ("" == name || "/dev/zero (deleted)" == name) {
+    return PSEUDODEVICE_ANONYMOUS;
+  }
+  if ("/SYSV" == name.substr(0, 5)) {
+    return PSEUDODEVICE_SYSV_SHM;
+  }
+  return PSEUDODEVICE_NONE;
+}
+
 /**
  * Iterate over /proc/maps segments for a task and verify that the
  * task's cached mapping matches the kernel's (given a lenient fuzz
@@ -865,16 +905,39 @@ void VerifyAddressSpace::assert_segments_match(Task* t) {
   assert(MERGING_KERNEL == phase);
   bool same_mapping = (m.start == km.start && m.end == km.end &&
                        m.prot == km.prot && m.flags == km.flags);
+  // When we stripped most identifying info from |r| with
+  // |to_kernel()|, we also lost its "is stack" flag.  So to check if
+  // it's a grows-down stack segment, we see if it has the special
+  // linux name for the process stack segment, "[stack]".
+  if (!same_mapping && km.start < m.start && "[stack]" == r.fsname) {
+    // TODO: the stack can grow down arbitrarily, and rr needs to be
+    // aware of the updated mapping in case the user tries to map or
+    // unmap pages near the stack.  But keeping track of expanded
+    // stack in general is somewhat difficult, because the stack grows
+    // without rr being notified.  So we just add a special early-exit
+    // case for the assert for now.
+    // We do fix up our current mapping to match the kernel as closely
+    // as possible. Then, if the grow-down VMA is split somehow, we know
+    // about the split parts.
+    t->vm()->fix_stack_segment_start(m, km.start);
+    return;
+  }
   if (!same_mapping) {
     LOG(error) << "cached mmap:";
     as->dump();
     LOG(error) << "/proc/" << t->tid << "/mmaps:";
     print_process_mmap(t);
 
-    ASSERT(t, same_mapping) << "\nCached mapping " << m << "should be " << km;
+    ASSERT(t, same_mapping) << "\nCached mapping " << m << " should be " << km;
   }
 }
 
+void AddressSpace::fix_stack_segment_start(const Mapping& mapping,
+                                           remote_ptr<void> new_start) {
+  auto it = mem.find(mapping);
+  it->first.update_start(new_start);
+}
+
 /**
  * Iterate over the segments that are parsed from
  * |/proc/[t->tid]/maps| and ensure that they match up with the cached
@@ -938,8 +1001,10 @@ void VerifyAddressSpace::assert_segments_match(Task* t) {
   assert(info.flags == (info.flags & Mapping::checkable_flags_mask));
   Mapping km(info.start_addr, info.end_addr, info.prot, info.flags,
              info.file_offset);
-  MappableResource kr(FileId(info.dev_major, info.dev_minor, info.inode),
-                      info.name);
+  PseudoDevice psdev = pseudodevice_for_name(info.name);
+  MappableResource kr = MappableResource(
+      FileId(info.dev_major, info.dev_minor, info.inode, psdev), info.name)
+                            .to_kernel();
 
   if (vas->INITING_KERNEL == vas->phase) {
     assert(kr == vas->r
@@ -955,6 +1020,7 @@ void VerifyAddressSpace::assert_segments_match(Task* t) {
            string::npos != kr.fsname.find(SHMEM_FS "/rr-emufs") ||
            string::npos != kr.fsname.find(SHMEM_FS2 "/rr-emufs"));
     vas->km = km;
+    vas->r = kr;
     vas->phase = vas->MERGING_KERNEL;
     return;
   }
@@ -973,7 +1039,7 @@ void VerifyAddressSpace::assert_segments_match(Task* t) {
 
 Mapping AddressSpace::vdso() const {
   assert(!vdso_start_addr.is_null());
-  return mapping_of(vdso_start_addr, 1).first;
+  return mapping_of(vdso_start_addr).first;
 }
 
 void AddressSpace::verify(Task* t) const {
@@ -986,11 +1052,17 @@ void AddressSpace::verify(Task* t) const {
   vas.assert_segments_match(t);
 }
 
-AddressSpace::AddressSpace(Task* t, const string& exe, Session& session)
-    : exe(exe), is_clone(false), session(&session), child_mem_fd(-1) {
+AddressSpace::AddressSpace(Task* t, const string& exe, uint32_t exec_count)
+    : exe(exe),
+      leader_tid_(t->rec_tid),
+      leader_serial(t->tuid().serial()),
+      exec_count(exec_count),
+      is_clone(false),
+      session_(&t->session()),
+      child_mem_fd(-1) {
   // TODO: this is a workaround of
   // https://github.com/mozilla/rr/issues/1113 .
-  if (session.can_validate()) {
+  if (session_->can_validate()) {
     iterate_memory_map(t, populate_address_space, this);
     assert(!vdso_start_addr.is_null());
   } else {
@@ -1011,50 +1083,198 @@ AddressSpace::AddressSpace(Task* t, const string& exe, Session& session)
   }
 }
 
-AddressSpace::AddressSpace(const AddressSpace& o)
-    // Whether the new VM wants our breakpoints our not,
-    // it's going to inherit them.  This is pretty much
-    // never what anyone wants, so a call to
-    // |remove_all_breakpoints()| is expected soon after
-    // the creation of this.
-    : breakpoints(o.breakpoints),
-      exe(o.exe),
+AddressSpace::AddressSpace(Task* t, const AddressSpace& o, pid_t leader_tid,
+                           uint32_t leader_serial, uint32_t exec_count)
+    : exe(o.exe),
+      leader_tid_(leader_tid),
+      leader_serial(leader_serial),
+      exec_count(exec_count),
       heap(o.heap),
       is_clone(true),
       mem(o.mem),
-      session(nullptr),
+      session_(nullptr),
       vdso_start_addr(o.vdso_start_addr),
       monkeypatch_state(o.monkeypatch_state),
       traced_syscall_ip_(o.traced_syscall_ip_),
       untraced_syscall_ip_(o.untraced_syscall_ip_),
       syscallbuf_lib_start_(o.syscallbuf_lib_start_),
       syscallbuf_lib_end_(o.syscallbuf_lib_end_) {
-  for (auto it = breakpoints.begin(); it != breakpoints.end(); ++it) {
-    it->second = it->second->clone();
+  for (auto& it : o.breakpoints) {
+    breakpoints.insert(make_pair(it.first, it.second));
+  }
+  for (auto& it : o.watchpoints) {
+    watchpoints.insert(make_pair(it.first, it.second));
+  }
+  // cloned tasks will automatically get cloned debug registers and
+  // cloned address-space memory, so we don't need to do any more work here.
+}
+
+void AddressSpace::copy_user_breakpoints_from(const AddressSpace& o) {
+  vector<remote_ptr<uint8_t> > addrs_to_remove;
+  for (auto& it : breakpoints) {
+    for (int i = 0; i < it.second.user_count; ++i) {
+      addrs_to_remove.push_back(it.first);
+    }
+  }
+  for (auto a : addrs_to_remove) {
+    remove_breakpoint(a, TRAP_BKPT_USER);
+  }
+
+  for (auto& it : o.breakpoints) {
+    for (int i = 0; i < it.second.user_count; ++i) {
+      add_breakpoint(it.first, TRAP_BKPT_USER);
+    }
   }
 }
 
-bool AddressSpace::allocate_watchpoints() {
-  Task::DebugRegs regs;
-  for (auto kv : watchpoints) {
-    const MemoryRange& r = kv.first;
-    int watching = kv.second->watched_bits();
-    if (EXEC_BIT & watching) {
-      regs.push_back(WatchConfig(r.addr, r.num_bytes, WATCH_EXEC));
+void AddressSpace::copy_watchpoints_from(const AddressSpace& o) {
+  watchpoints.clear();
+  for (auto& it : o.watchpoints) {
+    watchpoints.insert(make_pair(it.first, it.second));
+  }
+  allocate_watchpoints();
+}
+
+static bool try_split_unaligned_range(MemoryRange& range, size_t bytes,
+                                      vector<MemoryRange>& result) {
+  if ((range.addr.as_int() & (bytes - 1)) || range.num_bytes < bytes) {
+    return false;
+  }
+  result.push_back(MemoryRange(range.addr, bytes));
+  range.addr += bytes;
+  range.num_bytes -= bytes;
+  return true;
+}
+
+static vector<MemoryRange> split_range(const MemoryRange& range) {
+  vector<MemoryRange> result;
+  MemoryRange r = range;
+  while (r.num_bytes > 0) {
+    if ((sizeof(void*) < 8 || !try_split_unaligned_range(r, 8, result)) &&
+        !try_split_unaligned_range(r, 4, result) &&
+        !try_split_unaligned_range(r, 2, result)) {
+      bool ret = try_split_unaligned_range(r, 1, result);
+      assert(ret);
     }
-    if (!(READ_BIT & watching) && (WRITE_BIT & watching)) {
-      regs.push_back(WatchConfig(r.addr, r.num_bytes, WATCH_WRITE));
+  }
+  return result;
+}
+
+static void configure_watch_registers(vector<WatchConfig>& regs,
+                                      const MemoryRange& range, WatchType type,
+                                      vector<int8_t>* assigned_regs) {
+  auto split_ranges = split_range(range);
+
+  if (type == WATCH_WRITE && range.num_bytes > 1) {
+    // We can suppress spurious write-watchpoint triggerings by checking
+    // whether memory values have changed. So we can sometimes conserve
+    // debug registers by upgrading an unaligned range to an aligned range
+    // of a larger size.
+    uintptr_t align;
+    if (range.num_bytes <= 2) {
+      align = 2;
+    } else if (range.num_bytes <= 4 || sizeof(void*) <= 4) {
+      align = 4;
+    } else {
+      align = 8;
+    }
+    remote_ptr<void> aligned_start(range.addr.as_int() & ~(align - 1));
+    remote_ptr<void> aligned_end((range.end().as_int() + (align - 1)) &
+                                 ~(align - 1));
+    auto split = split_range(MemoryRange(aligned_start, aligned_end));
+    // If the aligned range doesn't reduce register usage, use the original
+    // split to avoid spurious triggerings
+    if (split.size() < split_ranges.size()) {
+      split_ranges = split;
+    }
+  }
+
+  for (auto& r : split_ranges) {
+    if (assigned_regs) {
+      assigned_regs->push_back(regs.size());
+    }
+    regs.push_back(WatchConfig(r.addr, r.num_bytes, type));
+  }
+}
+
+vector<WatchConfig> AddressSpace::get_watch_configs(
+    WillSetTaskState will_set_task_state) {
+  vector<WatchConfig> result;
+  for (auto& kv : watchpoints) {
+    vector<int8_t>* assigned_regs = nullptr;
+    if (will_set_task_state == SETTING_TASK_STATE) {
+      kv.second.debug_regs_for_exec_read.clear();
+      assigned_regs = &kv.second.debug_regs_for_exec_read;
+    }
+    const MemoryRange& r = kv.first;
+    int watching = kv.second.watched_bits();
+    if (EXEC_BIT & watching) {
+      configure_watch_registers(result, r, WATCH_EXEC, assigned_regs);
     }
     if (READ_BIT & watching) {
-      regs.push_back(WatchConfig(r.addr, r.num_bytes, WATCH_READWRITE));
+      configure_watch_registers(result, r, WATCH_READWRITE, assigned_regs);
+    } else if (WRITE_BIT & watching) {
+      configure_watch_registers(result, r, WATCH_WRITE, nullptr);
     }
   }
-  for (auto t : task_set()) {
-    if (!t->set_debug_regs(regs)) {
-      return false;
+  return result;
+}
+
+vector<WatchConfig> AddressSpace::get_watchpoints_internal(
+    WatchpointFilter filter) {
+  vector<WatchConfig> result;
+  for (auto& kv : watchpoints) {
+    if (filter == CHANGED_WATCHPOINTS) {
+      if (!kv.second.changed) {
+        continue;
+      }
+      kv.second.changed = false;
+    }
+    const MemoryRange& r = kv.first;
+    int watching = kv.second.watched_bits();
+    if (EXEC_BIT & watching) {
+      result.push_back(WatchConfig(r.addr, r.num_bytes, WATCH_EXEC));
+    }
+    if (READ_BIT & watching) {
+      result.push_back(WatchConfig(r.addr, r.num_bytes, WATCH_READWRITE));
+    } else if (WRITE_BIT & watching) {
+      result.push_back(WatchConfig(r.addr, r.num_bytes, WATCH_WRITE));
     }
   }
-  return true;
+  return result;
+}
+
+vector<WatchConfig> AddressSpace::consume_watchpoint_changes() {
+  return get_watchpoints_internal(CHANGED_WATCHPOINTS);
+}
+
+vector<WatchConfig> AddressSpace::all_watchpoints() {
+  return get_watchpoints_internal(ALL_WATCHPOINTS);
+}
+
+bool AddressSpace::allocate_watchpoints() {
+  Task::DebugRegs regs = get_watch_configs(SETTING_TASK_STATE);
+
+  if (regs.size() <= 0x7f) {
+    bool ok = true;
+    for (auto t : task_set()) {
+      if (!t->set_debug_regs(regs)) {
+        ok = false;
+      }
+    }
+    if (ok) {
+      return true;
+    }
+  }
+
+  regs.clear();
+  for (auto t2 : task_set()) {
+    t2->set_debug_regs(regs);
+  }
+  for (auto kv : watchpoints) {
+    kv.second.debug_regs_for_exec_read.clear();
+  }
+  return false;
 }
 
 void AddressSpace::coalesce_around(MemoryMap::iterator it) {
@@ -1095,7 +1315,7 @@ void AddressSpace::coalesce_around(MemoryMap::iterator it) {
 
 void AddressSpace::destroy_breakpoint(BreakpointMap::const_iterator it) {
   Task* t = *task_set().begin();
-  t->write_mem(it->first, it->second->overwritten_data);
+  t->write_mem(it->first, it->second.overwritten_data);
   breakpoints.erase(it);
 }
 
@@ -1104,9 +1324,9 @@ void AddressSpace::for_each_in_range(
     function<void(const Mapping& m, const MappableResource& r,
                   const Mapping& rem)> f,
     int how) {
-  num_bytes = ceil_page_size(num_bytes);
-  remote_ptr<void> last_unmapped_end = addr;
-  remote_ptr<void> region_end = addr + num_bytes;
+  remote_ptr<void> region_start = floor_page_size(addr);
+  remote_ptr<void> last_unmapped_end = region_start;
+  remote_ptr<void> region_end = ceil_page_size(addr + num_bytes);
   while (last_unmapped_end < region_end) {
     // Invariant: |rem| is always exactly the region of
     // memory remaining to be examined for pages to be
@@ -1127,7 +1347,7 @@ void AddressSpace::for_each_in_range(
       return;
     }
     if (ITERATE_CONTIGUOUS == how &&
-        !(m.start < addr || rem.start == m.start)) {
+        !(m.start < region_start || rem.start == m.start)) {
       LOG(debug) << "  discontiguous mapping at " << m.start << ", done.";
       return;
     }
@@ -1140,6 +1360,13 @@ void AddressSpace::for_each_in_range(
   }
 }
 
+void AddressSpace::for_all_mappings(
+    std::function<void(const Mapping& m, const MappableResource& r)> f) {
+  for (auto& m : mem) {
+    f(m.first, m.second);
+  }
+}
+
 void AddressSpace::map_and_coalesce(const Mapping& m,
                                     const MappableResource& r) {
   LOG(debug) << "  mapping " << m;
@@ -1147,6 +1374,8 @@ void AddressSpace::map_and_coalesce(const Mapping& m,
   auto ins = mem.insert(MemoryMap::value_type(m, r));
   assert(ins.second); // key didn't already exist
   coalesce_around(ins.first);
+
+  update_watchpoint_values(m.start, m.end);
 }
 
 /*static*/ void AddressSpace::populate_address_space(
@@ -1161,35 +1390,34 @@ void AddressSpace::map_and_coalesce(const Mapping& m,
                << " (end of text segment)";
   }
 
-  bool is_dynamic_heap = "[heap]" == info.name;
-  // This segment is adjacent to our previous guess at the start
-  // of the dynamic heap, but it's still not an explicit heap
-  // segment.  Update the guess.
-  if (as->heap.end == info.start_addr && !(info.prot & PROT_EXEC)) {
-    assert(as->heap.start == as->heap.end);
+  PseudoDevice psdev = pseudodevice_for_name(info.name);
+
+  // This segment is adjacent to our previous guess at the start of
+  // the dynamic heap, but it's still not an explicit heap segment.
+  // Or, in corner cases, the segment is the final mapping of the data
+  // segment of the exe image, but is not adjacent to the prior mapped
+  // segment of the exe.  (This is seen with x86-64 bash on Fedora
+  // Core 20.)  Update the guess.
+  if (!(info.prot & PROT_EXEC) &&
+      (as->heap.end == info.start_addr || as->exe == info.name)) {
+    assert(as->heap.start == as->heap.end || as->exe == info.name);
     as->update_heap(info.end_addr, info.end_addr);
     LOG(debug) << "  updating start-of-heap guess to " << as->heap.start
                << " (end of mapped-data segment)";
   }
 
-  FileId id;
-  if (is_dynamic_heap) {
-    id = FileId(PSEUDODEVICE_HEAP);
+  if (psdev == PSEUDODEVICE_HEAP) {
     if (!as->heap.start) {
       // No guess for the heap start. Assume it's just the [heap] segment.
       as->update_heap(info.start_addr, info.end_addr);
     } else {
       as->update_heap(as->heap.start, info.end_addr);
     }
-  } else if ("[stack]" == info.name) {
-    id = FileId(PSEUDODEVICE_STACK);
-  } else if ("[vdso]" == info.name) {
+  } else if (psdev == PSEUDODEVICE_VDSO) {
     assert(!as->vdso_start_addr);
     as->vdso_start_addr = info.start_addr;
-    id = FileId(PSEUDODEVICE_VDSO);
-  } else {
-    id = FileId(MKDEV(info.dev_major, info.dev_minor), info.inode);
   }
+  FileId id = FileId(MKDEV(info.dev_major, info.dev_minor), info.inode, psdev);
 
   as->map(info.start_addr, info.end_addr - info.start_addr, info.prot,
           info.flags, info.file_offset, MappableResource(id, info.name));
