@@ -14,6 +14,7 @@
 
 #include "kernel_abi.h"
 #include "Monkeypatcher.h"
+#include "TaskishUid.h"
 #include "TraceStream.h"
 #include "util.h"
 
@@ -50,6 +51,7 @@ enum PseudoDevice {
   PSEUDODEVICE_STACK,
   PSEUDODEVICE_SYSCALLBUF,
   PSEUDODEVICE_VDSO,
+  PSEUDODEVICE_SYSV_SHM
 };
 
 /**
@@ -82,6 +84,7 @@ public:
    * is false, return 0 (NO_INODE).
    */
   ino_t disp_inode() const;
+  ino_t internal_inode() const { return inode; }
   PseudoDevice psuedodevice() const { return psdev; }
 
   /**
@@ -97,14 +100,16 @@ public:
     if (psdev == PSEUDODEVICE_ANONYMOUS) {
       return true;
     }
-    if (dev_major() != o.dev_major()) {
-      return false;
-    }
-    // Allow device minor numbers to vary if the major device is
-    // 0. This was observed to be happening on
-    // "3.13.0-24-generic #46-Ubuntu SMP" in KVM with btrfs.
-    if (dev_major() != 0 && dev_minor() != o.dev_minor()) {
-      return false;
+    if (psdev != PSEUDODEVICE_SYSV_SHM) {
+      if (dev_major() != o.dev_major()) {
+        return false;
+      }
+      // Allow device minor numbers to vary if the major device is
+      // 0. This was observed to be happening on
+      // "3.13.0-24-generic #46-Ubuntu SMP" in KVM with btrfs.
+      if (dev_major() != 0 && dev_minor() != o.dev_minor()) {
+        return false;
+      }
     }
     return inode == o.inode;
   }
@@ -112,7 +117,9 @@ public:
    * Return true if this file is/was backed by an external
    * device, as opposed to a transient RAM mapping.
    */
-  bool is_real_device() const { return device > NO_DEVICE; }
+  bool is_real_device() const {
+    return device > NO_DEVICE || psdev == PSEUDODEVICE_SYSV_SHM;
+  }
   const char* special_name() const;
 
   bool operator<(const FileId& o) const {
@@ -231,6 +238,14 @@ struct Mapping {
     return str;
   }
 
+  /**
+   * Super-dangerous! Only call this if it's guaranteed not to change
+   * the ordering of Mappings.
+   */
+  void update_start(remote_ptr<void> new_start) const {
+    *const_cast<remote_ptr<void>*>(&start) = new_start;
+  }
+
   const remote_ptr<void> start;
   const remote_ptr<void> end;
   const int prot;
@@ -271,17 +286,41 @@ struct MappableResource {
   bool operator!=(const MappableResource& o) const { return !(*this == o); }
   bool is_scratch() const { return PSEUDODEVICE_SCRATCH == id.psuedodevice(); }
   bool is_shared_mmap_file() const {
-    return PSEUDODEVICE_SHARED_MMAP_FILE == id.psuedodevice();
+    return PSEUDODEVICE_SHARED_MMAP_FILE == id.psuedodevice() ||
+           PSEUDODEVICE_SYSV_SHM == id.psuedodevice();
   }
   bool is_stack() const { return PSEUDODEVICE_STACK == id.psuedodevice(); }
 
+  void remove_stackness() {
+    if (is_stack()) {
+      id = FileId(id.dev_major(), id.dev_minor(), id.internal_inode(),
+                  PSEUDODEVICE_ANONYMOUS);
+      fsname = "";
+    }
+  }
+
   /**
    * Return a representation of this resource that would be
    * parsed from /proc/maps if this were mapped.
    */
   MappableResource to_kernel() const {
+    PseudoDevice psdev;
+    switch (id.psuedodevice()) {
+      case PSEUDODEVICE_STACK:
+      case PSEUDODEVICE_SCRATCH:
+      case PSEUDODEVICE_ANONYMOUS:
+        psdev = PSEUDODEVICE_ANONYMOUS;
+        break;
+      case PSEUDODEVICE_SYSV_SHM:
+        psdev = PSEUDODEVICE_SYSV_SHM;
+        break;
+      default:
+        psdev = PSEUDODEVICE_NONE;
+        break;
+    }
+
     return MappableResource(
-        FileId(id.dev_major(), id.dev_minor(), id.disp_inode()),
+        FileId(id.dev_major(), id.dev_minor(), id.disp_inode(), psdev),
         fsname.c_str());
   }
 
@@ -311,6 +350,10 @@ struct MappableResource {
         FileId(FileId::NO_DEVICE, tid, PSEUDODEVICE_SCRATCH), "[scratch]");
   }
   static MappableResource shared_mmap_file(const TraceMappedRegion& file);
+  static MappableResource shared_mmap_anonymous(uint32_t unique_id) {
+    return MappableResource(
+        FileId(FileId::NO_DEVICE, unique_id, PSEUDODEVICE_SHARED_MMAP_FILE));
+  }
   static MappableResource stack(pid_t tid) {
     return MappableResource(FileId(FileId::NO_DEVICE, tid, PSEUDODEVICE_STACK),
                             "[stack]");
@@ -327,8 +370,6 @@ struct MappableResource {
   static ino_t nr_anonymous_maps;
 };
 
-struct Breakpoint;
-
 enum TrapType {
   TRAP_NONE = 0,
   // Trap for debugger 'stepi' request.
@@ -340,11 +381,6 @@ enum TrapType {
   TRAP_BKPT_USER,
 };
 
-// XXX one is tempted to merge Breakpoint and Watchpoint into a single
-// entity, but the semantics are just different enough that separate
-// objects are easier for now.
-class Watchpoint;
-
 enum WatchType {
   // NB: these random-looking enumeration values are chosen to
   // match the numbers programmed into x86 debug registers.
@@ -354,10 +390,6 @@ enum WatchType {
 };
 
 enum DebugStatus {
-  DS_WATCHPOINT0 = 1 << 0,
-  DS_WATCHPOINT1 = 1 << 1,
-  DS_WATCHPOINT2 = 1 << 2,
-  DS_WATCHPOINT3 = 1 << 3,
   DS_WATCHPOINT_ANY = 0xf,
   DS_SINGLESTEP = 1 << 14,
 };
@@ -368,6 +400,8 @@ enum DebugStatus {
 struct MemoryRange {
   MemoryRange(remote_ptr<void> addr, size_t num_bytes)
       : addr(addr), num_bytes(num_bytes) {}
+  MemoryRange(remote_ptr<void> addr, remote_ptr<void> end)
+      : addr(addr), num_bytes(end - addr) {}
   MemoryRange(const MemoryRange&) = default;
   MemoryRange& operator=(const MemoryRange&) = default;
 
@@ -378,6 +412,14 @@ struct MemoryRange {
     return addr != o.addr ? addr < o.addr : num_bytes < o.num_bytes;
   }
 
+  bool intersects(const MemoryRange& other) const {
+    remote_ptr<void> s = std::max(addr, other.addr);
+    remote_ptr<void> e = std::min(end(), other.end());
+    return s < e;
+  }
+
+  remote_ptr<void> end() const { return addr + num_bytes; }
+
   remote_ptr<void> addr;
   size_t num_bytes;
 };
@@ -403,11 +445,8 @@ class AddressSpace : public HasTaskSet {
   friend struct VerifyAddressSpace;
 
 public:
-  typedef std::map<remote_ptr<uint8_t>, std::shared_ptr<Breakpoint> >
-  BreakpointMap;
   typedef std::map<Mapping, MappableResource, MappingComparator> MemoryMap;
   typedef std::shared_ptr<AddressSpace> shr_ptr;
-  typedef std::map<MemoryRange, std::shared_ptr<Watchpoint> > WatchpointMap;
 
   ~AddressSpace();
 
@@ -443,6 +482,20 @@ public:
    */
   bool execed() const { return !is_clone; }
 
+  /**
+   * Return tid of the first task for this address space.
+   */
+  pid_t leader_tid() const { return leader_tid_; }
+
+  /**
+   * Return AddressSpaceUid for this address space.
+   */
+  AddressSpaceUid uid() const {
+    return AddressSpaceUid(leader_tid_, leader_serial, exec_count);
+  }
+
+  Session* session() const { return session_; }
+
   /**
    * Return the path this address space was exec()'d with.
    */
@@ -462,6 +515,14 @@ public:
    */
   TrapType get_breakpoint_type_at_addr(remote_ptr<uint8_t> addr);
 
+  /**
+   * The buffer |dest| of length |length| represents the contents of tracee
+   * memory at |addr|. Replace the bytes in |dest| that have been overwritten
+   * by breakpoints with the original data that was replaced by the breakpoints.
+   */
+  void replace_breakpoints_with_original_values(uint8_t* dest, size_t length,
+                                                remote_ptr<uint8_t> addr);
+
   /**
    * Map |num_bytes| into this address space at |addr|, with
    * |prot| protection and |flags|.  The pages are (possibly
@@ -471,11 +532,15 @@ public:
            off64_t offset_bytes, const MappableResource& res);
 
   /**
-   * Return the mapping and mapped resource that underly [addr,
-   * addr + num_bytes).  There must be exactly one such mapping.
+   * Return the mapping and mapped resource for the byte at address 'addr'.
+   * There must be such a mapping.
    */
-  MemoryMap::value_type mapping_of(remote_ptr<void> addr,
-                                   size_t num_bytes) const;
+  MemoryMap::value_type mapping_of(remote_ptr<void> addr) const;
+
+  /**
+   * Return true if there is some mapping for the byte at 'addr'.
+   */
+  bool has_mapping(remote_ptr<void> addr) const;
 
   /**
    * Return the memory map.
@@ -495,31 +560,78 @@ public:
   void remap(remote_ptr<void> old_addr, size_t old_num_bytes,
              remote_ptr<void> new_addr, size_t new_num_bytes);
 
+  /**
+   * Notify that the stack segment 'mapping' has grown down to a new start
+   * address.
+   */
+  void fix_stack_segment_start(const Mapping& mapping,
+                               remote_ptr<void> new_start);
+
+  /**
+   * Notify that data was written to this address space by rr or
+   * by the kernel.
+   */
+  void notify_written(remote_ptr<void> addr, size_t num_bytes);
+
+  /** Ensure a breakpoint of |type| is set at |addr|. */
+  bool add_breakpoint(remote_ptr<uint8_t> addr, TrapType type);
   /**
    * Remove a |type| reference to the breakpoint at |addr|.  If
    * the removed reference was the last, the breakpoint is
    * destroyed.
    */
   void remove_breakpoint(remote_ptr<uint8_t> addr, TrapType type);
-
-  /** Ensure a breakpoint of |type| is set at |addr|. */
-  bool set_breakpoint(remote_ptr<uint8_t> addr, TrapType type);
-
   /**
    * Destroy all breakpoints in this VM, regardless of their
    * reference counts.
    */
-  void destroy_all_breakpoints();
+  void remove_all_breakpoints();
 
   /**
    * Manage watchpoints.  Analogous to breakpoint-managing
    * methods above, except that watchpoints can be set for an
    * address range.
    */
+  bool add_watchpoint(remote_ptr<void> addr, size_t num_bytes, WatchType type);
   void remove_watchpoint(remote_ptr<void> addr, size_t num_bytes,
                          WatchType type);
-  bool set_watchpoint(remote_ptr<void> addr, size_t num_bytes, WatchType type);
-  void destroy_all_watchpoints();
+  void remove_all_watchpoints();
+  std::vector<WatchConfig> all_watchpoints();
+
+  /**
+   * Save all watchpoint state onto a stack.
+   */
+  void save_watchpoints();
+  /**
+   * Pop all watchpoint state from the saved-state stack.
+   */
+  bool restore_watchpoints();
+
+  /**
+   * Notify that at least one watchpoint was hit --- recheck them all.
+   * Returns true if any watchpoint actually triggered. Note that
+   * debug_status can indicate a hit watchpoint that doesn't actually
+   * trigger, because the value of a write-watchpoint did not change.
+   * Likewise, debug_status can indicate a watchpoint wasn't hit that
+   * actually was (because in some configurations, e.g. VMWare
+   * hypervisor with 32-bit x86 guest, debug_status watchpoint bits
+   * are known to not be set on singlestep).
+   */
+  bool notify_watchpoint_fired(uintptr_t debug_status);
+  /**
+   * Return all changed watchpoints in |watches| and clear their changed flags.
+   */
+  std::vector<WatchConfig> consume_watchpoint_changes();
+
+  /**
+   * Replace all our user breakpoints with the user breakpoints of 'o'.
+   * Asserts that there are no internal breakpoints currently set.
+   */
+  void copy_user_breakpoints_from(const AddressSpace& o);
+  /**
+   * Replace all our watchpoints with the watchpoints of 'o'.
+   */
+  void copy_watchpoints_from(const AddressSpace& o);
 
   /**
    * Make [addr, addr + num_bytes) inaccesible within this
@@ -536,6 +648,9 @@ public:
    */
   void verify(Task* t) const;
 
+  void for_all_mappings(
+      std::function<void(const Mapping& m, const MappableResource& r)> f);
+
   bool has_breakpoints() { return !breakpoints.empty(); }
   bool has_watchpoints() { return !watchpoints.empty(); }
 
@@ -619,11 +734,32 @@ public:
   remote_ptr<uint8_t> find_syscall_instruction(Task* t);
 
 private:
-  AddressSpace(Task* t, const std::string& exe, Session& session);
-  AddressSpace(const AddressSpace& o);
+  class Breakpoint;
+  typedef std::map<remote_ptr<uint8_t>, Breakpoint> BreakpointMap;
+  class Watchpoint;
+
+  AddressSpace(Task* t, const std::string& exe, uint32_t exec_count);
+  AddressSpace(Task* t, const AddressSpace& o, pid_t leader_tid,
+               uint32_t leader_serial, uint32_t exec_count);
 
   void map_rr_page(Task* t);
 
+  bool update_watchpoint_value(const MemoryRange& range,
+                               Watchpoint& watchpoint);
+  void update_watchpoint_values(remote_ptr<void> start, remote_ptr<void> end);
+  enum WatchpointFilter {
+    ALL_WATCHPOINTS,
+    CHANGED_WATCHPOINTS
+  };
+  std::vector<WatchConfig> get_watchpoints_internal(WatchpointFilter filter);
+
+  enum WillSetTaskState {
+    SETTING_TASK_STATE,
+    NOT_SETTING_TASK_STATE
+  };
+  std::vector<WatchConfig> get_watch_configs(
+      WillSetTaskState will_set_task_state);
+
   /**
    * Construct a minimal set of watchpoints to be enabled based
    * on |set_watchpoint()| calls, and program them for each task
@@ -679,11 +815,147 @@ private:
 
   template <typename Arch> void at_preload_init_arch(Task* t);
 
+  enum {
+    EXEC_BIT = 1 << 0,
+    READ_BIT = 1 << 1,
+    WRITE_BIT = 1 << 2
+  };
+
+  /** Return the access bits above needed to watch |type|. */
+  static int access_bits_of(WatchType type);
+
+  /**
+   * Represents a refcount set on a particular address.  Because there
+   * can be multiple refcounts of multiple types set on a single
+   * address, Breakpoint stores explicit USER and INTERNAL breakpoint
+   * refcounts.  Clients adding/removing breakpoints at this addr must
+   * call ref()/unref() as appropropiate.
+   */
+  struct Breakpoint {
+    Breakpoint() : internal_count(0), user_count(0) {}
+    Breakpoint(const Breakpoint& o) = default;
+    // AddressSpace::destroy_all_breakpoints() can cause this
+    // destructor to be invoked while we have nonzero total
+    // refcount, so the most we can assert is that the refcounts
+    // are valid.
+    ~Breakpoint() { assert(internal_count >= 0 && user_count >= 0); }
+
+    void ref(TrapType which) {
+      assert(internal_count >= 0 && user_count >= 0);
+      ++*counter(which);
+    }
+    int unref(TrapType which) {
+      assert(internal_count > 0 || user_count > 0);
+      --*counter(which);
+      assert(internal_count >= 0 && user_count >= 0);
+      return internal_count + user_count;
+    }
+
+    TrapType type() const {
+      // NB: USER breakpoints need to be processed before
+      // INTERNAL ones.  We want to give the debugger a
+      // chance to dispatch commands before we attend to the
+      // internal rr business.  So if there's a USER "ref"
+      // on the breakpoint, treat it as a USER breakpoint.
+      return user_count > 0 ? TRAP_BKPT_USER : TRAP_BKPT_INTERNAL;
+    }
+
+    size_t data_length() { return 1; }
+    uint8_t* original_data() { return &overwritten_data; }
+
+    // "Refcounts" of breakpoints set at |addr|.  The breakpoint
+    // object must be unique since we have to save the overwritten
+    // data, and we can't enforce the order in which breakpoints
+    // are set/removed.
+    int internal_count, user_count;
+    uint8_t overwritten_data;
+    static_assert(sizeof(overwritten_data) ==
+                      sizeof(AddressSpace::breakpoint_insn),
+                  "Must have the same size.");
+
+    int* counter(TrapType which) {
+      assert(TRAP_BKPT_INTERNAL == which || TRAP_BKPT_USER == which);
+      int* p = TRAP_BKPT_USER == which ? &user_count : &internal_count;
+      assert(*p >= 0);
+      return p;
+    }
+  };
+
+  // XXX one is tempted to merge Breakpoint and Watchpoint into a single
+  // entity, but the semantics are just different enough that separate
+  // objects are easier for now.
+  /**
+   * Track the watched accesses of a contiguous range of memory
+   * addresses.
+   */
+  class Watchpoint {
+  public:
+    Watchpoint(size_t num_bytes)
+        : exec_count(0),
+          read_count(0),
+          write_count(0),
+          value_bytes(num_bytes),
+          valid(false),
+          changed(false) {}
+    Watchpoint(const Watchpoint&) = default;
+    ~Watchpoint() { assert_valid(); }
+
+    void watch(int which) {
+      assert_valid();
+      exec_count += (EXEC_BIT & which) != 0;
+      read_count += (READ_BIT & which) != 0;
+      write_count += (WRITE_BIT & which) != 0;
+    }
+    int unwatch(int which) {
+      assert_valid();
+      if (EXEC_BIT & which) {
+        assert(exec_count > 0);
+        --exec_count;
+      }
+      if (READ_BIT & which) {
+        assert(read_count > 0);
+        --read_count;
+      }
+      if (WRITE_BIT & which) {
+        assert(write_count > 0);
+        --write_count;
+      }
+      return exec_count + read_count + write_count;
+    }
+
+    int watched_bits() const {
+      return (exec_count > 0 ? EXEC_BIT : 0) | (read_count > 0 ? READ_BIT : 0) |
+             (write_count > 0 ? WRITE_BIT : 0);
+    }
+
+    void assert_valid() const {
+      assert(exec_count >= 0 && read_count >= 0 && write_count >= 0);
+    }
+
+    // Watchpoints stay alive until all watched access typed have
+    // been cleared.  We track refcounts of each watchable access
+    // separately.
+    int exec_count, read_count, write_count;
+    // Debug registers allocated for read/exec access checking.
+    // Write watchpoints are always triggered by checking for actual memory
+    // value changes. Read/exec watchpoints can't be triggered that way, so
+    // we look for these registers being triggered instead.
+    std::vector<int8_t> debug_regs_for_exec_read;
+    std::vector<uint8_t> value_bytes;
+    bool valid;
+    bool changed;
+  };
+
   // All breakpoints set in this VM.
   BreakpointMap breakpoints;
   /* Path of the executable image this address space was
    * exec()'d with. */
   std::string exe;
+  /* Pid of first task for this address space */
+  pid_t leader_tid_;
+  /* Serial number of first task for this address space */
+  uint32_t leader_serial;
+  uint32_t exec_count;
   /* Track the special process-global heap in order to support
    * adjustments by brk(). */
   Mapping heap;
@@ -693,7 +965,7 @@ private:
   MemoryMap mem;
   // The session that created this.  We save a ref to it so that
   // we can notify it when we die.
-  Session* session;
+  Session* session_;
   /* First mapped byte of the vdso. */
   remote_ptr<void> vdso_start_addr;
   // The monkeypatcher that's handling this address space.
@@ -701,7 +973,8 @@ private:
   // The watchpoints set for tasks in this VM.  Watchpoints are
   // programmed per Task, but we track them per address space on
   // behalf of debuggers that assume that model.
-  WatchpointMap watchpoints;
+  std::map<MemoryRange, Watchpoint> watchpoints;
+  std::vector<std::map<MemoryRange, Watchpoint> > saved_watchpoints;
   // Tracee memory is read and written through this fd, which is
   // opened for the tracee's magic /proc/[tid]/mem device.  The
   // advantage of this over ptrace is that we can access it even
