@@ -2,6 +2,8 @@
 
 //#define DEBUGTAG "GdbConnection"
 
+#define REVERSE_EXECUTION
+
 /**
  * Much of this implementation is based on the documentation at
  *
@@ -56,8 +58,8 @@ static bool request_needs_immediate_response(const GdbRequest* req) {
   }
 }
 
-GdbConnection::GdbConnection(pid_t tgid)
-    : tgid(tgid), no_ack(false), inlen(0), outlen(0) {
+GdbConnection::GdbConnection(pid_t tgid, const Features& features)
+    : tgid(tgid), no_ack(false), inlen(0), outlen(0), features(features) {
   memset(&req, 0, sizeof(req));
 }
 
@@ -119,8 +121,9 @@ struct DebuggerParams {
 
 unique_ptr<GdbConnection> GdbConnection::await_client_connection(
     unsigned short desired_port, ProbePort probe, pid_t tgid,
-    const string& exe_image, ScopedFd* client_params_fd) {
-  auto dbg = unique_ptr<GdbConnection>(new GdbConnection(tgid));
+    const string& exe_image, const Features& features,
+    ScopedFd* client_params_fd) {
+  auto dbg = unique_ptr<GdbConnection>(new GdbConnection(tgid, features));
   unsigned short port = desired_port;
   ScopedFd listen_fd = open_socket(connection_addr, &port, probe);
   if (client_params_fd) {
@@ -132,9 +135,11 @@ unique_ptr<GdbConnection> GdbConnection::await_client_connection(
     ssize_t nwritten = write(*client_params_fd, &params, sizeof(params));
     assert(nwritten == sizeof(params));
   } else {
-    fprintf(stderr, "Attach to the rr debug server with this command:\n"
+    fprintf(stderr, "Launch gdb with\n"
+                    "  gdb %s\n"
+                    "and attach to the rr debug server with:\n"
                     "  target remote :%d\n",
-            port);
+            exe_image.c_str(), port);
   }
   LOG(debug) << "limiting debugger traffic to tgid " << tgid;
   dbg->await_debugger(listen_fd);
@@ -199,8 +204,6 @@ void GdbConnection::launch_gdb(ScopedFd& params_pipe_fd, const char* macros,
     args.push_back("-x");
     args.push_back(gdb_command_file_path);
   }
-  args.push_back("-l");
-  args.push_back("-1");
   if (macros) {
     string gdb_command_file = create_gdb_command_file(macros);
     args.push_back("-x");
@@ -639,11 +642,16 @@ bool GdbConnection::query(char* payload) {
     /* TODO process these */
     LOG(debug) << "gdb supports " << args;
 
+    const char* reverse_exec = "";
+#ifdef REVERSE_EXECUTION
+    if (features.reverse_execution) {
+      reverse_exec = ";ReverseContinue+;ReverseStep+";
+    }
+#endif
     snprintf(supported, sizeof(supported) - 1,
              "PacketSize=%zd;QStartNoAckMode+;qXfer:auxv:read+"
-             ";qXfer:siginfo:read+;qXfer:siginfo:write+"
-             ";multiprocess+",
-             sizeof(outbuf));
+             ";qXfer:siginfo:read+;qXfer:siginfo:write+;multiprocess+%s",
+             sizeof(outbuf), reverse_exec);
     write_packet(supported);
     return false;
   }
@@ -711,6 +719,23 @@ void GdbConnection::consume_request() {
   write_flush();
 }
 
+bool GdbConnection::process_bpacket(char* payload) {
+  if (strcmp(payload, "c") == 0) {
+    req.type = DREQ_CONTINUE;
+    req.run_direction = RUN_BACKWARD;
+    req.target = resume_thread;
+    return true;
+  } else if (strcmp(payload, "s") == 0) {
+    req.type = DREQ_STEP;
+    req.run_direction = RUN_BACKWARD;
+    req.target = resume_thread;
+    return true;
+  } else {
+    UNHANDLED_REQ() << "Unhandled gdb bpacket: b" << payload;
+    return false;
+  }
+}
+
 bool GdbConnection::process_vpacket(char* payload) {
   const char* name;
   char* args;
@@ -733,6 +758,7 @@ bool GdbConnection::process_vpacket(char* payload) {
       /* fall through */
       case 'c':
         req.type = DREQ_CONTINUE;
+        req.run_direction = RUN_FORWARD;
         req.target = resume_thread;
         return true;
       case 'S':
@@ -744,6 +770,7 @@ bool GdbConnection::process_vpacket(char* payload) {
       /* fall through */
       case 's':
         req.type = DREQ_STEP;
+        req.run_direction = RUN_FORWARD;
         if (args) {
           req.target = parse_threadid(args, &args);
           // If we get a step request for a
@@ -823,6 +850,11 @@ bool GdbConnection::process_vpacket(char* payload) {
     return true;
   }
 
+  if (name == strstr(name, "File:")) {
+    write_packet("");
+    return false;
+  }
+
   UNHANDLED_REQ() << "Unhandled gdb vpacket: v" << name;
   return false;
 }
@@ -852,6 +884,9 @@ bool GdbConnection::process_packet() {
       req.type = DREQ_INTERRUPT;
       ret = true;
       break;
+    case 'b':
+      ret = process_bpacket(payload);
+      break;
     case 'D':
       LOG(debug) << "gdb is detaching from us";
       req.type = DREQ_DETACH;
@@ -896,7 +931,7 @@ bool GdbConnection::process_packet() {
       req.mem.len = strtoul(payload, &payload, 16);
       assert('\0' == *payload);
 
-      LOG(debug) << "gdb requests memory (addr=" << req.mem.addr
+      LOG(debug) << "gdb requests memory (addr=" << HEX(req.mem.addr)
                  << ", len=" << req.mem.len << ")";
 
       ret = true;
@@ -984,7 +1019,7 @@ bool GdbConnection::process_packet() {
       assert('\0' == *payload);
 
       LOG(debug) << "gdb requests " << ('Z' == request ? "set" : "remove")
-                 << "breakpoint (addr=" << req.mem.addr
+                 << "breakpoint (addr=" << HEX(req.mem.addr)
                  << ", len=" << req.mem.len << ")";
 
       ret = true;
@@ -1070,7 +1105,7 @@ GdbRequest GdbConnection::get_request() {
     return req;
   }
 
-  while (1) {
+  while (true) {
     /* There's either new request data, or we have nothing
      * to do.  Either way, block until we read a complete
      * packet from gdb. */
@@ -1225,7 +1260,7 @@ void GdbConnection::notify_stop(GdbThreadId thread, int sig,
   // This isn't documented in the gdb remote protocol, but if we
   // don't do this, gdb will sometimes continue to send requests
   // for the previously-stopped thread when it obviously intends
-  // to making requests about the stopped thread.
+  // to be making requests about the stopped thread.
   LOG(debug) << "forcing query/resume thread to " << thread;
   query_thread = thread;
   resume_thread = thread;
@@ -1303,7 +1338,11 @@ void GdbConnection::reply_get_mem(const vector<uint8_t>& mem) {
   assert(DREQ_GET_MEM == req.type);
   assert(mem.size() <= req.mem.len);
 
-  write_hex_bytes_packet(mem.data(), mem.size());
+  if (req.mem.len > 0 && mem.size() == 0) {
+    write_packet("E01");
+  } else {
+    write_hex_bytes_packet(mem.data(), mem.size());
+  }
 
   consume_request();
 }
