@@ -11,7 +11,9 @@
 #include <vector>
 
 #include "GdbRegister.h"
+#include "Registers.h"
 #include "ReplaySession.h"
+#include "ReplayTimeline.h"
 
 /**
  * Descriptor for task within a task group.  Note: on linux, we can
@@ -41,7 +43,7 @@ inline std::ostream& operator<<(std::ostream& o, const GdbThreadId& t) {
  */
 struct GdbRegisterValue {
   enum {
-    MAX_SIZE = 16
+    MAX_SIZE = Registers::MAX_SIZE
   };
   GdbRegister name;
   uint8_t value[MAX_SIZE];
@@ -98,7 +100,7 @@ enum GdbRequestType {
   DREQ_GET_REG,
   DREQ_SET_REG,
 
-  /* No parameters. */
+  /* Use params.run_direction. */
   DREQ_CONTINUE,
   DREQ_INTERRUPT,
   DREQ_STEP,
@@ -153,6 +155,8 @@ struct GdbRequest {
       int param;
       GdbRestartType type;
     } restart;
+
+    RunDirection run_direction;
   };
 
   /**
@@ -203,9 +207,14 @@ public:
     DONT_PROBE = 0,
     PROBE_PORT
   };
+  struct Features {
+    Features() : reverse_execution(true) {}
+    bool reverse_execution;
+  };
   static std::unique_ptr<GdbConnection> await_client_connection(
       unsigned short desired_port, ProbePort probe, pid_t tgid,
-      const std::string& exe_image, ScopedFd* client_params_fd = nullptr);
+      const std::string& exe_image, const Features& features,
+      ScopedFd* client_params_fd = nullptr);
 
   /**
    * Exec gdb using the params that were written to
@@ -379,7 +388,7 @@ public:
   ReplaySession::shr_ptr get_checkpoint(int checkpoint_id);
 
 private:
-  GdbConnection(pid_t tgid);
+  GdbConnection(pid_t tgid, const Features& features);
 
   /**
    * Wait for a debugger client to connect to |dbg|'s socket.  Blocks
@@ -442,6 +451,11 @@ private:
    * false if we already handled the packet internally.
    */
   bool process_vpacket(char* payload);
+  /**
+   * Return true if we need to do something in a debugger request,
+   * false if we already handled the packet internally.
+   */
+  bool process_bpacket(char* payload);
   /**
    * Return true if we need to do something in a debugger request,
    * false if we already handled the packet internally.
@@ -471,6 +485,7 @@ private:
   ssize_t packetend;     /* index of '#' character */
   uint8_t outbuf[32768]; /* buffered output for gdb */
   ssize_t outlen;
+  Features features;
 };
 
 #endif /* RR_GDB_CONNECTION_H_ */
