@@ -17,6 +17,9 @@
 
 #include "kernel_metadata.h"
 #include "log.h"
+#include "ReplaySession.h"
+#include "ScopedFd.h"
+#include "task.h"
 #include "util.h"
 
 using namespace rr;
@@ -24,7 +27,7 @@ using namespace std;
 
 /**
  * 32-bit writes to DBG_COMMAND_MAGIC_ADDRESS by the debugger trigger
- * rr commands
+ * rr commands.
  */
 static const uintptr_t DBG_COMMAND_MAGIC_ADDRESS = 29298; // 'rr'
 
@@ -44,8 +47,15 @@ static const uintptr_t DBG_COMMAND_MSG_CREATE_CHECKPOINT = 0x01000000;
  * command parameter.
  */
 static const uintptr_t DBG_COMMAND_MSG_DELETE_CHECKPOINT = 0x02000000;
+
 static const uintptr_t DBG_COMMAND_PARAMETER_MASK = 0x00FFFFFF;
 
+/**
+ * 64-bit reads from DBG_WHEN_MAGIC_ADDRESS return the current trace frame's
+ * event number (the event we're working towards).
+ */
+static const uintptr_t DBG_WHEN_MAGIC_ADDRESS = DBG_COMMAND_MAGIC_ADDRESS + 4;
+
 // Special-sauce macros defined by rr when launching the gdb client,
 // which implement functionality outside of the gdb remote protocol.
 // (Don't stare at them too long or you'll go blind ;).)
@@ -66,16 +76,95 @@ static const char gdb_rr_macros[] =
     "define restart\n"
     "  run c$arg0\n"
     "end\n"
+    "define when\n"
+    "  p *(long long int*)(29298 + 4)\n"
+    "end\n"
+    // In gdb version "Fedora 7.8.1-30.fc21", a raw "run" command
+    // issued before any user-generated resume-execution command
+    // results in gdb hanging just after the inferior hits an internal
+    // gdb breakpoint.  This happens outside of rr, with gdb
+    // controlling gdbserver, as well.  We work around that by
+    // ensuring *some* resume-execution command has been issued before
+    // restarting the session.  But, only if the inferior hasn't
+    // already finished execution ($_thread != 0).  If it has and we
+    // issue the "stepi" command, then gdb refuses to restart
+    // execution.
+    "define hook-run\n"
+    "  if $_thread != 0 && !$suppress_run_hook\n"
+    "    stepi\n"
+    "  end\n"
+    "end\n"
+    "define hookpost-continue\n"
+    "  set $suppress_run_hook = 1\n"
+    "end\n"
+    "define hookpost-step\n"
+    "  set $suppress_run_hook = 1\n"
+    "end\n"
+    "define hookpost-stepi\n"
+    "  set $suppress_run_hook = 1\n"
+    "end\n"
+    "define hookpost-next\n"
+    "  set $suppress_run_hook = 1\n"
+    "end\n"
+    "define hookpost-nexti\n"
+    "  set $suppress_run_hook = 1\n"
+    "end\n"
+    "define hookpost-finish\n"
+    "  set $suppress_run_hook = 1\n"
+    "end\n"
+    "define hookpost-reverse-continue\n"
+    "  set $suppress_run_hook = 1\n"
+    "end\n"
+    "define hookpost-reverse-step\n"
+    "  set $suppress_run_hook = 1\n"
+    "end\n"
+    "define hookpost-reverse-stepi\n"
+    "  set $suppress_run_hook = 1\n"
+    "end\n"
+    "define hookpost-reverse-finish\n"
+    "  set $suppress_run_hook = 1\n"
+    "end\n"
+    "define hookpost-run\n"
+    "  set $suppress_run_hook = 0\n"
+    "end\n"
+    // Try both "set target-async" and "maint set target-async" since
+    // that changed recently.
+    "set target-async 0\n"
+    "maint set target-async 0\n"
     "handle SIGURG stop\n";
 
+/**
+ * Attempt to find the value of |regname| (a DebuggerRegister
+ * name), and if so (i) write it to |buf|; (ii)
+ * set |*defined = true|; (iii) return the size of written
+ * data.  If |*defined == false|, the value of |buf| is
+ * meaningless.
+ *
+ * This helper can fetch the values of both general-purpose
+ * and "extra" registers.
+ *
+ * NB: |buf| must be large enough to hold the largest register
+ * value that can be named by |regname|.
+ */
+static size_t get_reg(const Registers& regs, const ExtraRegisters& extra_regs,
+                      uint8_t* buf, GdbRegister regname, bool* defined) {
+  size_t num_bytes = regs.read_register(buf, regname, defined);
+  if (!*defined) {
+    num_bytes = extra_regs.read_register(buf, regname, defined);
+  }
+  return num_bytes;
+}
+
 /**
  * Return the register |which|, which may not have a defined value.
  */
-static GdbRegisterValue get_reg(Task* t, GdbRegister which) {
+static GdbRegisterValue get_reg(const Registers& regs,
+                                const ExtraRegisters& extra_regs,
+                                GdbRegister which) {
   GdbRegisterValue reg;
   memset(&reg, 0, sizeof(reg));
   reg.name = which;
-  reg.size = t->get_reg(&reg.value[0], which, &reg.defined);
+  reg.size = get_reg(regs, extra_regs, &reg.value[0], which, &reg.defined);
   return reg;
 }
 
@@ -86,6 +175,11 @@ static GdbThreadId get_threadid(Task* t) {
   return thread;
 }
 
+static bool matches_threadid(Task* t, const GdbThreadId& target) {
+  return (!target.pid || target.pid == t->tgid()) &&
+         (!target.tid || target.tid == t->rec_tid);
+}
+
 static WatchType watchpoint_type(GdbRequestType req) {
   switch (req) {
     case DREQ_SET_HW_BREAK:
@@ -110,8 +204,9 @@ static WatchType watchpoint_type(GdbRequestType req) {
   }
 }
 
-void GdbServer::maybe_singlestep_for_event(Task* t, GdbRequest* req) {
-  if (trace_instructions_up_to_event(session->current_trace_frame().time())) {
+static void maybe_singlestep_for_event(Task* t, GdbRequest* req) {
+  if (trace_instructions_up_to_event(
+          t->replay_session().current_trace_frame().time())) {
     fputs("Stepping: ", stderr);
     t->regs().print_register_file_compact(stderr);
     fprintf(stderr, " ticks:%" PRId64 "\n", t->tick_count());
@@ -121,24 +216,6 @@ void GdbServer::maybe_singlestep_for_event(Task* t, GdbRequest* req) {
   }
 }
 
-ReplaySession::shr_ptr GdbServer::get_checkpoint(int checkpoint_id) {
-  auto it = checkpoints.find(checkpoint_id);
-  if (it == checkpoints.end()) {
-    return nullptr;
-  }
-  return it->second;
-}
-
-void GdbServer::delete_checkpoint(int checkpoint_id) {
-  auto it = checkpoints.find(checkpoint_id);
-  if (it == checkpoints.end()) {
-    return;
-  }
-
-  it->second->kill_all_tasks();
-  checkpoints.erase(it);
-}
-
 bool GdbServer::maybe_process_magic_command(Task* t, const GdbRequest& req) {
   if (!(req.mem.addr == DBG_COMMAND_MAGIC_ADDRESS && req.mem.len == 4)) {
     return false;
@@ -148,14 +225,19 @@ bool GdbServer::maybe_process_magic_command(Task* t, const GdbRequest& req) {
   uintptr_t param = cmd & DBG_COMMAND_PARAMETER_MASK;
   switch (cmd & DBG_COMMAND_MSG_MASK) {
     case DBG_COMMAND_MSG_CREATE_CHECKPOINT: {
-      ReplaySession::shr_ptr checkpoint = session->clone();
-      delete_checkpoint(param);
-      checkpoints[param] = checkpoint;
+      if (timeline.can_add_checkpoint()) {
+        checkpoints[param] = timeline.add_explicit_checkpoint();
+      }
       break;
     }
-    case DBG_COMMAND_MSG_DELETE_CHECKPOINT:
-      delete_checkpoint(param);
+    case DBG_COMMAND_MSG_DELETE_CHECKPOINT: {
+      auto it = checkpoints.find(param);
+      if (it != checkpoints.end()) {
+        timeline.remove_explicit_checkpoint(it->second);
+        checkpoints.erase(it);
+      }
       break;
+    }
     default:
       return false;
   }
@@ -163,6 +245,30 @@ bool GdbServer::maybe_process_magic_command(Task* t, const GdbRequest& req) {
   return true;
 }
 
+bool GdbServer::maybe_process_magic_read(Task* t, const GdbRequest& req) {
+  if (req.mem.addr == DBG_WHEN_MAGIC_ADDRESS && req.mem.len == 8) {
+    vector<uint8_t> mem;
+    mem.resize(req.mem.len);
+    int64_t when = t->session().as_replay()
+                       ? int64_t(t->current_trace_frame().time())
+                       : int64_t(-1);
+    memcpy(mem.data(), &when, mem.size());
+    dbg->reply_get_mem(mem);
+    return true;
+  }
+  return false;
+}
+
+void GdbServer::dispatch_regs_request(const Registers& regs,
+                                      const ExtraRegisters& extra_regs) {
+  size_t n_regs = regs.total_registers();
+  GdbRegisterFile file(n_regs);
+  for (size_t i = 0; i < n_regs; ++i) {
+    file.regs[i] = get_reg(regs, extra_regs, GdbRegister(i));
+  }
+  dbg->reply_get_regs(file);
+}
+
 void GdbServer::dispatch_debugger_request(Session& session, Task* t,
                                           const GdbRequest& req) {
   assert(!req.is_resume_request());
@@ -180,11 +286,12 @@ void GdbServer::dispatch_debugger_request(Session& session, Task* t,
       dbg->reply_get_offsets();
       return;
     case DREQ_GET_THREAD_LIST: {
-      auto tasks = t->session().tasks();
       vector<GdbThreadId> tids;
-      for (auto& kv : tasks) {
-        Task* t = kv.second;
-        tids.push_back(get_threadid(t));
+      // When replay ends, there is still the last_task() around.
+      if (!session.as_replay() || !session.as_replay()->last_task()) {
+        for (auto& kv : session.tasks()) {
+          tids.push_back(get_threadid(kv.second));
+        }
       }
       dbg->reply_get_thread_list(tids);
       return;
@@ -263,11 +370,16 @@ void GdbServer::dispatch_debugger_request(Session& session, Task* t,
       return;
     }
     case DREQ_GET_MEM: {
+      if (maybe_process_magic_read(target, req)) {
+        return;
+      }
       vector<uint8_t> mem;
       mem.resize(req.mem.len);
       ssize_t nread =
           target->read_bytes_fallible(req.mem.addr, req.mem.len, mem.data());
       mem.resize(max(ssize_t(0), nread));
+      target->vm()->replace_breakpoints_with_original_values(
+          mem.data(), mem.size(), req.mem.addr);
       dbg->reply_get_mem(mem);
       return;
     }
@@ -298,17 +410,13 @@ void GdbServer::dispatch_debugger_request(Session& session, Task* t,
       return;
     }
     case DREQ_GET_REG: {
-      GdbRegisterValue reg = get_reg(target, req.reg.name);
+      GdbRegisterValue reg =
+          get_reg(target->regs(), target->extra_regs(), req.reg.name);
       dbg->reply_get_reg(reg);
       return;
     }
     case DREQ_GET_REGS: {
-      size_t n_regs = target->regs().total_registers();
-      GdbRegisterFile file(n_regs);
-      for (size_t i = 0; i < n_regs; ++i) {
-        file.regs[i] = get_reg(target, GdbRegister(i));
-      }
-      dbg->reply_get_regs(file);
+      dispatch_regs_request(target->regs(), target->extra_regs());
       return;
     }
     case DREQ_SET_REG: {
@@ -342,28 +450,43 @@ void GdbServer::dispatch_debugger_request(Session& session, Task* t,
     case DREQ_SET_SW_BREAK: {
       ASSERT(target, (req.mem.len == sizeof(AddressSpace::breakpoint_insn)))
           << "Debugger setting bad breakpoint insn";
-      bool ok = target->vm()->set_breakpoint(req.mem.addr, TRAP_BKPT_USER);
+      bool ok =
+          &session == &timeline.current_session()
+              ? timeline.add_breakpoint(target, req.mem.addr)
+              : target->vm()->add_breakpoint(req.mem.addr, TRAP_BKPT_USER);
       dbg->reply_watchpoint_request(ok);
       return;
     }
     case DREQ_REMOVE_SW_BREAK:
-      target->vm()->remove_breakpoint(req.mem.addr, TRAP_BKPT_USER);
+      if (&session == &timeline.current_session()) {
+        timeline.remove_breakpoint(target, req.mem.addr);
+      } else {
+        target->vm()->remove_breakpoint(req.mem.addr, TRAP_BKPT_USER);
+      }
       dbg->reply_watchpoint_request(true);
       return;
     case DREQ_REMOVE_HW_BREAK:
     case DREQ_REMOVE_RD_WATCH:
     case DREQ_REMOVE_WR_WATCH:
     case DREQ_REMOVE_RDWR_WATCH:
-      target->vm()->remove_watchpoint(req.mem.addr, req.mem.len,
-                                      watchpoint_type(req.type));
+      if (&session == &timeline.current_session()) {
+        timeline.remove_watchpoint(target, req.mem.addr, req.mem.len,
+                                   watchpoint_type(req.type));
+      } else {
+        target->vm()->remove_watchpoint(req.mem.addr, req.mem.len,
+                                        watchpoint_type(req.type));
+      }
       dbg->reply_watchpoint_request(true);
       return;
     case DREQ_SET_HW_BREAK:
     case DREQ_SET_RD_WATCH:
     case DREQ_SET_WR_WATCH:
     case DREQ_SET_RDWR_WATCH: {
-      bool ok = target->vm()->set_watchpoint(req.mem.addr, req.mem.len,
-                                             watchpoint_type(req.type));
+      bool ok = &session == &timeline.current_session()
+                    ? timeline.add_watchpoint(target, req.mem.addr, req.mem.len,
+                                              watchpoint_type(req.type))
+                    : target->vm()->add_watchpoint(req.mem.addr, req.mem.len,
+                                                   watchpoint_type(req.type));
       dbg->reply_watchpoint_request(ok);
       return;
     }
@@ -387,7 +510,9 @@ void GdbServer::dispatch_debugger_request(Session& session, Task* t,
  *
  * The received request is returned through |req|.
  */
-Task* GdbServer::diverter_process_debugger_requests(Task* t, GdbRequest* req) {
+Task* GdbServer::diverter_process_debugger_requests(
+    Task* t, DiversionSession& diversion_session, uint32_t& diversion_refcount,
+    GdbRequest* req) {
   while (true) {
     *req = dbg->get_request();
 
@@ -413,8 +538,12 @@ Task* GdbServer::diverter_process_debugger_requests(Task* t, GdbRequest* req) {
         continue;
       }
       case DREQ_SET_QUERY_THREAD: {
-        Task* next_task = t->session().find_task(req->target.tid);
-        t = next_task ? next_task : t;
+        if (req->target.tid) {
+          Task* next = t->session().find_task(req->target.tid);
+          if (next) {
+            t = next;
+          }
+        }
         break;
       }
       case DREQ_WRITE_SIGINFO:
@@ -450,7 +579,27 @@ Task* GdbServer::diverter_process_debugger_requests(Task* t, GdbRequest* req) {
         break;
     }
 
-    dispatch_debugger_request(*diversion_session, t, *req);
+    dispatch_debugger_request(diversion_session, t, *req);
+  }
+}
+
+void GdbServer::maybe_notify_stop(const BreakStatus& break_status) {
+  int sig = -1;
+  remote_ptr<void> watch_addr;
+  if (!break_status.watchpoints_hit.empty()) {
+    sig = SIGTRAP;
+    watch_addr = break_status.watchpoints_hit[0].addr;
+  }
+  if (break_status.breakpoint_hit || break_status.singlestep_complete) {
+    sig = SIGTRAP;
+  }
+  if (break_status.signal) {
+    sig = break_status.signal;
+  }
+  if (sig >= 0) {
+    /* Notify the debugger and process any new requests
+     * that might have triggered before resuming. */
+    dbg->notify_stop(get_threadid(break_status.task), sig, watch_addr.as_int());
   }
 }
 
@@ -468,56 +617,56 @@ Task* GdbServer::diverter_process_debugger_requests(Task* t, GdbRequest* req) {
 GdbRequest GdbServer::divert(ReplaySession& replay, pid_t task) {
   GdbRequest req;
   LOG(debug) << "Starting debugging diversion for " << &replay;
-  assert(!diversion_session && diversion_refcount == 0);
 
-  diversion_session = replay.clone_diversion();
-  diversion_refcount = 1;
+  DiversionSession::shr_ptr diversion_session = replay.clone_diversion();
+  uint32_t diversion_refcount = 1;
 
   Task* t = diversion_session->find_task(task);
   while (true) {
-    if (!(t = diverter_process_debugger_requests(t, &req))) {
+    if (!(t = diverter_process_debugger_requests(t, *diversion_session,
+                                                 diversion_refcount, &req))) {
       break;
     }
 
-    ReplaySession::RunCommand command =
-        (DREQ_STEP == req.type && get_threadid(t) == req.target)
-            ? Session::RUN_SINGLESTEP
-            : Session::RUN_CONTINUE;
+    if (req.run_direction == RUN_BACKWARD) {
+      // We don't support reverse execution in a diversion. Just issue
+      // an immediate stop.
+      dbg->notify_stop(get_threadid(t), SIGTRAP, 0);
+      continue;
+    }
+
+    RunCommand command =
+        (DREQ_STEP == req.type && matches_threadid(t, req.target))
+            ? RUN_SINGLESTEP
+            : RUN_CONTINUE;
     auto result = diversion_session->diversion_step(t, command);
 
     if (result.status == DiversionSession::DIVERSION_EXITED) {
       diversion_refcount = 0;
+      req.type = DREQ_NONE;
       dbg->notify_exit_code(0);
       break;
     }
 
     assert(result.status == DiversionSession::DIVERSION_CONTINUE);
-    if (result.break_status.reason == Session::BREAK_NONE) {
-      continue;
-    }
 
-    int sig = SIGTRAP;
-    remote_ptr<void> watch_addr = nullptr;
-    switch (result.break_status.reason) {
-      case Session::BREAK_SIGNAL:
-        sig = result.break_status.signal;
-        break;
-      case Session::BREAK_WATCHPOINT:
-        watch_addr = result.break_status.watch_address;
-        break;
-      default:
-        break;
-    }
-    /* Notify the debugger and process any new requests
-     * that might have triggered before resuming. */
-    dbg->notify_stop(get_threadid(result.break_status.task), sig,
-                     watch_addr.as_int());
+    maybe_notify_stop(result.break_status);
   }
 
   LOG(debug) << "... ending debugging diversion";
   assert(diversion_refcount == 0);
+
+  // Replace original task's breakpoints with the breakpoints set in the
+  // diversion. gdb thinks they're the same, so keeping breakpoints in sync
+  // is the right thing to do.
+  t = diversion_session->find_task(task);
+  Task* orig = replay.find_task(task);
+  if (t && orig) {
+    orig->vm()->copy_user_breakpoints_from(*t->vm());
+    orig->vm()->copy_watchpoints_from(*t->vm());
+  }
+
   diversion_session->kill_all_tasks();
-  diversion_session = nullptr;
   return req;
 }
 
@@ -526,9 +675,14 @@ GdbRequest GdbServer::divert(ReplaySession& replay, pid_t task) {
  * execution.
  */
 GdbRequest GdbServer::process_debugger_requests(Task* t) {
-  while (1) {
+  while (true) {
     GdbRequest req = dbg->get_request();
     req.suppress_debugger_stop = false;
+    if (timeline.is_running() && t) {
+      TaskUid tuid = t->tuid();
+      try_lazy_reverse_singlesteps(t, req);
+      t = timeline.current_session().find_task(tuid);
+    }
 
     if (req.type == DREQ_READ_SIGINFO) {
       // TODO: we send back a dummy siginfo_t to gdb
@@ -543,14 +697,16 @@ GdbRequest GdbServer::process_debugger_requests(Task* t) {
       memset(si_bytes.data(), 0, si_bytes.size());
       dbg->reply_read_siginfo(si_bytes);
 
-      req = divert(*session, t->rec_tid);
+      req = divert(t->replay_session(), t->rec_tid);
+      if (req.type == DREQ_NONE) {
+        continue;
+      }
       // Carry on to process the request that was rejected by
       // the diversion session
     }
 
     if (req.is_resume_request()) {
       maybe_singlestep_for_event(t, &req);
-      LOG(debug) << "  is resume request";
       return req;
     }
 
@@ -561,102 +717,85 @@ GdbRequest GdbServer::process_debugger_requests(Task* t) {
       return req;
     }
 
-    dispatch_debugger_request(*session, t, req);
+    dispatch_debugger_request(t ? t->session() : timeline.current_session(), t,
+                              req);
   }
 }
 
-GdbRequest GdbServer::replay_one_step() {
-  Task* t = session->current_task();
-  ReplaySession::RunCommand command;
-  GdbRequest req;
+void GdbServer::try_lazy_reverse_singlesteps(Task* t, GdbRequest& req) {
+  ReplayTimeline::Mark now;
+  bool need_seek = false;
 
-  if (debugger_active) {
-    req = process_debugger_requests(t);
+  while (req.type == DREQ_STEP && req.run_direction == RUN_BACKWARD &&
+         matches_threadid(t, req.target) && !req.suppress_debugger_stop &&
+         debugger_active) {
+    if (!now) {
+      now = timeline.mark();
+    }
+    ReplayTimeline::Mark previous = timeline.lazy_reverse_singlestep(now, t);
+    if (!previous) {
+      break;
+    }
+
+    now = previous;
+    need_seek = true;
+    BreakStatus break_status;
+    break_status.task = t;
+    break_status.singlestep_complete = true;
+    LOG(debug) << "  using lazy reverse-singlestep";
+    maybe_notify_stop(break_status);
+
+    while (true) {
+      req = dbg->get_request();
+      req.suppress_debugger_stop = false;
+      if (req.type != DREQ_GET_REGS) {
+        break;
+      }
+      LOG(debug) << "  using lazy reverse-singlestep registers";
+      dispatch_regs_request(now.regs(), now.extra_regs());
+    }
+  }
+
+  if (need_seek) {
+    timeline.seek_to_mark(now);
+  }
+}
+
+ReplayStatus GdbServer::replay_one_step() {
+  ReplayResult result;
+  bool suppress_debugger_stop = false;
+  Task* t = timeline.current_session().current_task();
+
+  if (debugger_active && t && t->task_group()->tguid() == debuggee_tguid) {
+    TaskUid tuid = t->tuid();
+    GdbRequest req = process_debugger_requests(t);
+    t = timeline.current_session().find_task(tuid);
     if (DREQ_RESTART == req.type) {
-      return req;
+      maybe_restart_session(req);
+      return REPLAY_CONTINUE;
     }
+    suppress_debugger_stop = req.suppress_debugger_stop;
     assert(req.is_resume_request());
-    command = (DREQ_STEP == req.type && get_threadid(t) == req.target)
-                  ? Session::RUN_SINGLESTEP
-                  : Session::RUN_CONTINUE;
+    RunCommand command =
+        (DREQ_STEP == req.type && matches_threadid(t, req.target))
+            ? RUN_SINGLESTEP
+            : RUN_CONTINUE;
+    result = timeline.replay_step(
+        command, req.run_direction,
+        req.run_direction == RUN_FORWARD ? target.event : 0);
   } else {
-    command = Session::RUN_CONTINUE;
+    result = timeline.replay_step(RUN_CONTINUE, RUN_FORWARD, target.event);
   }
 
-  auto result = session->replay_step(command, target.event);
+  if (result.status == REPLAY_EXITED) {
+    return result.status;
+  }
+  assert(result.status == REPLAY_CONTINUE);
 
-  GdbRequest no_restart;
-  no_restart.type = DREQ_NONE;
-  if (result.status == ReplaySession::REPLAY_EXITED) {
-    return no_restart;
-  }
-  assert(result.status == ReplaySession::REPLAY_CONTINUE);
-  if (result.break_status.reason == Session::BREAK_NONE) {
-    return no_restart;
-  }
-
-  if (debugger_active && !req.suppress_debugger_stop) {
-    int sig = SIGTRAP;
-    remote_ptr<void> watch_addr = nullptr;
-    switch (result.break_status.reason) {
-      case Session::BREAK_SIGNAL:
-        sig = result.break_status.signal;
-        break;
-      case Session::BREAK_WATCHPOINT:
-        watch_addr = result.break_status.watch_address;
-        break;
-      default:
-        break;
-    }
-    /* Notify the debugger and process any new requests
-     * that might have triggered before resuming. */
-    dbg->notify_stop(get_threadid(result.break_status.task), sig,
-                     watch_addr.as_int());
-  }
-  return no_restart;
-}
-
-/**
- * Return true if |t| appears to have entered but not exited an atomic
- * syscall (one that can't be interrupted).
- */
-static bool is_atomic_syscall(Task* t) {
-  return (t->is_probably_replaying_syscall() &&
-          (is_execve_syscall(t->regs().original_syscallno(), t->arch()) ||
-           (-ENOSYS == t->regs().syscall_result_signed() &&
-            !is_always_emulated_syscall(t->regs().original_syscallno(),
-                                        t->arch()))));
-}
-
-/**
- * Return true if it's possible/meaningful to make a checkpoint at the
- * |frame| that |t| will replay.
- */
-static bool can_checkpoint_at(Task* t, const TraceFrame& frame) {
-  Event ev(frame.event());
-  if (is_atomic_syscall(t)) {
-    return false;
-  }
-  if (ev.has_ticks_slop()) {
-    return false;
-  }
-  switch (ev.type()) {
-    case EV_EXIT:
-    case EV_UNSTABLE_EXIT:
-    // At exits, we can't clone the exiting tasks, so
-    // don't event bother trying to checkpoint.
-    case EV_SYSCALLBUF_RESET:
-    // RESETs are usually inserted in between syscall
-    // entry/exit.  Help the |is_atomic_syscall()|
-    // heuristics by not attempting to checkpoint at
-    // RESETs.  Users would never want to do that anyway.
-    case EV_TRACE_TERMINATION:
-      // There's nothing to checkpoint at the end of an
-      // early-terminated trace.
-      return false;
-    default:
-      return true;
+  if (debugger_active && !suppress_debugger_stop) {
+    maybe_notify_stop(result.break_status);
   }
+  return result.status;
 }
 
 /**
@@ -673,10 +812,10 @@ void GdbServer::maybe_connect_debugger(const ConnectionFlags& flags) {
   }
   // Don't launch the debugger for the initial rr fork child.
   // No one ever wants that to happen.
-  if (!session->can_validate()) {
+  if (!timeline.current_session().can_validate()) {
     return;
   }
-  Task* t = session->current_task();
+  Task* t = timeline.current_session().current_task();
   if (!t) {
     return;
   }
@@ -690,7 +829,7 @@ void GdbServer::maybe_connect_debugger(const ConnectionFlags& flags) {
   // So we make the decision to create the debugger based on the
   // frame we're *about to* replay, without modifying the
   // TraceIfstream.
-  TraceFrame next_frame = session->current_trace_frame();
+  TraceFrame next_frame = timeline.current_session().current_trace_frame();
   TraceFrame::Time event_now = next_frame.time();
   // NB: we'll happily attach to whichever task within the
   // group happens to be scheduled here.  We don't take
@@ -698,7 +837,7 @@ void GdbServer::maybe_connect_debugger(const ConnectionFlags& flags) {
   // leader".
   if (event_now < target.event || (target.pid && t->tgid() != target.pid) ||
       (target.pid && target.require_exec && !t->vm()->execed()) ||
-      !can_checkpoint_at(t, next_frame)) {
+      !timeline.can_add_checkpoint()) {
     return;
   }
 
@@ -710,18 +849,13 @@ void GdbServer::maybe_connect_debugger(const ConnectionFlags& flags) {
             target.pid, event_now);
   }
 
-  // Call set_debugged_tgid now so it will be cloned into the
-  // checkpoint below if one is created.
-  t->replay_session().set_debugged_tgid(t->tgid());
-
   // Have the "checkpoint" be the original replay
   // session, and then switch over to using the cloned
   // session.  The cloned tasks will look like children
   // of the clonees, so this scheme prevents |pstree|
   // output from getting /too/ far out of whack.
-  debugger_restart_checkpoint = session;
-  session = session->clone();
-  t = session->find_task(t->rec_tid);
+  debugger_restart_mark = timeline.add_explicit_checkpoint();
+  t = timeline.current_session().find_task(t->rec_tid);
 
   // Store the current tgid and event as the "execution target"
   // for the next replay session, if we end up restarting.  This
@@ -741,37 +875,43 @@ void GdbServer::maybe_connect_debugger(const ConnectionFlags& flags) {
     auto probe = flags.dbg_port > 0 ? GdbConnection::DONT_PROBE
                                     : GdbConnection::PROBE_PORT;
     dbg = GdbConnection::await_client_connection(
-        port, probe, t->tgid(), t->vm()->exe_image(),
+        port, probe, t->tgid(), t->vm()->exe_image(), GdbConnection::Features(),
         flags.debugger_params_write_pipe);
     if (flags.debugger_params_write_pipe) {
       flags.debugger_params_write_pipe->close();
     }
   }
 
+  debuggee_tguid = t->task_group()->tguid();
   debugger_active = true;
 }
 
 void GdbServer::maybe_restart_session(const GdbRequest& req) {
-  if (req.type != DREQ_RESTART) {
-    return;
-  }
+  assert(req.type == DREQ_RESTART);
   assert(dbg);
 
-  ReplaySession::shr_ptr checkpoint_to_restore;
+  ReplayTimeline::Mark mark_to_restore;
   if (req.restart.type == RESTART_FROM_CHECKPOINT) {
-    checkpoint_to_restore = get_checkpoint(req.restart.param);
-    if (!checkpoint_to_restore) {
+    auto it = checkpoints.find(req.restart.param);
+    if (it == checkpoints.end()) {
       LOG(info) << "Checkpoint " << req.restart.param << " not found.";
       dbg->notify_restart_failed();
       return;
     }
+    mark_to_restore = it->second;
   } else if (req.restart.type == RESTART_FROM_PREVIOUS) {
-    checkpoint_to_restore = debugger_restart_checkpoint;
+    mark_to_restore = debugger_restart_mark;
   }
-  if (checkpoint_to_restore) {
-    debugger_restart_checkpoint = checkpoint_to_restore;
-    session = checkpoint_to_restore->clone();
-    debugger_active = true;
+  timeline.remove_breakpoints_and_watchpoints();
+  if (mark_to_restore) {
+    timeline.seek_to_mark(mark_to_restore);
+    if (debugger_restart_mark) {
+      timeline.remove_explicit_checkpoint(debugger_restart_mark);
+    }
+    debugger_restart_mark = mark_to_restore;
+    if (timeline.can_add_checkpoint()) {
+      timeline.add_explicit_checkpoint();
+    }
     return;
   }
 
@@ -781,21 +921,23 @@ void GdbServer::maybe_restart_session(const GdbRequest& req) {
   // Note that we don't reset the target pid; we intentionally keep targeting
   // the same process no matter what is running when we hit the event.
   target.event = req.restart.param;
-  if (session->trace_reader().time() > target.event) {
-    // We weren't able to reuse a saved session, so
-    // just discard it and create a fresh one that's back
-    // at beginning-of-trace.
-    session = ReplaySession::create(session->trace_reader().dir());
-  }
+  timeline.seek_to_before_event(target.event);
 }
 
 void GdbServer::serve_replay(const ConnectionFlags& flags) {
   while (true) {
-    while (!session->last_task()) {
+    while (true) {
       maybe_connect_debugger(flags);
 
-      GdbRequest restart_request = replay_one_step();
-      maybe_restart_session(restart_request);
+      if (debugger_active &&
+          !timeline.current_session().find_task_group(debuggee_tguid)) {
+        // Our debugee must have exited.
+        break;
+      }
+
+      if (replay_one_step() == REPLAY_EXITED) {
+        break;
+      }
     }
     LOG(info) << ("Replayer successfully finished.");
 
@@ -806,7 +948,8 @@ void GdbServer::serve_replay(const ConnectionFlags& flags) {
 
     // TODO return real exit code, if it's useful.
     dbg->notify_exit_code(0);
-    GdbRequest req = process_debugger_requests(session->last_task());
+    GdbRequest req =
+        process_debugger_requests(timeline.current_session().last_task());
     if (DREQ_RESTART == req.type) {
       maybe_restart_session(req);
       continue;
@@ -828,14 +971,22 @@ void GdbServer::emergency_debug(Task* t) {
   // this.  Unlike in that context though, we don't know if |t|
   // overshot an internal breakpoint.  If it did, cover that
   // breakpoint up.
-  t->vm()->destroy_all_breakpoints();
+  if (t->vm()) {
+    t->vm()->remove_all_breakpoints();
+  }
 
   // Don't launch a debugger on fatal errors; the user is most
   // likely already in a debugger, and wouldn't be able to
   // control another session. Instead, launch a new GdbServer and wait for
   // the user to connect from another window.
+  GdbConnection::Features features;
+  // Don't advertise reverse_execution to gdb becase a) it won't work and
+  // b) some gdb versions will fail if the user doesn't turn off async
+  // mode (and we don't want to require users to do that)
+  features.reverse_execution = false;
   unique_ptr<GdbConnection> dbg = GdbConnection::await_client_connection(
-      t->tid, GdbConnection::PROBE_PORT, t->tgid(), string());
+      t->tid, GdbConnection::PROBE_PORT, t->tgid(), t->vm()->exe_image(),
+      features);
 
   GdbServer(dbg).process_debugger_requests(t);
 }
