@@ -11,13 +11,12 @@
 #include "log.h"
 #include "record_signal.h"
 #include "record_syscall.h"
+#include "seccomp-bpf.h"
 #include "task.h"
 
 using namespace rr;
 using namespace std;
 
-extern char** environ;
-
 /**
  * Create a pulseaudio client config file with shm disabled.  That may
  * be the cause of a mysterious divergence.  Return an envpair to set
@@ -111,61 +110,106 @@ static bool handle_ptrace_exit_event(Task* t) {
   return true;
 }
 
-static void handle_seccomp_event(Task* t) {
-  if (t->regs().original_syscallno() < 0) {
-    // negative syscall numbers after a SECCOMP event
-    // are treated as "skip this syscall". There will be one syscall event
-    // recorded instead of two. So, record an enter-syscall event now
-    // and treat the other event as the exit.
-    t->push_event(SyscallEvent(t->regs().original_syscallno(), t->arch()));
-    ASSERT(t, EV_SYSCALL == t->ev().type());
-    t->ev().Syscall().state = ENTERING_SYSCALL;
-    t->record_current_event();
-  }
-}
-
-void RecordSession::handle_ptrace_event(Task* t, ForceSyscall* force_cont) {
-  /* handle events */
+bool RecordSession::handle_ptrace_event(Task* t, StepState* step_state) {
   int event = t->ptrace_event();
-  if (event != PTRACE_EVENT_NONE) {
-    LOG(debug) << "  " << t->tid << ": handle_ptrace_event " << event
-               << ": event " << t->ev();
+  if (event == PTRACE_EVENT_NONE) {
+    return false;
   }
-  switch (event) {
-    case PTRACE_EVENT_NONE:
-    case PTRACE_EVENT_STOP:
-      break;
 
+  LOG(debug) << "  " << t->tid << ": handle_ptrace_event " << event
+             << ": event " << t->ev();
+
+  switch (event) {
     case PTRACE_EVENT_SECCOMP_OBSOLETE:
     case PTRACE_EVENT_SECCOMP:
-      handle_seccomp_event(t);
-      // The next resume_execution needs to be a PTRACE_SYSCALL to observe
-      // the enter-syscall event (or, for negative syscall numbers, exit).
-      *force_cont = FORCE_SYSCALL;
+      t->seccomp_bpf_enabled = true;
+      if (t->get_ptrace_eventmsg_seccomp_data() == SECCOMP_RET_DATA) {
+        if (t->regs().original_syscallno() < 0) {
+          // negative syscall numbers after a SECCOMP event
+          // are treated as "skip this syscall". There will be one syscall event
+          // reported instead of two. So, record an enter-syscall event now
+          // and treat the other event as the exit.
+          t->fixup_syscall_regs();
+          t->push_event(
+              SyscallEvent(t->regs().original_syscallno(), t->arch()));
+          ASSERT(t, EV_SYSCALL == t->ev().type());
+          t->ev().Syscall().state = ENTERING_SYSCALL;
+          t->record_current_event();
+          // Don't continue yet. At the next iteration of record_step, we'll
+          // enter syscall_state_changed and that will trigger a continue to
+          // the syscall exit.
+          step_state->continue_type = DONT_CONTINUE;
+        } else {
+          // The next continue needs to be a PTRACE_SYSCALL to observe
+          // the enter-syscall event.
+          step_state->continue_type = CONTINUE_SYSCALL;
+        }
+      } else {
+        // user seccomp filter that would have returned SECOMP_RET_ERRNO.
+        // Emulate that.
+        t->fixup_syscall_regs();
+
+        if (!t->is_in_untraced_syscall()) {
+          t->push_event(
+              SyscallEvent(t->regs().original_syscallno(), t->arch()));
+          ASSERT(t, EV_SYSCALL == t->ev().type());
+          t->ev().Syscall().state = ENTERING_SYSCALL;
+          t->record_current_event();
+        }
+
+        Registers r = t->regs();
+        // Cause kernel processing to skip the syscall
+        r.set_original_syscallno(-1);
+        r.set_syscall_result(-t->get_ptrace_eventmsg_seccomp_data());
+        t->set_regs(r);
+        // Don't continue yet. At the next iteration of record_step, if we
+        // recorded the syscall-entry we'll enter syscall_state_changed and
+        // that will trigger a continue to the syscall exit.
+        step_state->continue_type = DONT_CONTINUE;
+      }
       break;
 
-    case PTRACE_EVENT_CLONE:
-    case PTRACE_EVENT_FORK: {
-      pid_t new_tid = t->get_ptrace_eventmsg_pid();
+    case PTRACE_EVENT_CLONE: {
       remote_ptr<void> stack;
       remote_ptr<int>* ptid_not_needed = nullptr;
       remote_ptr<void> tls;
       remote_ptr<int> ctid;
       extract_clone_parameters(t, &stack, ptid_not_needed, &tls, &ctid);
-      // fork and can never share these resources, only
+      // fork can never share these resources, only
       // copy, so the flags here aren't meaningful for it.
       unsigned long flags_arg =
           is_clone_syscall(t->regs().original_syscallno(), t->arch())
               ? t->regs().arg1()
               : 0;
-      clone(t, clone_flags_to_task_flags(flags_arg), stack, tls, ctid, new_tid);
+
+      // Ideally we'd just use t->get_ptrace_eventmsg_pid() here, but
+      // kernels failed to translate that value from other pid namespaces to
+      // our pid namespace until June 2014:
+      // https://github.com/torvalds/linux/commit/4e52365f279564cef0ddd41db5237f0471381093
+      pid_t new_tid;
+      if (flags_arg & CLONE_THREAD) {
+        new_tid = t->find_newborn_thread();
+      } else {
+        new_tid = t->find_newborn_child_process();
+      }
+      Task* new_task = clone(t, clone_flags_to_task_flags(flags_arg), stack,
+                             tls, ctid, new_tid);
+      rec_set_syscall_new_task(t, new_task);
       // Skip past the ptrace event.
-      t->cont_syscall();
-      assert(t->pending_sig() == 0);
+      step_state->continue_type = CONTINUE_SYSCALL;
       break;
     }
 
-    case PTRACE_EVENT_EXEC: {
+    case PTRACE_EVENT_FORK: {
+      pid_t new_tid = t->find_newborn_child_process();
+      Task* new_task = clone(t, 0, nullptr, nullptr, nullptr, new_tid);
+      rec_set_syscall_new_task(t, new_task);
+      // Skip past the ptrace event.
+      step_state->continue_type = CONTINUE_SYSCALL;
+      break;
+    }
+
+    case PTRACE_EVENT_EXEC:
       /* The initial tracee, if it's still around, is now
        * for sure not running in the initial rr address
        * space, so we can unblock signals. */
@@ -180,23 +224,26 @@ void RecordSession::handle_ptrace_event(Task* t, ForceSyscall* force_cont) {
       t->pop_syscall();
 
       // Skip past the ptrace event.
-      t->cont_syscall();
-      assert(t->pending_sig() == 0);
+      step_state->continue_type = CONTINUE_SYSCALL;
       break;
-    }
 
-    case PTRACE_EVENT_EXIT: {
-      FATAL() << "PTRACE_EVENT_EXIT should already have been handled";
+    case PTRACE_EVENT_STOP:
+      last_task_switchable = ALLOW_SWITCH;
+      step_state->continue_type = DONT_CONTINUE;
       break;
-    }
 
+    // We map vfork() to fork() so we don't expect to see these:
     case PTRACE_EVENT_VFORK:
     case PTRACE_EVENT_VFORK_DONE:
+    // This is handled separately:
+    case PTRACE_EVENT_EXIT:
     default:
-      FATAL() << "Unhandled ptrace event " << ptrace_event_name(event) << "("
-              << event << ")";
+      ASSERT(t, false) << "Unhandled ptrace event " << ptrace_event_name(event)
+                       << "(" << event << ")";
       break;
   }
+
+  return true;
 }
 
 static void debug_exec_state(const char* msg, Task* t) {
@@ -204,23 +251,27 @@ static void debug_exec_state(const char* msg, Task* t) {
              << " pevent=" << t->ptrace_event();
 }
 
-static void task_continue(Task* t, ForceSyscall force_cont, int sig) {
+void RecordSession::task_continue(Task* t, const StepState& step_state) {
+  ASSERT(t, step_state.continue_type != DONT_CONTINUE);
+
   bool may_restart = t->at_may_restart_syscall();
 
-  if (sig) {
-    LOG(debug) << "  delivering " << signal_name(sig) << " to " << t->tid;
+  if (step_state.continue_sig) {
+    LOG(debug) << "  delivering " << signal_name(step_state.continue_sig)
+               << " to " << t->tid;
   }
   if (may_restart && t->seccomp_bpf_enabled) {
     LOG(debug) << "  PTRACE_SYSCALL to possibly-restarted " << t->ev();
   }
 
-  if (!t->seccomp_bpf_enabled || FORCE_SYSCALL == force_cont || may_restart) {
+  if (!t->seccomp_bpf_enabled || CONTINUE_SYSCALL == step_state.continue_type ||
+      may_restart) {
     /* We won't receive PTRACE_EVENT_SECCOMP events until
      * the seccomp filter is installed by the
      * syscall_buffer lib in the child, therefore we must
      * record in the traditional way (with PTRACE_SYSCALL)
      * until it is installed. */
-    t->cont_syscall_nonblocking(sig,
+    t->cont_syscall_nonblocking(step_state.continue_sig,
                                 t->record_session().scheduler().max_ticks());
   } else {
     /* When the seccomp filter is on, instead of capturing
@@ -233,41 +284,19 @@ static void task_continue(Task* t, ForceSyscall force_cont, int sig) {
      * process to continue to the actual entry point of
      * the syscall (using cont_syscall_block()) and then
      * using the same logic as before. */
-    t->cont_nonblocking(sig, t->record_session().scheduler().max_ticks());
+    t->cont_nonblocking(step_state.continue_sig,
+                        t->record_session().scheduler().max_ticks());
   }
 }
 
 /**
- * Resume execution of |t| to the next notable event, such as a
- * syscall.
+ * Step |t| forward until the tracee syscall that disarms the desched
+ * event. If a signal becomes pending in the interim, we stash it.
+ * This allows the caller to deliver the signal after this returns.
+ * (In reality the desched event will already have been disarmed before we
+ * enter this function.)
  */
-enum NeedTaskContinue {
-  DONT_NEED_TASK_CONTINUE = 0,
-  NEED_TASK_CONTINUE
-};
-static void resume_execution(Task* t, NeedTaskContinue need_task_continue,
-                             ForceSyscall force_cont = DEFAULT_CONT) {
-  assert(!t->may_be_blocked());
-
-  debug_exec_state("EXEC_START", t);
-
-  if (need_task_continue) {
-    task_continue(t, force_cont, /*no sig*/ 0);
-    t->wait();
-  }
-
-  if (t->is_ptrace_seccomp_event()) {
-    t->seccomp_bpf_enabled = true;
-  }
-}
-
-/**
- * Step |t| forward until the desched event is disarmed.  If a signal
- * becomes pending in the interim, the |waitpid()| status is returned,
- * and |si| is filled in.  This allows the caller to deliver the
- * signal after this returns and the desched event is disabled.
- */
-static void disarm_desched(Task* t) {
+static void advance_to_disarm_desched_syscall(Task* t) {
   int old_sig = 0;
 
   LOG(debug) << "desched: DISARMING_DESCHED_EVENT";
@@ -284,14 +313,14 @@ static void disarm_desched(Task* t) {
      * copy-out the buffered out data to the user's
      * buffer.  This happens in the interval where we're
      * reaching the disarm-desched ioctl, so that code is
-     * susceptible to receiving SIG_TIMESLICE.  If it
-     * does, we'll try to stepi the tracee to a safe point
-     * ... through a practically unbounded memcpy(), which
-     * can be very expensive. */
+     * susceptible to receiving SIG_TIMESLICE. */
     int sig = t->pending_sig();
     if (PerfCounters::TIME_SLICE_SIGNAL == sig) {
       continue;
     }
+    // We should not receive SYSCALLBUF_DESCHED_SIGNAL since it should already
+    // have been disarmed.
+    ASSERT(t, SYSCALLBUF_DESCHED_SIGNAL != sig);
     if (sig && sig == old_sig) {
       LOG(debug) << "  coalescing pending " << signal_name(sig);
       continue;
@@ -301,6 +330,9 @@ static void disarm_desched(Task* t) {
       t->stash_sig();
     }
   } while (!t->is_disarm_desched_event_syscall());
+
+  // Exit the syscall.
+  t->cont_syscall();
 }
 
 /**
@@ -308,7 +340,7 @@ static void disarm_desched(Task* t) {
  * changed.  (For now, changes except the original desched'd syscall
  * being restarted.)
  */
-static void desched_state_changed(Task* t) {
+void RecordSession::desched_state_changed(Task* t) {
   switch (t->ev().Desched().state) {
     case IN_SYSCALL:
       LOG(debug) << "desched: IN_SYSCALL";
@@ -323,7 +355,7 @@ static void desched_state_changed(Task* t) {
       t->ev().Desched().state = DISARMING_DESCHED_EVENT;
     /* fall through */
     case DISARMING_DESCHED_EVENT: {
-      disarm_desched(t);
+      advance_to_disarm_desched_syscall(t);
 
       t->ev().Desched().state = DISARMED_DESCHED_EVENT;
       t->record_current_event();
@@ -340,7 +372,7 @@ static void desched_state_changed(Task* t) {
       // We were just descheduled for potentially a long
       // time, and may have just had a signal become
       // pending.  Ensure we get another chance to run.
-      t->switchable = PREVENT_SWITCH;
+      last_task_switchable = PREVENT_SWITCH;
       return;
     }
     default:
@@ -404,7 +436,7 @@ static void maybe_discard_syscall_interruption(Task* t, int ret) {
   syscallno = t->ev().Syscall().number;
   if (0 > ret) {
     syscall_not_restarted(t);
-  } else if (0 < ret) {
+  } else {
     ASSERT(t, syscallno == ret) << "Interrupted call was "
                                 << t->syscall_name(syscallno)
                                 << " and sigreturn claims to be restarting "
@@ -425,7 +457,7 @@ static void copy_syscall_arg_regs(Registers* to, const Registers& from) {
   to->set_arg6(from.arg6());
 }
 
-static void syscall_state_changed(Task* t, bool by_waitpid) {
+void RecordSession::syscall_state_changed(Task* t, StepState* step_state) {
   switch (t->ev().Syscall().state) {
     case ENTERING_SYSCALL: {
       debug_exec_state("EXEC_SYSCALL_ENTRY", t);
@@ -442,38 +474,53 @@ static void syscall_state_changed(Task* t, bool by_waitpid) {
         t->ev().Syscall().regs = t->regs();
       }
 
-      t->switchable = rec_prepare_syscall(t);
+      last_task_switchable = rec_prepare_syscall(t);
+
+      debug_exec_state("after cont", t);
+      t->ev().Syscall().state = PROCESSING_SYSCALL;
 
       // Resume the syscall execution in the kernel context.
-      t->cont_syscall_nonblocking();
-      debug_exec_state("after cont", t);
+      step_state->continue_type = CONTINUE_SYSCALL;
+
+      if (t->session().can_validate() && Flags::get().check_cached_mmaps) {
+        t->vm()->verify(t);
+      }
+
+      if (t->desched_rec() && t->is_in_untraced_syscall() &&
+          t->ev().Syscall().is_restart && t->has_stashed_sig()) {
+        // We have a signal to deliver but we're about to restart an untraced
+        // syscall that may block and the desched event has been disarmed.
+        // Rearm the desched event so if the syscall blocks, it will be
+        // interrupted and we'll have a chance to deliver our signal.
+        arm_desched_event(t);
+      }
 
-      t->ev().Syscall().state = PROCESSING_SYSCALL;
       return;
     }
     case PROCESSING_SYSCALL:
       debug_exec_state("EXEC_IN_SYSCALL", t);
 
-      assert(by_waitpid);
       // Linux kicks tasks out of syscalls before delivering
       // signals.
       ASSERT(t, !t->pending_sig()) << "Signal " << signal_name(t->pending_sig())
                                    << " pending while in syscall???";
 
       t->ev().Syscall().state = EXITING_SYSCALL;
-      t->switchable = PREVENT_SWITCH;
+      step_state->continue_type = DONT_CONTINUE;
       return;
 
     case EXITING_SYSCALL: {
-      int syscallno = t->ev().Syscall().number;
-      int may_restart;
-      int retval;
-
       debug_exec_state("EXEC_SYSCALL_DONE", t);
 
       assert(t->pending_sig() == 0);
 
-      retval = t->regs().syscall_result();
+      if (t->desched_rec()) {
+        // If we enabled the desched event above, disable it.
+        disarm_desched_event(t);
+      }
+
+      int syscallno = t->ev().Syscall().number;
+      int retval = t->regs().syscall_result();
 
       // sigreturn is a special snowflake, because it
       // doesn't actually return.  Instead, it undoes the
@@ -481,8 +528,7 @@ static void syscall_state_changed(Task* t, bool by_waitpid) {
       // preparing the tracee for a restart-syscall.  So we
       // take this opportunity to possibly pop an
       // interrupted-syscall event.
-      if (is_sigreturn_syscall(syscallno, t->arch()) ||
-          is_rt_sigreturn_syscall(syscallno, t->arch())) {
+      if (is_sigreturn(syscallno, t->arch())) {
         ASSERT(t, t->regs().original_syscallno() == -1);
         t->record_current_event();
         t->pop_syscall();
@@ -492,9 +538,13 @@ static void syscall_state_changed(Task* t, bool by_waitpid) {
         t->record_event(Event(EV_EXIT_SIGHANDLER, NO_EXEC_INFO, t->arch()));
 
         maybe_discard_syscall_interruption(t, retval);
-        // XXX probably not necessary to make the
-        // tracee unswitchable
-        t->switchable = PREVENT_SWITCH;
+
+        if (EV_DESCHED == t->ev().type()) {
+          LOG(debug) << "  exiting desched critical section";
+          desched_state_changed(t);
+        }
+
+        // XXX probably not necessary to leave the tracee unswitchable
         return;
       }
 
@@ -510,12 +560,12 @@ static void syscall_state_changed(Task* t, bool by_waitpid) {
 
       /* TODO: is there any reason a restart_syscall can't
        * be interrupted by a signal and itself restarted? */
-      may_restart = !is_restart_syscall_syscall(syscallno, t->arch())
-                    // SYS_pause is either interrupted or
-                    // never returns.  It doesn't restart.
-                    &&
-                    !is_pause_syscall(syscallno, t->arch()) &&
-                    t->regs().syscall_may_restart();
+      bool may_restart = !is_restart_syscall_syscall(syscallno, t->arch())
+                         // SYS_pause is either interrupted or
+                         // never returns.  It doesn't restart.
+                         &&
+                         !is_pause_syscall(syscallno, t->arch()) &&
+                         t->regs().syscall_may_restart();
       /* no need to process the syscall in case its
        * restarted this will be done in the exit from the
        * restart_syscall */
@@ -553,10 +603,8 @@ static void syscall_state_changed(Task* t, bool by_waitpid) {
         }
       } else {
         t->ev().transform(EV_SYSCALL_INTERRUPTION);
-        t->ev().Syscall().is_restart = 1;
+        t->ev().Syscall().is_restart = true;
       }
-
-      t->switchable = ALLOW_SWITCH;
       return;
     }
 
@@ -605,21 +653,89 @@ void RecordSession::check_perf_counters_working(Task* t,
   }
 }
 
+static void inject_signal(Task* t) {
+  int sig = t->ev().Signal().siginfo.si_signo;
+
+  /* Signal injection is tricky. Per the ptrace(2) man page, injecting
+   * a signal while the task is not in a signal-stop is not guaranteed to work
+   * (and indeed, we see that the kernel sometimes ignores such signals).
+   * But some signals must be delayed until after the signal-stop that notified
+   * us of them.
+   * So, first we check if we're in a signal-stop that we can use to inject
+   * a signal. Some (all?) SIGTRAP stops are *not* usable for signal injection.
+   */
+  if (t->pending_sig() && t->pending_sig() != SIGTRAP) {
+    LOG(debug) << "    in signal-stop for " << signal_name(t->pending_sig());
+  } else {
+    /* We're not in a usable signal-stop. Force a signal-stop by sending
+     * a new signal with tgkill (as the ptrace(2) man page recommends).
+     */
+    LOG(debug) << "    maybe not in signal-stop; tgkill(" << signal_name(sig)
+               << ")";
+    t->tgkill(sig);
+
+    /* Now singlestep the task until we're in a signal-stop for the signal
+     * we've just sent. We must absorb and forget that signal here since we
+     * don't want it delivered to the task for real.
+     */
+    while (true) {
+      auto old_ip = t->ip();
+      t->cont_singlestep();
+      ASSERT(t, old_ip == t->ip());
+      ASSERT(t, t->pending_sig());
+      if (t->pending_sig() == sig) {
+        LOG(debug) << "    stopped with signal " << signal_name(sig);
+        break;
+      }
+      /* It's possible for other signals to arrive while we're trying to
+       * get to the signal-stop for the signal we just sent. Stash them for
+       * later delivery.
+       */
+      if (t->pending_sig() == SYSCALLBUF_DESCHED_SIGNAL) {
+        LOG(debug) << "    stopped with signal " << signal_name(sig)
+                   << "; ignoring it and carrying on";
+      } else {
+        LOG(debug) << "    stopped with signal " << signal_name(sig)
+                   << "; stashing it and carrying on";
+        t->stash_sig();
+      }
+    }
+    /* We're now in a signal-stop (and for the right signal too, though that
+     * doesn't really matter).
+     */
+  }
+
+  /* Now that we're in a signal-stop, we can inject our signal and advance
+   * to the signal handler with one single-step.
+   */
+  LOG(debug) << "    injecting signal number " << t->ev().Signal().siginfo;
+  t->set_siginfo(t->ev().Signal().siginfo);
+  t->cont_singlestep(sig);
+  ASSERT(t, t->pending_sig() == SIGTRAP);
+  ASSERT(t, t->get_signal_user_handler(sig) == t->ip());
+
+  // It's been observed that when tasks enter
+  // sighandlers, the singlestep operation above
+  // doesn't retire any instructions; and
+  // indeed, if an instruction could be retired,
+  // this code wouldn't work.  This also
+  // cross-checks the sighandler information we
+  // maintain in |t->sighandlers|.
+  assert(!PerfCounters::extra_perf_counters_enabled() ||
+         0 == t->hpc.read_extra().instructions_retired);
+}
+
 /**
  * |t| is being delivered a signal, and its state changed.
- * |by_waitpid| is true if the status change was observed by a
- * waitpid() call.
  *
  * Return true if execution was incidentally resumed to a new event,
  * false otherwise.
  */
-static bool signal_state_changed(Task* t, bool by_waitpid) {
-  int sig = t->ev().Signal().number;
+void RecordSession::signal_state_changed(Task* t, StepState* step_state) {
+  int sig = t->ev().Signal().siginfo.si_signo;
 
   switch (t->ev().type()) {
     case EV_SIGNAL: {
-      assert(!by_waitpid);
-
       // This event is used by the replayer to advance to
       // the point of signal delivery.
       t->record_current_event();
@@ -631,20 +747,7 @@ static bool signal_state_changed(Task* t, bool by_waitpid) {
         LOG(debug) << "  " << t->tid << ": " << signal_name(sig)
                    << " has user handler";
 
-        t->cont_singlestep(sig);
-        ASSERT(t, t->stop_sig() == SIGTRAP) << "Expected single-step trap,"
-                                               " got siginfo "
-                                            << t->get_siginfo();
-
-        // It's been observed that when tasks enter
-        // sighandlers, the singlestep operation above
-        // doesn't retire any instructions; and
-        // indeed, if an instruction could be retired,
-        // this code wouldn't work.  This also
-        // cross-checks the sighandler information we
-        // maintain in |t->sighandlers|.
-        assert(!PerfCounters::extra_perf_counters_enabled() ||
-               0 == t->hpc.read_extra().instructions_retired);
+        inject_signal(t);
 
         // It's somewhat difficult engineering-wise to
         // compute the sigframe size at compile time,
@@ -668,10 +771,24 @@ static bool signal_state_changed(Task* t, bool by_waitpid) {
 
         t->ev().transform(EV_SIGNAL_HANDLER);
         t->signal_delivered(sig);
-        t->ev().Signal().delivered = 1;
+        // We already continued! Don't continue now, and allow switching.
+        step_state->continue_type = DONT_CONTINUE;
+        last_task_switchable = ALLOW_SWITCH;
       } else {
         LOG(debug) << "  " << t->tid << ": no user handler for "
                    << signal_name(sig);
+        // Don't do another task continue. We want to deliver the signal
+        // as the next thing that the task does.
+        step_state->continue_type = DONT_CONTINUE;
+        // If we didn't set up the sighandler frame, we need
+        // to ensure that this tracee is scheduled next so
+        // that we can deliver the signal normally.  We have
+        // to do that because setting up the sighandler frame
+        // is synchronous, but delivery otherwise is async.
+        // But right after this, we may have to process some
+        // syscallbuf state, so we can't let the tracee race
+        // with us.
+        last_task_switchable = PREVENT_SWITCH;
       }
 
       // We record this data regardless to simplify replay.
@@ -683,76 +800,40 @@ static bool signal_state_changed(Task* t, bool by_waitpid) {
       // handler.
       t->record_current_event();
 
-      // If we didn't set up the sighandler frame, we need
-      // to ensure that this tracee is scheduled next so
-      // that we can deliver the signal normally.  We have
-      // to do that because setting up the sighandler frame
-      // is synchronous, but delivery otherwise is async.
-      // But right after this, we may have to process some
-      // syscallbuf state, so we can't let the tracee race
-      // with us.
-      t->switchable =
-          t->ev().Signal().delivered ? ALLOW_SWITCH : PREVENT_SWITCH;
-      return false;
+      // If this is the signal delivered by a sigsuspend, then clear
+      // sigsuspend_blocked_sigs to indicate that future signals are not
+      // being delivered by sigsuspend.
+      t->sigsuspend_blocked_sigs = nullptr;
+      break;
     }
-    case EV_SIGNAL_DELIVERY:
-      if (!t->ev().Signal().delivered) {
-        task_continue(t, DEFAULT_CONT, sig);
-        if (possibly_destabilizing_signal(t, sig,
-                                          t->ev().Signal().deterministic)) {
-          LOG(warn) << "Delivered core-dumping signal; may misrecord "
-                       "CLONE_CHILD_CLEARTID memory race";
-          t->destabilize_task_group();
-          t->switchable = ALLOW_SWITCH;
-        }
-        t->signal_delivered(sig);
-        t->ev().Signal().delivered = 1;
-        return false;
-      }
 
-      // The tracee's waitpid status has changed, so we're finished
-      // delivering the signal.
-      assert(by_waitpid);
+    case EV_SIGNAL_DELIVERY:
+      step_state->continue_sig = sig;
+      t->signal_delivered(sig);
+      if (possibly_destabilizing_signal(t, sig,
+                                        t->ev().Signal().deterministic)) {
+        LOG(warn) << "Delivered core-dumping signal; may misrecord "
+                     "CLONE_CHILD_CLEARTID memory race";
+        t->destabilize_task_group();
+        t->pop_signal_delivery();
+        last_task_switchable = ALLOW_SWITCH;
+        break;
+      }
       t->pop_signal_delivery();
-      // The event we just |task_continue()|d to above is
-      // ready to be prepared.
-      return true;
+      break;
 
     default:
       FATAL() << "Unhandled signal state " << t->ev().type();
-      return false; // not reached
+      break;
   }
 }
 
-/**
- * The execution of |t| has just been resumed, and it most likely has
- * a new event that needs to be processed.  Prepare that new event.
- * Pass |si| to force-override signal status.
- */
-void RecordSession::runnable_state_changed(Task* t, RecordResult* step_result) {
-  // Have to disable context-switching until we know it's safe
-  // to allow switching the context.
-  t->switchable = PREVENT_SWITCH;
-
-  if (t->ptrace_event()) {
-    // A ptrace event arrived. The steps below are irrelevant
-    // and potentially wrong because no ev() was pushed.
-    return;
+bool RecordSession::handle_signal_event(Task* t, StepState* step_state) {
+  int sig = t->pending_sig();
+  if (!sig) {
+    return false;
   }
-
-  siginfo_t* si = nullptr;
-  siginfo_t stash;
-  if (t->has_stashed_sig()) {
-    stash = t->pop_stash_sig();
-    si = &stash;
-    LOG(debug) << "pulled " << signal_name(t->pending_sig()) << " out of stash";
-  }
-
-  if (t->pending_sig() && can_deliver_signals) {
-    // This will either push a new signal event, new
-    // desched + syscall-interruption events, or no-op.
-    handle_signal(t, si);
-  } else if (t->pending_sig()) {
+  if (!can_deliver_signals) {
     // If the initial tracee isn't prepared to handle
     // signals yet, then us ignoring the ptrace
     // notification here will have the side effect of
@@ -764,26 +845,76 @@ void RecordSession::runnable_state_changed(Task* t, RecordResult* step_result) {
               << " because it can't be delivered yet";
     // No events to be recorded, so no syscallbuf updates
     // needed.
-    return;
+    return true;
   }
+  if (is_deterministic_signal(t->get_siginfo()) ||
+      sig == SYSCALLBUF_DESCHED_SIGNAL) {
+    // Don't stash these signals; deliver them immediately.
+    // We don't want them to be reordered around other signals.
+    siginfo_t siginfo = t->get_siginfo();
+    switch (handle_signal(t, &siginfo)) {
+      case SIGNAL_PTRACE_STOP:
+        // Emulated ptrace-stop. Don't run the task again yet.
+        last_task_switchable = ALLOW_SWITCH;
+        break;
+      case DEFER_SIGNAL:
+        ASSERT(t, false) << "Can't defer deterministic or internal signals";
+        break;
+      case SIGNAL_HANDLED:
+        break;
+    }
+    return false;
+  }
+  if (sig == PerfCounters::TIME_SLICE_SIGNAL) {
+    auto& si = t->get_siginfo();
+    /* This implementation will of course fall over if rr tries to
+     * record itself.
+     *
+     * NB: we can't check that the ticks is >= the programmed
+     * target, because this signal may have become pending before
+     * we reset the HPC counters.  There be a way to handle that
+     * more elegantly, but bridge will be crossed in due time.
+     *
+     * We check this here because later, when this signal is delivered,
+     * t->hpc.ticks_fd() may have changed.
+     */
+    ASSERT(t, (PerfCounters::TIME_SLICE_SIGNAL == si.si_signo &&
+               si.si_fd == t->hpc.ticks_fd() && POLL_IN == si.si_code))
+        << "Tracee is using SIGSTKFLT??? (code=" << si.si_code
+        << ", fd=" << si.si_fd << ")";
+  }
+  t->stash_sig();
+  return true;
+}
 
+/**
+ * The execution of |t| has just been resumed, and it most likely has
+ * a new event that needs to be processed.  Prepare that new event.
+ */
+void RecordSession::runnable_state_changed(Task* t, RecordResult* step_result,
+                                           bool can_consume_wait_status,
+                                           StepState* step_state) {
   switch (t->ev().type()) {
     case EV_NOOP:
       t->pop_noop();
       break;
     case EV_SEGV_RDTSC:
+      t->record_current_event();
+      t->pop_event(t->ev().type());
+      break;
     case EV_SCHED:
       t->record_current_event();
       t->pop_event(t->ev().type());
-      t->switchable = ALLOW_SWITCH;
-      break;
-    case EV_SIGNAL:
-      signal_state_changed(t, false);
+      last_task_switchable = ALLOW_SWITCH;
+      step_state->continue_type = DONT_CONTINUE;
       break;
 
     case EV_SENTINEL:
     case EV_SIGNAL_HANDLER:
     case EV_SYSCALL_INTERRUPTION:
+      if (!can_consume_wait_status) {
+        return;
+      }
       // We just entered a syscall.
       if (!maybe_restart_syscall(t)) {
         if (t->vm()->monkeypatcher().try_patch_syscall(t)) {
@@ -793,7 +924,6 @@ void RecordSession::runnable_state_changed(Task* t, RecordResult* step_result) {
         }
 
         t->push_event(SyscallEvent(t->regs().original_syscallno(), t->arch()));
-        rec_before_record_syscall_entry(t, t->ev().Syscall().number);
       }
       ASSERT(t, EV_SYSCALL == t->ev().type());
       check_perf_counters_working(t, step_result);
@@ -802,11 +932,39 @@ void RecordSession::runnable_state_changed(Task* t, RecordResult* step_result) {
       break;
 
     default:
-      ASSERT(t, false) << t->ev()
-                       << " can't be on event stack at start of new event";
+      return;
+  }
+
+  maybe_reset_syscallbuf(t);
+}
+
+bool RecordSession::prepare_to_inject_signal(Task* t, StepState* step_state) {
+  if (!t->has_stashed_sig() || !can_deliver_signals ||
+      step_state->continue_type != CONTINUE) {
+    return false;
+  }
+  siginfo_t si = t->peek_stash_sig();
+  if (si.si_signo == get_ignore_sig()) {
+    LOG(info) << "Declining to deliver " << signal_name(si.si_signo)
+              << " by user request";
+    t->pop_stash_sig();
+    return false;
+  }
+  switch (handle_signal(t, &si)) {
+    case SIGNAL_PTRACE_STOP:
+      // Emulated ptrace-stop. Don't run the task again yet.
+      last_task_switchable = ALLOW_SWITCH;
+      break;
+    case DEFER_SIGNAL:
+      // Leave signal on the stack and continue task execution. We'll try again
+      // later.
+      return false;
+    case SIGNAL_HANDLED:
       break;
   }
-  maybe_reset_syscallbuf(t);
+  step_state->continue_type = DONT_CONTINUE;
+  t->pop_stash_sig();
+  return true;
 }
 
 static string find_syscall_buffer_library() {
@@ -830,7 +988,8 @@ static string find_syscall_buffer_library() {
 }
 
 /*static*/ RecordSession::shr_ptr RecordSession::create(
-    const vector<string>& argv, uint32_t flags) {
+    const vector<string>& argv, uint32_t flags,
+    const vector<string>& extra_env) {
   // The syscallbuf library interposes some critical
   // external symbols like XShmQueryExtension(), so we
   // preload it whether or not syscallbuf is enabled. Indicate here whether
@@ -846,6 +1005,7 @@ static string find_syscall_buffer_library() {
   for (; *envp; ++envp) {
     env.push_back(*envp);
   }
+  env.insert(env.end(), extra_env.begin(), extra_env.end());
 
   char cwd[PATH_MAX] = "";
   getcwd(cwd, sizeof(cwd));
@@ -881,6 +1041,8 @@ static string find_syscall_buffer_library() {
     env.push_back(env_pair);
   }
 
+  env.push_back("LIBGL_ALWAYS_INDIRECT=1");
+
   shr_ptr session(new RecordSession(argv, env, cwd, flags));
   return session;
 }
@@ -892,6 +1054,7 @@ RecordSession::RecordSession(const std::vector<std::string>& argv,
       scheduler_(*this),
       last_recorded_task(nullptr),
       ignore_sig(0),
+      last_task_switchable(PREVENT_SWITCH),
       use_syscall_buffer_(!(flags & DISABLE_SYSCALL_BUF)),
       can_deliver_signals(false) {
   last_recorded_task = Task::spawn(*this, trace_out);
@@ -910,8 +1073,9 @@ RecordSession::RecordResult RecordSession::record_step() {
 
   result.status = STEP_CONTINUE;
 
-  bool by_waitpid;
-  Task* t = scheduler().get_next_thread(last_recorded_task, &by_waitpid);
+  bool did_wait;
+  Task* t = scheduler().get_next_thread(last_recorded_task,
+                                        last_task_switchable, &did_wait);
   if (!t) {
     // The scheduler was waiting for some task to become active, but was
     // interrupted by a signal. Yield to our caller now to give the caller
@@ -921,58 +1085,70 @@ RecordSession::RecordResult RecordSession::record_step() {
   }
   last_recorded_task = t;
 
+  // Have to disable context-switching until we know it's safe
+  // to allow switching the context.
+  last_task_switchable = PREVENT_SWITCH;
+
   LOG(debug) << "line " << t->trace_time() << ": Active task is " << t->tid
              << ". Events:";
 #ifdef DEBUGTAG
   t->log_pending_events();
 #endif
-  ASSERT(t, (!by_waitpid || t->may_be_blocked() || t->ptrace_event()))
-      << "unexpectedly runnable (" << HEX(t->status()) << ") by waitpid";
   if (handle_ptrace_exit_event(t)) {
     // t is dead and has been deleted.
     last_recorded_task = nullptr;
     return result;
   }
 
-  NeedTaskContinue need_task_continue = NEED_TASK_CONTINUE;
-  ForceSyscall force_cont = DEFAULT_CONT;
-
-  handle_ptrace_event(t, &force_cont);
-
   if (t->unstable) {
-    // Do not record non-ptrace events for tasks in
+    // Do not record non-ptrace-exit events for tasks in
     // an unstable exit. We can't replay them.
     LOG(debug) << "Task in unstable exit; "
                   "refusing to record non-ptrace events";
+    last_task_switchable = ALLOW_SWITCH;
     return result;
   }
 
-  switch (t->ev().type()) {
-    case EV_DESCHED:
-      desched_state_changed(t);
-      return result;
-    case EV_SYSCALL:
-      syscall_state_changed(t, by_waitpid);
-      return result;
-    case EV_SIGNAL_DELIVERY: {
-      if (signal_state_changed(t, by_waitpid)) {
-        need_task_continue = DONT_NEED_TASK_CONTINUE;
-        break;
-      }
+  StepState step_state(CONTINUE);
+
+  if (!(did_wait && handle_ptrace_event(t, &step_state)) &&
+      !(did_wait && handle_signal_event(t, &step_state))) {
+    runnable_state_changed(t, &result, did_wait, &step_state);
+    if (result.status != STEP_CONTINUE ||
+        step_state.continue_type == DONT_CONTINUE) {
       return result;
     }
-    default:
-      /* No special handling needed; continue on
-       * below. */
-      break;
+
+    switch (t->ev().type()) {
+      case EV_DESCHED:
+        desched_state_changed(t);
+        break;
+      case EV_SYSCALL:
+        syscall_state_changed(t, &step_state);
+        break;
+      case EV_SIGNAL:
+      case EV_SIGNAL_DELIVERY:
+        signal_state_changed(t, &step_state);
+        break;
+      default:
+        break;
+    }
   }
 
-  if (!t->has_stashed_sig()) {
-    resume_execution(t, need_task_continue, force_cont);
-  }
+  // We try to inject a signal if there's one pending; otherwise we continue
+  // task execution.
+  if (!prepare_to_inject_signal(t, &step_state) &&
+      step_state.continue_type != DONT_CONTINUE) {
+    // Ensure that we aren't allowing switches away from a running task.
+    // Only tasks blocked in a syscall can be switched away from, otherwise
+    // we have races.
+    ASSERT(t, last_task_switchable == PREVENT_SWITCH || t->unstable ||
+                  t->may_be_blocked());
 
-  // runnable_state_changed can detect errors that
-  runnable_state_changed(t, &result);
+    debug_exec_state("EXEC_START", t);
+
+    task_continue(t, step_state);
+  }
 
   return result;
 }
@@ -985,9 +1161,10 @@ void RecordSession::terminate_recording() {
   LOG(info) << "Processing termination request ...";
   LOG(info) << "  recording final TRACE_TERMINATION event ...";
 
-  TraceFrame frame(
-      trace_out.time(), last_recorded_task ? last_recorded_task->tid : 0,
-      Event(EV_TRACE_TERMINATION, NO_EXEC_INFO, RR_NATIVE_ARCH).encode());
+  TraceFrame frame(trace_out.time(),
+                   last_recorded_task ? last_recorded_task->tid : 0,
+                   Event(EV_TRACE_TERMINATION, NO_EXEC_INFO, RR_NATIVE_ARCH),
+                   last_recorded_task ? last_recorded_task->tick_count() : 0);
   trace_out.write_frame(frame);
   trace_out.close();
 }
