@@ -15,6 +15,7 @@
 #include "task.h"
 
 using namespace rr;
+using namespace std;
 
 struct RegisterValue {
   // The name of this register.
@@ -263,7 +264,6 @@ void Registers::print_register_file_for_trace_arch(
         assert(0 && "bad register size");
     }
   }
-  fprintf(f, "\n");
 }
 
 void Registers::print_register_file_compact(FILE* f) const {
@@ -444,6 +444,27 @@ size_t Registers::read_register(uint8_t* buf, GdbRegister regno,
   RR_ARCH_FUNCTION(read_register_arch, arch(), buf, regno, defined);
 }
 
+template <typename Arch>
+size_t Registers::read_register_by_user_offset_arch(uint8_t* buf,
+                                                    uintptr_t offset,
+                                                    bool* defined) const {
+  for (size_t regno = 0; regno < RegisterInfo<Arch>::num_registers; ++regno) {
+    RegisterValue& rv = RegisterInfo<Arch>::registers[regno];
+    if (rv.offset == offset) {
+      return read_register_arch<Arch>(buf, GdbRegister(regno), defined);
+    }
+  }
+
+  *defined = false;
+  return 0;
+}
+
+size_t Registers::read_register_by_user_offset(uint8_t* buf, uintptr_t offset,
+                                               bool* defined) const {
+  RR_ARCH_FUNCTION(read_register_by_user_offset_arch, arch(), buf, offset,
+                   defined);
+}
+
 template <typename Arch>
 void Registers::write_register_arch(GdbRegister regno, const uint8_t* value,
                                     size_t value_size) {
@@ -498,9 +519,9 @@ void convert_x86(X86Arch::user_regs_struct& x86,
   same(x86.xss, x64.ss);
 }
 
-static void to_x86_narrow(int32_t& r32, uint64_t& r64) { r32 = r64; }
-static void to_x86_same(int32_t& r32, uint32_t& r64) { r32 = r64; }
-static void from_x86_narrow(int32_t& r32, uint64_t& r64) {
+void to_x86_narrow(int32_t& r32, uint64_t& r64) { r32 = r64; }
+void to_x86_same(int32_t& r32, uint32_t& r64) { r32 = r64; }
+void from_x86_narrow(int32_t& r32, uint64_t& r64) {
   // We must zero-extend 32-bit register values to 64-bit values when we
   // do a PTRACE_SETREGS. Most of the time the upper 32 bits are irrelevant for
   // a 32-bit tracee, but when setting up a signal handler frame, the kernel
@@ -509,7 +530,7 @@ static void from_x86_narrow(int32_t& r32, uint64_t& r64) {
   // sign-extended to point outside the 32-bit address space.
   r64 = (uint32_t)r32;
 }
-static void from_x86_same(int32_t& r32, uint32_t& r64) { r64 = r32; }
+void from_x86_same(int32_t& r32, uint32_t& r64) { r64 = r32; }
 
 void Registers::set_from_ptrace(const struct user_regs_struct& ptrace_regs) {
   if (arch() == NativeArch::arch()) {
@@ -529,7 +550,7 @@ void Registers::set_from_ptrace(const struct user_regs_struct& ptrace_regs) {
  * 64-bit rr. In that case the user_regs_struct is 64-bit and we copy
  * the 32-bit register values from u.x86regs into it.
  */
-struct user_regs_struct Registers::get_ptrace() {
+struct user_regs_struct Registers::get_ptrace() const {
   struct user_regs_struct result;
   if (arch() == NativeArch::arch()) {
     memcpy(&result, &u, sizeof(result));
@@ -539,6 +560,55 @@ struct user_regs_struct Registers::get_ptrace() {
   assert(arch() == x86 && NativeArch::arch() == x86_64);
   memset(&result, 0, sizeof(result));
   convert_x86<from_x86_narrow, from_x86_same>(
-      u.x86regs, *reinterpret_cast<X64Arch::user_regs_struct*>(&result));
+      const_cast<Registers*>(this)->u.x86regs,
+      *reinterpret_cast<X64Arch::user_regs_struct*>(&result));
   return result;
 }
+
+vector<uint8_t> Registers::get_ptrace_for_arch(SupportedArch arch) const {
+  Registers tmp_regs(arch);
+  tmp_regs.set_from_ptrace(get_ptrace());
+
+  vector<uint8_t> result;
+  switch (arch) {
+    case x86:
+      result.resize(sizeof(u.x86regs));
+      memcpy(result.data(), &u.x86regs, result.size());
+      break;
+    case x86_64:
+      result.resize(sizeof(u.x64regs));
+      memcpy(result.data(), &u.x64regs, result.size());
+      break;
+    default:
+      assert(0 && "Unknown arch");
+      break;
+  }
+  return result;
+}
+
+bool Registers::clear_singlestep_flag() {
+  switch (arch()) {
+    case x86:
+      if (u.x86regs.eflags & X86_TF_FLAG) {
+        u.x86regs.eflags &= ~X86_TF_FLAG;
+        return true;
+      }
+      return false;
+    case x86_64:
+      if (u.x64regs.eflags & X86_TF_FLAG) {
+        u.x64regs.eflags &= ~X86_TF_FLAG;
+        return true;
+      }
+      return false;
+    default:
+      assert(0 && "Unknown arch");
+      return false;
+  }
+}
+
+ostream& operator<<(ostream& stream, const Registers& r) {
+  stream << "{ args:(" << HEX(r.arg1()) << "," << HEX(r.arg2()) << ","
+         << HEX(r.arg3()) << "," << HEX(r.arg4()) << "," << HEX(r.arg5()) << ","
+         << r.arg6() << ") orig_syscall:" << r.original_syscallno() << " }";
+  return stream;
+}
