@@ -19,6 +19,8 @@
 
 using namespace std;
 
+static int DUMP_STATS_PERIOD = 0;
+
 class ReplayCommand : public Command {
 public:
   virtual int run(std::vector<std::string>& args);
@@ -40,13 +42,18 @@ ReplayCommand ReplayCommand::singleton(
     "<EVENT-NUM>\n"
     "                             in the trace.  See -m above.\n"
     "  -p, --onprocess=<PID>|<COMMAND>\n"
-    "                             start a debug server when <PID> or <COMMAND>\n"
-    "                             has been exec()d, AND the target event has been\n"
+    "                             start a debug server when <PID> or "
+    "<COMMAND>\n"
+    "                             has been exec()d, AND the target event has "
+    "been\n"
     "                             reached.\n"
     "  -q, --no-redirect-output   don't replay writes to stdout/stderr\n"
     "  -s, --dbgport=<PORT>       only start a debug server on <PORT>;\n"
     "                             don't automatically launch the debugger\n"
     "                             client too.\n"
+    "  -t, --trace=<EVENT>        singlestep instructions and dump register\n"
+    "                             states when replaying towards <EVENT> or\n"
+    "                             later\n"
     "  -x, --gdb-x=<FILE>         execute gdb commands from <FILE>\n");
 
 struct ReplayFlags {
@@ -55,6 +62,8 @@ struct ReplayFlags {
   // been "created".
   TraceFrame::Time goto_event;
 
+  TraceFrame::Time singlestep_to_event;
+
   pid_t target_process;
 
   string target_command;
@@ -86,6 +95,7 @@ struct ReplayFlags {
 
   ReplayFlags()
       : goto_event(0),
+        singlestep_to_event(0),
         target_process(0),
         process_created_how(CREATED_NONE),
         dont_launch_debugger(false),
@@ -102,6 +112,7 @@ static bool parse_replay_arg(std::vector<std::string>& args,
   static const OptionSpec options[] = { { 'a', "autopilot", NO_PARAMETER },
                                         { 's', "dbgport", HAS_PARAMETER },
                                         { 'g', "goto", HAS_PARAMETER },
+                                        { 't', "trace", HAS_PARAMETER },
                                         { 'q', "no-redirect-output",
                                           NO_PARAMETER },
                                         { 'f', "onfork", HAS_PARAMETER },
@@ -151,6 +162,12 @@ static bool parse_replay_arg(std::vector<std::string>& args,
       flags.dbg_port = opt.int_value;
       flags.dont_launch_debugger = true;
       break;
+    case 't':
+      if (!opt.verify_valid_int(1, INT32_MAX)) {
+        return false;
+      }
+      flags.singlestep_to_event = opt.int_value;
+      break;
     case 'x':
       flags.gdb_command_file_path = opt.value;
       break;
@@ -182,6 +199,30 @@ static int find_pid_for_command(const string& trace_dir,
   return -1;
 }
 
+static bool pid_exists(const string& trace_dir, pid_t pid) {
+  TraceReader trace(trace_dir);
+
+  while (trace.good()) {
+    auto e = trace.read_task_event();
+    if (e.tid() == pid) {
+      return true;
+    }
+  }
+  return false;
+}
+
+static bool pid_execs(const string& trace_dir, pid_t pid) {
+  TraceReader trace(trace_dir);
+
+  while (trace.good()) {
+    auto e = trace.read_task_event();
+    if (e.tid() == pid && e.type() == TraceTaskEvent::EXEC) {
+      return true;
+    }
+  }
+  return false;
+}
+
 // The parent process waits until the server, |waiting_for_child|, creates a
 // debug socket. Then the parent exec()s the debugger over itself. While it's
 // waiting for the child, this is the child's pid.
@@ -200,25 +241,59 @@ static void set_sig_blockedness(int sig, int blockedness) {
   }
 }
 
-static void setup_session_from_flags(ReplaySession& session,
-                                     const ReplayFlags& flags) {
-  session.set_redirect_stdio(flags.redirect);
+static ReplaySession::Flags session_flags(ReplayFlags flags) {
+  ReplaySession::Flags result;
+  result.redirect_stdio = flags.redirect;
+  return result;
+}
+
+static uint64_t to_microseconds(const struct timeval& tv) {
+  return (uint64_t)tv.tv_sec * 1000000 + tv.tv_usec;
 }
 
 static void serve_replay_no_debugger(const string& trace_dir,
                                      const ReplayFlags& flags) {
   ReplaySession::shr_ptr replay_session = ReplaySession::create(trace_dir);
-  setup_session_from_flags(*replay_session, flags);
+  replay_session->set_flags(session_flags(flags));
+  uint32_t step_count = 0;
+  struct timeval last_dump_time;
+  Session::Statistics last_stats;
+  gettimeofday(&last_dump_time, NULL);
 
   while (true) {
-    auto result = replay_session->replay_step(ReplaySession::RUN_CONTINUE);
+    RunCommand cmd = RUN_CONTINUE;
+    if (flags.singlestep_to_event > 0 &&
+        replay_session->trace_reader().time() >= flags.singlestep_to_event) {
+      cmd = RUN_SINGLESTEP;
+      fputs("Stepping from:", stderr);
+      Task* t = replay_session->current_task();
+      t->regs().print_register_file_compact(stderr);
+      fprintf(stderr, " ticks:%" PRId64 "\n", t->tick_count());
+    }
+    auto result = replay_session->replay_step(cmd);
+    ++step_count;
+    if (DUMP_STATS_PERIOD > 0 && step_count % DUMP_STATS_PERIOD == 0) {
+      struct timeval now;
+      gettimeofday(&now, NULL);
+      Session::Statistics stats = replay_session->statistics();
+      printf(
+          "[ReplayStatistics] ticks %lld syscalls %lld bytes_written %lld "
+          "microseconds %lld\n",
+          (long long)(stats.ticks_processed - last_stats.ticks_processed),
+          (long long)(stats.syscalls_performed - last_stats.syscalls_performed),
+          (long long)(stats.bytes_written - last_stats.bytes_written),
+          (long long)(to_microseconds(now) - to_microseconds(last_dump_time)));
+      last_dump_time = now;
+      last_stats = stats;
+    }
 
-    if (result.status == ReplaySession::REPLAY_EXITED) {
+    if (result.status == REPLAY_EXITED) {
       break;
     }
-    assert(result.status == ReplaySession::REPLAY_CONTINUE);
-    assert(result.break_status.reason == Session::BREAK_NONE ||
-           result.break_status.reason == Session::BREAK_SIGNAL);
+    assert(result.status == REPLAY_CONTINUE);
+    assert(result.break_status.watchpoints_hit.empty());
+    assert(!result.break_status.breakpoint_hit);
+    assert(cmd == RUN_SINGLESTEP || !result.break_status.singlestep_complete);
   }
 
   LOG(info) << ("Replayer successfully finished.");
@@ -264,10 +339,9 @@ static int replay(const string& trace_dir, const ReplayFlags& flags) {
       serve_replay_no_debugger(trace_dir, flags);
     } else {
       auto session = ReplaySession::create(trace_dir);
-      setup_session_from_flags(*session, flags);
       GdbServer::ConnectionFlags conn_flags;
       conn_flags.dbg_port = flags.dbg_port;
-      GdbServer::serve(session, target, conn_flags);
+      GdbServer::serve(session, target, conn_flags, session_flags(flags));
     }
     return 0;
   }
@@ -294,11 +368,10 @@ static int replay(const string& trace_dir, const ReplayFlags& flags) {
     // block it.
     set_sig_blockedness(SIGINT, SIG_BLOCK);
     auto session = ReplaySession::create(trace_dir);
-    setup_session_from_flags(*session, flags);
     GdbServer::ConnectionFlags conn_flags;
     conn_flags.dbg_port = flags.dbg_port;
     conn_flags.debugger_params_write_pipe = &debugger_params_write_pipe;
-    GdbServer::serve(session, target, conn_flags);
+    GdbServer::serve(session, target, conn_flags, session_flags(flags));
     return 0;
   }
   // Ensure only the child has the write end of the pipe open. Then if
@@ -335,12 +408,18 @@ static int replay(const string& trace_dir, const ReplayFlags& flags) {
 }
 
 int ReplayCommand::run(std::vector<std::string>& args) {
-  ReplayFlags flags;
-  while (parse_replay_arg(args, flags)) {
-  }
-
+  bool found_dir = false;
   string trace_dir;
-  if (!parse_optional_trace_dir(args, &trace_dir)) {
+  ReplayFlags flags;
+
+  while (!args.empty()) {
+    if (parse_replay_arg(args, flags)) {
+      continue;
+    }
+    if (!found_dir && parse_optional_trace_dir(args, &trace_dir)) {
+      found_dir = true;
+      continue;
+    }
     print_help(stderr);
     return 1;
   }
@@ -354,6 +433,20 @@ int ReplayCommand::run(std::vector<std::string>& args) {
       return 2;
     }
   }
+  if (flags.process_created_how != ReplayFlags::CREATED_NONE) {
+    if (!pid_exists(trace_dir, flags.target_process)) {
+      fprintf(stderr, "No process %d found in trace. Try 'rr ps'.\n",
+              flags.target_process);
+      return 2;
+    }
+    if (flags.process_created_how == ReplayFlags::CREATED_EXEC &&
+        !pid_execs(trace_dir, flags.target_process)) {
+      fprintf(stderr, "Process %d never exec()ed. Try 'rr ps', or use "
+                      "'-f'.\n",
+              flags.target_process);
+      return 2;
+    }
+  }
 
   assert_prerequisites();
   check_performance_settings();
