@@ -3,6 +3,8 @@
 #ifndef RR_REPLAY_SESSION_H_
 #define RR_REPLAY_SESSION_H_
 
+#include <memory>
+
 #include "CPUIDBugDetector.h"
 #include "DiversionSession.h"
 #include "EmuFs.h"
@@ -73,9 +75,6 @@ struct ReplayFlushBufferedSyscallState {
 enum ReplayTraceStepType {
   TSTEP_NONE,
 
-  /* Frame has been replayed, done. */
-  TSTEP_RETIRE,
-
   /* Enter/exit a syscall.  |syscall| describe what should be
    * done at entry/exit. */
   TSTEP_ENTER_SYSCALL,
@@ -101,6 +100,9 @@ enum ReplayTraceStepType {
   /* Emulate arming or disarming the desched event.  |desched|
    * tracks the replay state. */
   TSTEP_DESCHED,
+
+  /* Frame has been replayed, done. */
+  TSTEP_RETIRE,
 };
 
 enum ExecOrEmulate {
@@ -128,13 +130,6 @@ struct ReplayTraceStep {
       /* Is the kernel entry and exit for this
        * syscall emulated, that is, not executed? */
       ExecOrEmulate emu;
-      /* The number of outparam arguments that are
-       * set from what was recorded.
-       * Only used when action is TSTEP_EXIT_SYSCALL. */
-      ssize_t num_emu_args;
-      /* Nonzero if the return from the syscall
-       * should be emulated.  |emu| implies this. */
-      ExecOrEmulateReturn emu_ret;
     } syscall;
 
     int signo;
@@ -150,6 +145,53 @@ struct ReplayTraceStep {
   };
 };
 
+enum ReplayStatus {
+  // Some execution was replayed. replay_step() can be called again.
+  REPLAY_CONTINUE,
+  // All tracees are dead. replay_step() should not be called again.
+  REPLAY_EXITED
+};
+
+struct ReplayResult {
+  ReplayResult(ReplayStatus status = REPLAY_CONTINUE)
+      : status(status), did_fast_forward(false) {}
+  ReplayStatus status;
+  BreakStatus break_status;
+  // True if we did a fast-forward operation, in which case
+  // break_status.singlestep_complete might indicate the completion of more
+  // than one instruction.
+  bool did_fast_forward;
+};
+
+/**
+ * An indicator of how much progress the ReplaySession has made within a given
+ * (TraceFrame::Time, Ticks) pair. These can only be used for comparisons, to
+ * check whether two ReplaySessions are in the same state and to help
+ * order their states temporally.
+ */
+class ReplayStepKey {
+public:
+  /**
+   * Construct the "none" key; this value is before or equal to every other
+   * key value.
+   */
+  ReplayStepKey() : action(TSTEP_NONE) {}
+  explicit ReplayStepKey(ReplayTraceStepType action) : action(action) {}
+
+  bool operator==(const ReplayStepKey& other) const {
+    return action == other.action;
+  }
+  bool operator<(const ReplayStepKey& other) const {
+    return action < other.action;
+  }
+
+  bool in_execution() const { return action != TSTEP_NONE; }
+  int as_int() const { return (int)action; }
+
+private:
+  ReplayTraceStepType action;
+};
+
 /** Encapsulates additional session state related to replay. */
 class ReplaySession : public Session {
 public:
@@ -165,9 +207,21 @@ public:
    *
    * This operation is also called "checkpointing" the replay
    * session.
+   *
+   * The returned clone is only partially initialized. This uses less
+   * system resources than a fully-initialized session, so if you're going
+   * to keep a session around inactive, keep the clone and not the original
+   * session. Partially initialized sessions automatically finish
+   * initializing when necessary.
    */
   shr_ptr clone();
 
+  /**
+   * Return true if we're in a state where it's OK to clone. For example,
+   * we can't clone in some syscalls.
+   */
+  bool can_clone();
+
   /**
    * Like |clone()|, but return a session in "diversion" mode,
    * which allows free execution.
@@ -180,6 +234,7 @@ public:
   void gc_emufs();
 
   TraceReader& trace_reader() { return trace_in; }
+  const TraceReader& trace_reader() const { return trace_in; }
 
   /**
    * The trace record that we are working on --- the next event
@@ -190,23 +245,17 @@ public:
   /**
    * The Task for the current trace record.
    */
-  Task* current_task() const { return find_task(trace_frame.tid()); }
+  Task* current_task() {
+    finish_initializing();
+    return find_task(trace_frame.tid());
+  }
 
   /**
-   * Set |tgid| as the one that's being debugged in this
-   * session.
-   *
-   * Little hack: technically replayer doesn't know about the
-   * fact that GdbConnection hides all but one tgid from the gdb
-   * client.  But to recognize the last_task below (another
-   * little hack), we need to known when an exiting thread from
-   * the target task group is the last.
+   * The current ReplayStepKey.
    */
-  void set_debugged_tgid(pid_t tgid) {
-    assert(0 == tgid_debugged);
-    tgid_debugged = tgid;
+  ReplayStepKey current_step_key() const {
+    return ReplayStepKey(current_step.action);
   }
-  pid_t debugged_tgid() const { return tgid_debugged; }
 
   /**
    * If we've finished replaying (all tracees terminated), return the last
@@ -221,25 +270,41 @@ public:
    */
   static shr_ptr create(const std::string& dir);
 
-  enum ReplayStatus {
-    // Some execution was replayed. replay_step() can be called again.
-    REPLAY_CONTINUE,
-    // All tracees are dead. replay_step() should not be called again.
-    REPLAY_EXITED
-  };
-  struct ReplayResult {
-    ReplayStatus status;
-    // When status == STEP_CONTINUE
-    BreakStatus break_status;
+  struct StepConstraints {
+    explicit StepConstraints(RunCommand command)
+        : command(command), stop_at_time(0), ticks_target(0) {}
+    RunCommand command;
+    TraceFrame::Time stop_at_time;
+    Ticks ticks_target;
+    // When the RunCommand is RUN_SINGLESTEP_FAST_FORWARD, stop if the next
+    // singlestep would enter one of the register states in this (null-
+    // terminated) list. RUN_SINGLESTEP_FAST_FORWARD will always singlestep
+    // at least once regardless.
+    std::vector<const Registers*> stop_before_states;
+
+    bool is_singlestep() const {
+      return command == RUN_SINGLESTEP ||
+             command == RUN_SINGLESTEP_FAST_FORWARD;
+    }
   };
   /**
    * Take a single replay step.
    * Ensure we stop at event stop_at_time. If this is not specified,
    * optimizations may cause a replay_step to pass straight through
    * stop_at_time.
+   * Outside of replay_step, no internal breakpoints will be set for any
+   * task in this session.
+   * Stop when the current event reaches stop_at_time (i.e. this event has
+   * is the next event to be replayed).
+   * If ticks_target is nonzero, stop before the current task's ticks
+   * reaches ticks_target (but not too far before, unless we hit a breakpoint
+   * or stop_at_time). Only useful for RUN_CONTINUE.
+   * Always stops on a switch to a new task.
    */
-  ReplayResult replay_step(RunCommand command = RUN_CONTINUE,
-                           TraceFrame::Time stop_at_time = 0);
+  ReplayResult replay_step(const StepConstraints& constraints);
+  ReplayResult replay_step(RunCommand command) {
+    return replay_step(StepConstraints(command));
+  }
 
   virtual ReplaySession* as_replay() { return this; }
 
@@ -254,31 +319,34 @@ public:
    */
   static bool is_ignored_signal(int sig);
 
-  bool redirect_stdio() { return redirect_stdio_; }
+  struct Flags {
+    Flags() : redirect_stdio(false) {}
+    Flags(const Flags& other) = default;
+    bool redirect_stdio;
+  };
+  bool redirect_stdio() { return flags.redirect_stdio; }
 
-  void set_redirect_stdio(bool redirect) { redirect_stdio_ = redirect; }
+  void set_flags(const Flags& flags) { this->flags = flags; }
 
 private:
   ReplaySession(const std::string& dir)
       : emu_fs(EmuFs::create()),
         last_debugged_task(nullptr),
-        tgid_debugged(0),
         trace_in(dir),
         trace_frame(),
-        current_step(),
-        redirect_stdio_(false) {
+        current_step() {
     advance_to_next_trace_frame(0);
   }
 
   ReplaySession(const ReplaySession& other)
-      : emu_fs(other.emu_fs->clone()),
+      : Session(other),
+        emu_fs(other.emu_fs->clone()),
         last_debugged_task(nullptr),
-        tgid_debugged(other.tgid_debugged),
         trace_in(other.trace_in),
         trace_frame(other.trace_frame),
         current_step(other.current_step),
         cpuid_bug_detector(other.cpuid_bug_detector),
-        redirect_stdio_(other.redirect_stdio_) {
+        flags(other.flags) {
     assert(!other.last_debugged_task);
   }
 
@@ -294,8 +362,6 @@ private:
     last_debugged_task = t;
   }
 
-  void copy_state_to(Session& dest, EmuFs& dest_emu_fs);
-
   const struct syscallbuf_hdr* syscallbuf_flush_buffer_hdr() {
     return (const struct syscallbuf_hdr*)syscallbuf_flush_buffer_array;
   }
@@ -303,16 +369,18 @@ private:
   void setup_replay_one_trace_frame(Task* t);
   void advance_to_next_trace_frame(TraceFrame::Time stop_at_time);
   Completion emulate_signal_delivery(Task* oldtask, int sig,
-                                     TraceFrame::Time stop_at_time);
-  Completion try_one_trace_step(Task* t, RunCommand stepi,
-                                TraceFrame::Time stop_at_time);
+                                     const StepConstraints& constraints);
+  Completion try_one_trace_step(Task* t,
+                                const StepConstraints& step_constraints);
   Completion cont_syscall_boundary(Task* t, ExecOrEmulate emu,
-                                   RunCommand stepi);
-  Completion enter_syscall(Task* t, RunCommand stepi);
-  Completion exit_syscall(Task* t, RunCommand stepi);
+                                   const StepConstraints& constraints);
+  Completion enter_syscall(Task* t, const StepConstraints& constraints);
+  Completion exit_syscall(Task* t, const StepConstraints& constraints);
   Ticks get_ticks_slack(Task* t);
   void check_ticks_consistency(Task* t, const Event& ev);
-  void continue_or_step(Task* t, RunCommand stepi, int64_t tick_period = 0);
+  void check_pending_sig(Task* t);
+  void continue_or_step(Task* t, const StepConstraints& constraints,
+                        int64_t tick_period = 0);
   enum ExecStateType {
     UNKNOWN,
     NOT_AT_TARGET,
@@ -320,31 +388,39 @@ private:
   };
   TrapType compute_trap_type(Task* t, int target_sig,
                              SignalDeterministic deterministic,
-                             ExecStateType exec_state, RunCommand stepi);
+                             ExecStateType exec_state,
+                             const StepConstraints& constraints);
   bool is_debugger_trap(Task* t, int target_sig,
                         SignalDeterministic deterministic,
-                        ExecStateType exec_state, RunCommand stepi);
+                        ExecStateType exec_state,
+                        const StepConstraints& constraints);
   Completion advance_to(Task* t, const Registers& regs, int sig,
-                        RunCommand stepi, Ticks ticks);
-  Completion emulate_deterministic_signal(Task* t, int sig, RunCommand stepi);
-  Completion emulate_async_signal(Task* t, int sig, RunCommand stepi,
+                        const StepConstraints& constraints, Ticks ticks);
+  Completion advance_to_ticks_target(Task* t,
+                                     const StepConstraints& constraints);
+  Completion emulate_deterministic_signal(Task* t, int sig,
+                                          const StepConstraints& constraints);
+  Completion emulate_async_signal(Task* t, int sig,
+                                  const StepConstraints& constraints,
                                   Ticks ticks);
   Completion skip_desched_ioctl(Task* t, ReplayDeschedState* ds,
-                                RunCommand stepi);
+                                const StepConstraints& constraints);
   void prepare_syscallbuf_records(Task* t);
-  Completion flush_one_syscall(Task* t, RunCommand stepi);
-  Completion flush_syscallbuf(Task* t, RunCommand stepi);
-  Completion patch_next_syscall(Task* t, RunCommand stepi);
-  bool is_last_interesting_task(Task* t);
+  Completion flush_one_syscall(Task* t, const StepConstraints& constraints);
+  Completion flush_syscallbuf(Task* t, const StepConstraints& constraints);
+  Completion patch_next_syscall(Task* t, const StepConstraints& constraints);
+  void check_approaching_ticks_target(Task* t,
+                                      const StepConstraints& constraints,
+                                      BreakStatus& break_status);
 
   std::shared_ptr<EmuFs> emu_fs;
   Task* last_debugged_task;
-  pid_t tgid_debugged;
   TraceReader trace_in;
   TraceFrame trace_frame;
   ReplayTraceStep current_step;
   CPUIDBugDetector cpuid_bug_detector;
-  bool redirect_stdio_;
+  Flags flags;
+  bool did_fast_forward;
   /**
    * Buffer for recorded syscallbuf bytes.  By definition buffer flushes
    * must be replayed sequentially, so we can use one buffer for all
