@@ -25,7 +25,30 @@ using namespace std;
 // MUST increment this version number.  Otherwise users' old traces
 // will become unreplayable and they won't know why.
 //
-#define TRACE_VERSION 17
+#define TRACE_VERSION 26
+
+struct SubstreamData {
+  const char* name;
+  size_t block_size;
+  int threads;
+};
+
+static const SubstreamData substreams[TraceStream::SUBSTREAM_COUNT] = {
+  { "events", 1024 * 1024, 1 },
+  { "data_header", 1024 * 1024, 1 },
+  { "data", 8 * 1024 * 1024, 3 },
+  { "mmaps", 64 * 1024, 1 },
+  { "tasks", 64 * 1024, 1 }
+};
+
+static const SubstreamData& substream(TraceStream::Substream s) {
+  return substreams[s];
+}
+
+static TraceStream::Substream operator++(TraceStream::Substream& s) {
+  s = (TraceStream::Substream)(s + 1);
+  return s;
+}
 
 static string default_rr_trace_dir() { return string(getenv("HOME")) + "/.rr"; }
 
@@ -63,35 +86,56 @@ static void ensure_default_rr_trace_dir() {
   }
 }
 
+string TraceStream::path(Substream s) {
+  return trace_dir + "/" + substream(s).name;
+}
+
 bool TraceWriter::good() const {
-  return events.good() && data.good() && data_header.good() && mmaps.good() &&
-         tasks.good();
+  for (auto& w : writers) {
+    if (!w->good()) {
+      return false;
+    }
+  }
+  return true;
 }
 
 bool TraceReader::good() const {
-  return events.good() && data.good() && data_header.good() && mmaps.good() &&
-         tasks.good();
+  for (auto& r : readers) {
+    if (!r->good()) {
+      return false;
+    }
+  }
+  return true;
 }
 
+struct BasicInfo {
+  TraceFrame::Time global_time;
+  pid_t tid_;
+  EncodedEvent ev;
+  Ticks ticks_;
+};
+
 void TraceWriter::write_frame(const TraceFrame& frame) {
-  events.write(&frame.basic_info, sizeof(frame.basic_info));
+  auto& events = writer(EVENTS);
+
+  BasicInfo basic_info = { frame.time(),           frame.tid(),
+                           frame.event().encode(), frame.ticks() };
+  events << basic_info;
   if (!events.good()) {
-    FATAL() << "Tried to save " << sizeof(frame.basic_info)
+    FATAL() << "Tried to save " << sizeof(basic_info)
             << " bytes to the trace, but failed";
   }
   // TODO: only store exec info for non-async-sig events when
   // debugging assertions are enabled.
-  if (frame.event().has_exec_info == HAS_EXEC_INFO) {
-    events.write(&frame.exec_info, sizeof(frame.exec_info));
+  if (frame.event().has_exec_info() == HAS_EXEC_INFO) {
+    events << frame.regs() << frame.extra_perf_values();
     if (!events.good()) {
-      FATAL() << "Tried to save " << sizeof(frame.exec_info)
-              << " bytes to the trace, but failed";
+      FATAL() << "Tried to save registers to the trace, but failed";
     }
 
     int extra_reg_bytes = frame.extra_regs().data_size();
     char extra_reg_format = (char)frame.extra_regs().format();
-    events.write(&extra_reg_format, sizeof(extra_reg_format));
-    events.write((char*)&extra_reg_bytes, sizeof(extra_reg_bytes));
+    events << extra_reg_format << extra_reg_bytes;
     if (!events.good()) {
       FATAL() << "Tried to save "
               << sizeof(extra_reg_bytes) + sizeof(extra_reg_format)
@@ -106,6 +150,9 @@ void TraceWriter::write_frame(const TraceFrame& frame) {
       }
     }
   }
+  if (frame.event().is_signal_event()) {
+    events << frame.event().Signal().signal_data();
+  }
 
   tick_time();
 }
@@ -113,95 +160,52 @@ void TraceWriter::write_frame(const TraceFrame& frame) {
 TraceFrame TraceReader::read_frame() {
   // Read the common event info first, to see if we also have
   // exec info to read.
-  TraceFrame frame;
-  events.read(&frame.basic_info, sizeof(frame.basic_info));
-  if (frame.event().has_exec_info) {
-    events.read(&frame.exec_info, sizeof(frame.exec_info));
+  auto& events = reader(EVENTS);
+  BasicInfo basic_info;
+  events >> basic_info;
+  TraceFrame frame(basic_info.global_time, basic_info.tid_,
+                   Event(basic_info.ev), basic_info.ticks_);
+  if (frame.event().has_exec_info() == HAS_EXEC_INFO) {
+    events >> frame.recorded_regs >> frame.extra_perf;
 
     int extra_reg_bytes;
     char extra_reg_format;
-    events.read(&extra_reg_format, sizeof(extra_reg_format));
-    events.read((char*)&extra_reg_bytes, sizeof(extra_reg_bytes));
+    events >> extra_reg_format >> extra_reg_bytes;
     if (extra_reg_bytes > 0) {
-      std::vector<uint8_t> data;
+      vector<uint8_t> data;
       data.resize(extra_reg_bytes);
       events.read((char*)data.data(), extra_reg_bytes);
+      frame.recorded_extra_regs.set_arch(frame.event().arch());
       frame.recorded_extra_regs.set_to_raw_data(
           (ExtraRegisters::Format)extra_reg_format, data);
     } else {
       assert(extra_reg_format == ExtraRegisters::NONE);
-      frame.recorded_extra_regs = ExtraRegisters();
+      frame.recorded_extra_regs = ExtraRegisters(frame.event().arch());
     }
   }
+  if (frame.event().is_signal_event()) {
+    uint64_t signal_data;
+    events >> signal_data;
+    frame.ev.Signal().set_signal_data(signal_data);
+  }
 
   tick_time();
   assert(time() == frame.time());
   return frame;
 }
 
-template <typename T>
-static CompressedWriter& operator<<(CompressedWriter& out, const T& value) {
-  out.write(&value, sizeof(value));
-  return out;
-}
-
-template <typename T>
-static CompressedReader& operator>>(CompressedReader& in, T& value) {
-  in.read(&value, sizeof(value));
-  return in;
-}
-
-static CompressedWriter& operator<<(CompressedWriter& out,
-                                    const string& value) {
-  out.write(value.c_str(), value.size() + 1);
-  return out;
-}
-
-static CompressedReader& operator>>(CompressedReader& in, string& value) {
-  value.empty();
-  while (true) {
-    char ch;
-    in.read(&ch, 1);
-    if (ch == 0) {
-      break;
-    }
-    value.append(1, ch);
-  }
-  return in;
-}
-
-template <typename T>
-static CompressedWriter& operator<<(CompressedWriter& out,
-                                    const std::vector<T>& value) {
-  out << value.size();
-  for (auto& i : value) {
-    out << i;
-  }
-  return out;
-}
-
-template <typename T>
-static CompressedReader& operator>>(CompressedReader& in,
-                                    std::vector<T>& value) {
-  size_t len;
-  in >> len;
-  value.resize(0);
-  for (size_t i = 0; i < len; ++i) {
-    T v;
-    in >> v;
-    value.push_back(v);
-  }
-  return in;
-}
-
 void TraceWriter::write_task_event(const TraceTaskEvent& event) {
+  auto& tasks = writer(TASKS);
   tasks << event.type() << event.tid();
   switch (event.type()) {
     case TraceTaskEvent::CLONE:
       tasks << event.parent_tid() << event.clone_flags();
       break;
+    case TraceTaskEvent::FORK:
+      tasks << event.parent_tid();
+      break;
     case TraceTaskEvent::EXEC:
-      tasks << event.file_name() << event.cmd_line();
+      tasks << event.file_name() << event.cmd_line() << event.fds_to_close();
       break;
     case TraceTaskEvent::EXIT:
       break;
@@ -212,14 +216,18 @@ void TraceWriter::write_task_event(const TraceTaskEvent& event) {
 }
 
 TraceTaskEvent TraceReader::read_task_event() {
+  auto& tasks = reader(TASKS);
   TraceTaskEvent r;
   tasks >> r.type_ >> r.tid_;
   switch (r.type()) {
     case TraceTaskEvent::CLONE:
       tasks >> r.parent_tid_ >> r.clone_flags_;
       break;
+    case TraceTaskEvent::FORK:
+      tasks >> r.parent_tid_;
+      break;
     case TraceTaskEvent::EXEC:
-      tasks >> r.file_name_ >> r.cmd_line_;
+      tasks >> r.file_name_ >> r.cmd_line_ >> r.fds_to_close_;
       break;
     case TraceTaskEvent::EXIT:
       break;
@@ -251,9 +259,12 @@ string TraceWriter::try_hardlink_file(const string& file_name) {
 
 TraceWriter::RecordInTrace TraceWriter::write_mapped_region(
     const TraceMappedRegion& map, int prot, int flags) {
+  auto& mmaps = writer(MMAPS);
   TraceReader::MappedDataSource source;
   string backing_file_name;
-  if (map.stat().st_ino == 0) {
+  if (map.type() == TraceMappedRegion::SYSV_SHM) {
+    source = TraceReader::SOURCE_TRACE;
+  } else if (map.stat().st_ino == 0) {
     source = TraceReader::SOURCE_ZERO;
   } else {
     if (should_copy_mmap_region(map.file_name(), &map.stat(), prot, flags)) {
@@ -268,8 +279,8 @@ TraceWriter::RecordInTrace TraceWriter::write_mapped_region(
       backing_file_name = try_hardlink_file(map.file_name());
     }
   }
-  mmaps << source << map.file_name() << map.stat() << map.start() << map.end()
-        << map.offset_pages() << backing_file_name;
+  mmaps << source << map.type() << map.file_name() << map.stat() << map.start()
+        << map.end() << map.offset_pages() << backing_file_name;
   ++mmap_count;
   return source == TraceReader::SOURCE_TRACE ? RECORD_IN_TRACE
                                              : DONT_RECORD_IN_TRACE;
@@ -295,10 +306,11 @@ static void verify_backing_file(const TraceMappedRegion& map,
 }
 
 TraceMappedRegion TraceReader::read_mapped_region(MappedData* data) {
+  auto& mmaps = reader(MMAPS);
   TraceMappedRegion map;
   string backing_file_name;
-  mmaps >> data->source >> map.filename >> map.stat_ >> map.start_ >>
-      map.end_ >> map.file_offset_pages >> backing_file_name;
+  mmaps >> data->source >> map.type_ >> map.filename >> map.stat_ >>
+      map.start_ >> map.end_ >> map.file_offset_pages >> backing_file_name;
   if (data->source == SOURCE_FILE) {
     if (backing_file_name[0] != '/') {
       backing_file_name = dir() + "/" + backing_file_name;
@@ -331,11 +343,15 @@ static istream& operator>>(istream& in, vector<string>& vs) {
 }
 
 void TraceWriter::write_raw(const void* d, size_t len, remote_ptr<void> addr) {
+  auto& data = writer(RAW_DATA);
+  auto& data_header = writer(RAW_DATA_HEADER);
   data_header << global_time << addr.as_int() << len;
   data.write(d, len);
 }
 
 TraceReader::RawData TraceReader::read_raw_data() {
+  auto& data = reader(RAW_DATA);
+  auto& data_header = reader(RAW_DATA_HEADER);
   TraceFrame::Time time;
   RawData d;
   size_t num_bytes;
@@ -347,6 +363,7 @@ TraceReader::RawData TraceReader::read_raw_data() {
 }
 
 bool TraceReader::read_raw_data_for_frame(const TraceFrame& frame, RawData& d) {
+  auto& data_header = reader(RAW_DATA_HEADER);
   while (!data_header.at_end()) {
     TraceFrame::Time time;
     data_header.save_state();
@@ -364,11 +381,9 @@ bool TraceReader::read_raw_data_for_frame(const TraceFrame& frame, RawData& d) {
 }
 
 void TraceWriter::close() {
-  events.close();
-  data.close();
-  data_header.close();
-  mmaps.close();
-  tasks.close();
+  for (auto& w : writers) {
+    w->close();
+  }
 }
 
 static string make_trace_dir(const string& exe_path) {
@@ -393,23 +408,22 @@ static string make_trace_dir(const string& exe_path) {
   return dir;
 }
 
-TraceWriter::TraceWriter(const std::vector<std::string>& argv,
-                         const std::vector<std::string>& envp,
+TraceWriter::TraceWriter(const vector<string>& argv, const vector<string>& envp,
                          const string& cwd, int bind_to_cpu)
     : TraceStream(make_trace_dir(argv[0]),
                   // Somewhat arbitrarily start the
                   // global time from 1.
-                  1),
-      events(events_path(), 1024 * 1024, 1),
-      data(data_path(), 8 * 1024 * 1024, 3),
-      data_header(data_header_path(), 1024 * 1024, 1),
-      mmaps(mmaps_path(), 64 * 1024, 1),
-      tasks(tasks_path(), 64 * 1024, 1) {
+                  1) {
   this->argv = argv;
   this->envp = envp;
   this->cwd = cwd;
   this->bind_to_cpu = bind_to_cpu;
 
+  for (Substream s = SUBSTREAM_FIRST; s < SUBSTREAM_COUNT; ++s) {
+    writers[s] = unique_ptr<CompressedWriter>(new CompressedWriter(
+        path(s), substream(s).block_size, substream(s).threads));
+  }
+
   string ver_path = version_path();
   fstream version(ver_path.c_str(), fstream::out);
   if (!version.good()) {
@@ -444,6 +458,7 @@ TraceWriter::TraceWriter(const std::vector<std::string>& argv,
 }
 
 TraceFrame TraceReader::peek_frame() {
+  auto& events = reader(EVENTS);
   events.save_state();
   auto saved_time = global_time;
   TraceFrame frame;
@@ -455,15 +470,16 @@ TraceFrame TraceReader::peek_frame() {
   return frame;
 }
 
-TraceFrame TraceReader::peek_to(pid_t pid, EventType type,
-                                SyscallEntryOrExit state) {
+TraceFrame TraceReader::peek_to(pid_t pid, EventType type, SyscallState state) {
+  auto& events = reader(EVENTS);
   TraceFrame frame;
   events.save_state();
   auto saved_time = global_time;
   while (good() && !at_end()) {
     frame = read_frame();
-    if (frame.tid() == pid && frame.event().type == type &&
-        frame.event().state == state) {
+    if (frame.tid() == pid && frame.event().type() == type &&
+        (!frame.event().is_syscall_event() ||
+         frame.event().Syscall().state == state)) {
       events.restore_state();
       global_time = saved_time;
       return frame;
@@ -475,11 +491,9 @@ TraceFrame TraceReader::peek_to(pid_t pid, EventType type,
 }
 
 void TraceReader::rewind() {
-  events.rewind();
-  data.rewind();
-  data_header.rewind();
-  mmaps.rewind();
-  tasks.rewind();
+  for (Substream s = SUBSTREAM_FIRST; s < SUBSTREAM_COUNT; ++s) {
+    reader(s).rewind();
+  }
   global_time = 0;
   assert(good());
 }
@@ -490,12 +504,11 @@ TraceReader::TraceReader(const string& dir)
                   // that when we tick it when reading
                   // the first trace, it matches the
                   // initial global time at recording, 1.
-                  0),
-      events(events_path()),
-      data(data_path()),
-      data_header(data_header_path()),
-      mmaps(mmaps_path()),
-      tasks(tasks_path()) {
+                  0) {
+  for (Substream s = SUBSTREAM_FIRST; s < SUBSTREAM_COUNT; ++s) {
+    readers[s] = unique_ptr<CompressedReader>(new CompressedReader(path(s)));
+  }
+
   string path = version_path();
   fstream vfile(path.c_str(), fstream::in);
   if (!vfile.good()) {
@@ -538,14 +551,36 @@ TraceReader::TraceReader(const string& dir)
   in >> bind_to_cpu;
 }
 
+/**
+ * Create a copy of this stream that has exactly the same
+ * state as 'other', but for which mutations of this
+ * clone won't affect the state of 'other' (and vice versa).
+ */
+TraceReader::TraceReader(const TraceReader& other)
+    : TraceStream(other.dir(), other.time()) {
+  for (Substream s = SUBSTREAM_FIRST; s < SUBSTREAM_COUNT; ++s) {
+    readers[s] =
+        unique_ptr<CompressedReader>(new CompressedReader(other.reader(s)));
+  }
+
+  argv = other.argv;
+  envp = other.envp;
+  cwd = other.cwd;
+  bind_to_cpu = other.bind_to_cpu;
+}
+
 uint64_t TraceReader::uncompressed_bytes() const {
-  return events.uncompressed_bytes() + data.uncompressed_bytes() +
-         data_header.uncompressed_bytes() + mmaps.uncompressed_bytes() +
-         tasks.uncompressed_bytes();
+  uint64_t total = 0;
+  for (Substream s = SUBSTREAM_FIRST; s < SUBSTREAM_COUNT; ++s) {
+    total += reader(s).uncompressed_bytes();
+  }
+  return total;
 }
 
 uint64_t TraceReader::compressed_bytes() const {
-  return events.compressed_bytes() + data.compressed_bytes() +
-         data_header.compressed_bytes() + mmaps.compressed_bytes() +
-         tasks.uncompressed_bytes();
+  uint64_t total = 0;
+  for (Substream s = SUBSTREAM_FIRST; s < SUBSTREAM_COUNT; ++s) {
+    total += reader(s).compressed_bytes();
+  }
+  return total;
 }
