@@ -3,17 +3,25 @@
 #ifndef RR_KERNEL_ABI_H
 #define RR_KERNEL_ABI_H
 
+// Include remote_ptr.h first since it (indirectly) requires a definition of
+// ERANGE, which other headers below #undef :-(
+#include "remote_ptr.h"
+
 // Get all the kernel definitions so we can verify our alternative versions.
 #include <arpa/inet.h>
 #include <asm/ldt.h>
 #include <elf.h>
 #include <fcntl.h>
 #include <linux/ethtool.h>
+#include <linux/filter.h>
 #include <linux/ipc.h>
 #include <linux/msg.h>
 #include <linux/net.h>
+#include <linux/sem.h>
+#include <linux/shm.h>
 #include <linux/sockios.h>
 #include <linux/sysctl.h>
+#include <linux/videodev2.h>
 #include <linux/wireless.h>
 #include <poll.h>
 #include <signal.h>
@@ -38,8 +46,6 @@
 
 #include <vector>
 
-#include "remote_ptr.h"
-
 class Task;
 
 enum SupportedArch {
@@ -124,6 +130,8 @@ struct FcntlConstants {
     SETLKW64 = 14,
     SETOWN_EX = 15,
     GETOWN_EX = 16,
+    // Linux-specific operations
+    DUPFD_CLOEXEC = 0x400 + 6,
   };
 };
 
@@ -144,6 +152,7 @@ struct WordSize32Defs : public KernelConstants {
   typedef uint32_t unsigned_word;
 
   typedef uint32_t size_t;
+  typedef int32_t ssize_t;
 
   // These really only exist as proper abstractions so that adding x32
   // (x86-64's ILP32 ABI) support is relatively easy.
@@ -175,6 +184,7 @@ struct WordSize64Defs : public KernelConstants {
   typedef uint64_t unsigned_word;
 
   typedef uint64_t size_t;
+  typedef int64_t ssize_t;
 
   // These really only exist as proper abstractions so that adding x32
   // (x86-64's ILP32 ABI) support is relatively easy.
@@ -189,6 +199,11 @@ struct WordSize64Defs : public KernelConstants {
   typedef Elf64_Sym ElfSym;
 };
 
+/**
+ * Structs defined in BaseArch and its derivatives should not contain any
+ * holes. Holes can cause divergence if such structs are copied from rr to
+ * the tracee.
+ */
 template <SupportedArch arch_, typename wordsize>
 struct BaseArch : public wordsize, public FcntlConstants {
   static SupportedArch arch() { return arch_; }
@@ -203,6 +218,7 @@ struct BaseArch : public wordsize, public FcntlConstants {
   typedef typename wordsize::unsigned_long unsigned_long;
   typedef typename wordsize::unsigned_word unsigned_word;
   typedef typename wordsize::size_t size_t;
+  typedef typename wordsize::ssize_t ssize_t;
   typedef typename wordsize::sigchld_clock_t sigchld_clock_t;
   typedef typename wordsize::__statfs_word __statfs_word;
 
@@ -234,6 +250,7 @@ struct BaseArch : public wordsize, public FcntlConstants {
   typedef int64_t __kernel_loff_t;
 
   template <typename T> struct ptr {
+    typedef T Referent;
     unsigned_word val;
     template <typename U> operator remote_ptr<U>() const { return rptr(); }
     /**
@@ -248,7 +265,7 @@ struct BaseArch : public wordsize, public FcntlConstants {
       return *this;
     }
     operator bool() const { return val; }
-    size_t referent_size() const { return sizeof(T); }
+    static size_t referent_size() { return sizeof(T); }
   };
 
   union sigval_t {
@@ -296,6 +313,7 @@ struct BaseArch : public wordsize, public FcntlConstants {
   struct msghdr {
     ptr<void> msg_name;
     socklen_t msg_namelen;
+    char _padding[sizeof(ptr<void>) - sizeof(socklen_t)];
 
     ptr<iovec> msg_iov;
     size_t msg_iovlen;
@@ -422,6 +440,7 @@ struct BaseArch : public wordsize, public FcntlConstants {
     tcflag_t c_lflag;
     cc_t c_line;
     cc_t c_cc[32];
+    char _padding[3];
     speed_t c_ispeed;
     speed_t c_ospeed;
   };
@@ -442,16 +461,16 @@ struct BaseArch : public wordsize, public FcntlConstants {
     __kernel_uid32_t cuid;
     __kernel_gid32_t cgid;
     __kernel_mode_t mode;
-    unsigned char __pad1[4 - sizeof(__kernel_mode_t)];
     unsigned_short seq;
     unsigned_short __pad2;
+    char __pad3[sizeof(__kernel_ulong_t) - 2 * sizeof(unsigned_short)];
     __kernel_ulong_t unused1;
     __kernel_ulong_t unused2;
   };
   RR_VERIFY_TYPE(ipc64_perm);
 
   struct msqid64_ds {
-    struct ipc64_perm msg_perm;
+    ipc64_perm msg_perm;
     // These msg*time fields are really __kernel_time_t plus
     // appropiate padding.  We don't touch the fields, though.
     //
@@ -483,6 +502,70 @@ struct BaseArch : public wordsize, public FcntlConstants {
   };
   RR_VERIFY_TYPE(msginfo);
 
+  struct shmid64_ds {
+    ipc64_perm shm_perm;
+    size_t shm_segsz;
+    uint64_t shm_atime_only_little_endian;
+    uint64_t shm_dtime_only_little_endian;
+    uint64_t shm_ctime_only_little_endian;
+    __kernel_pid_t shm_cpid;
+    __kernel_pid_t shm_lpid;
+    __kernel_ulong_t shm_nattch;
+    __kernel_ulong_t unused4;
+    __kernel_ulong_t unused5;
+  };
+  RR_VERIFY_TYPE(shmid64_ds);
+
+  struct shminfo64 {
+    __kernel_ulong_t shmmax;
+    __kernel_ulong_t shmmin;
+    __kernel_ulong_t shmmni;
+    __kernel_ulong_t shmseg;
+    __kernel_ulong_t shmall;
+    __kernel_ulong_t unused1;
+    __kernel_ulong_t unused2;
+    __kernel_ulong_t unused3;
+    __kernel_ulong_t unused4;
+  };
+  RR_VERIFY_TYPE(shminfo64);
+
+  struct shm_info {
+    int used_ids;
+    char __pad[sizeof(__kernel_ulong_t) - sizeof(int)];
+    __kernel_ulong_t shm_tot;
+    __kernel_ulong_t shm_rss;
+    __kernel_ulong_t shm_swp;
+    __kernel_ulong_t swap_attempts;
+    __kernel_ulong_t swap_successes;
+  };
+  RR_VERIFY_TYPE(shm_info);
+
+  struct semid64_ds {
+    ipc64_perm sem_perm;
+    __kernel_time_t sem_otime;
+    __kernel_ulong_t __unused1;
+    __kernel_time_t sem_ctime;
+    __kernel_ulong_t __unused2;
+    __kernel_ulong_t sem_nsems;
+    __kernel_ulong_t __unused3;
+    __kernel_ulong_t __unused4;
+  };
+  RR_VERIFY_TYPE(semid64_ds);
+
+  struct seminfo {
+    int semmap;
+    int semmni;
+    int semmns;
+    int semmnu;
+    int semmsl;
+    int semopm;
+    int semume;
+    int semusz;
+    int semvmx;
+    int semaem;
+  };
+  RR_VERIFY_TYPE(seminfo);
+
   // The clone(2) syscall has four (!) different calling conventions,
   // depending on what architecture it's being compiled for.  We describe
   // the orderings for x86oids here.
@@ -591,6 +674,7 @@ struct BaseArch : public wordsize, public FcntlConstants {
 
   struct ifconf {
     signed_int ifc_len;
+    char __pad[sizeof(ptr<void>) - sizeof(int)];
     union {
       ptr<char> ifcu_buf;
       ptr<ifreq> ifcu_req;
@@ -683,6 +767,7 @@ struct BaseArch : public wordsize, public FcntlConstants {
   struct flock {
     signed_short l_type;
     signed_short l_whence;
+    char __pad[sizeof(off_t) - 2 * sizeof(short)];
     off_t l_start;
     off_t l_len;
     pid_t l_pid;
@@ -692,6 +777,8 @@ struct BaseArch : public wordsize, public FcntlConstants {
   struct flock64 {
     signed_short l_type;
     signed_short l_whence;
+    // No padding on 32-bit, 4 bytes of padding on 64-bit
+    char __pad[sizeof(uint32_t) - 2 * sizeof(short)];
     uint64_t l_start;
     uint64_t l_len;
     pid_t l_pid;
@@ -709,6 +796,7 @@ struct BaseArch : public wordsize, public FcntlConstants {
   // them here makes their definitions more concise.
   struct accept_args {
     signed_int sockfd;
+    char __pad[sizeof(ptr<void>) - sizeof(int)];
     ptr<sockaddr> addr;
     ptr<socklen_t> addrlen;
   };
@@ -719,6 +807,7 @@ struct BaseArch : public wordsize, public FcntlConstants {
 
   struct getsockname_args {
     signed_int sockfd;
+    char __pad[sizeof(ptr<void>) - sizeof(int)];
     ptr<sockaddr> addr;
     ptr<socklen_t> addrlen;
   };
@@ -727,12 +816,14 @@ struct BaseArch : public wordsize, public FcntlConstants {
     signed_int sockfd;
     signed_int level;
     signed_int optname;
+    char __pad[sizeof(ptr<void>) - sizeof(int)];
     ptr<void> optval;
     ptr<socklen_t> optlen;
   };
 
   struct recv_args {
     signed_int sockfd;
+    char __pad[sizeof(ptr<void>) - sizeof(int)];
     ptr<void> buf;
     size_t len;
     signed_int flags;
@@ -749,20 +840,30 @@ struct BaseArch : public wordsize, public FcntlConstants {
 
   struct recvmsg_args {
     signed_int fd;
+    char __pad[sizeof(ptr<void>) - sizeof(int)];
     ptr<msghdr> msg;
     signed_int flags;
   };
 
   struct recvmmsg_args {
     signed_int sockfd;
+    char __pad[sizeof(ptr<void>) - sizeof(int)];
     ptr<mmsghdr> msgvec;
     unsigned_int vlen;
     unsigned_int flags;
     ptr<timespec> timeout;
   };
 
+  struct sendmsg_args {
+    signed_int fd;
+    char __pad[sizeof(ptr<void>) - sizeof(int)];
+    ptr<msghdr> msg;
+    signed_int flags;
+  };
+
   struct sendmmsg_args {
     signed_int sockfd;
+    char __pad[sizeof(ptr<void>) - sizeof(int)];
     ptr<mmsghdr> msgvec;
     unsigned_int vlen;
     unsigned_int flags;
@@ -772,6 +873,7 @@ struct BaseArch : public wordsize, public FcntlConstants {
     signed_int domain;
     signed_int type;
     signed_int protocol;
+    char __pad[sizeof(ptr<void>) - sizeof(int)];
     ptr<signed_int> sv; // int sv[2]
   };
 
@@ -790,6 +892,7 @@ struct BaseArch : public wordsize, public FcntlConstants {
     signed_int prot;
     signed_int flags;
     signed_int fd;
+    char __pad[sizeof(off_t) - sizeof(int)];
     off_t offset;
   };
 
@@ -813,6 +916,7 @@ struct BaseArch : public wordsize, public FcntlConstants {
 
   struct select_args {
     signed_int n_fds;
+    char __pad[sizeof(ptr<void>) - sizeof(int)];
     ptr<fd_set> read_fds;
     ptr<fd_set> write_fds;
     ptr<fd_set> except_fds;
@@ -831,6 +935,7 @@ struct BaseArch : public wordsize, public FcntlConstants {
   struct __sysctl_args {
     ptr<signed_int> name;
     signed_int nlen;
+    char __pad[sizeof(ptr<void>) - sizeof(int)];
     ptr<void> oldval;
     ptr<size_t> oldlenp;
     ptr<void> newval;
@@ -839,19 +944,6 @@ struct BaseArch : public wordsize, public FcntlConstants {
   };
   RR_VERIFY_TYPE(__sysctl_args);
 
-  // The getgroups syscall (as well as several others) differs between
-  // architectures depending on whether they ever supported 16-bit
-  // {U,G}IDs or not.  Architectures such as x86, which did support
-  // 16-bit {U,G}IDs, have a getgroups syscall for the 16-bit GID case
-  // and a getgroups32 syscall for the 32-bit GID case.  Architectures
-  // such as as x86-64, which support 32-bit GIDs exclusively, have only
-  // a getgroups syscall.  We need to know which one we're dealing with
-  // when recording and replaying getgroups and related syscalls.
-  enum UserAndGroupIDSizesSupported {
-    Mixed16And32Bit,
-    Only32Bit,
-  };
-
   typedef struct {
     unsigned_long __val[1024 / (8 * sizeof(unsigned_long))];
   } __sigset_t;
@@ -941,14 +1033,15 @@ struct BaseArch : public wordsize, public FcntlConstants {
   RR_VERIFY_TYPE(itimerval);
 
   struct itimerspec {
-    struct timespec it_interval;
-    struct timespec it_value;
+    timespec it_interval;
+    timespec it_value;
   };
   RR_VERIFY_TYPE(itimerspec);
 
   typedef struct sigaltstack {
     ptr<void> ss_sp;
     int ss_flags;
+    char __pad[sizeof(size_t) - sizeof(int)];
     size_t ss_size;
   } stack_t;
   RR_VERIFY_TYPE(stack_t);
@@ -964,6 +1057,7 @@ struct BaseArch : public wordsize, public FcntlConstants {
     __kernel_ulong_t freeswap;
     uint16_t procs;
     uint16_t pad;
+    char __pad[sizeof(__kernel_ulong_t) - 2 * sizeof(uint16_t)];
     __kernel_ulong_t totalhigh;
     __kernel_ulong_t freehigh;
     uint32_t mem_unit;
@@ -1002,6 +1096,55 @@ struct BaseArch : public wordsize, public FcntlConstants {
   static size_t cmsg_len(size_t len) {
     return cmsg_align(sizeof(cmsghdr)) + len;
   }
+
+  struct v4l2_timecode {
+    uint32_t type;
+    uint32_t flags;
+    uint8_t frames;
+    uint8_t seconds;
+    uint8_t minutes;
+    uint8_t hours;
+    uint8_t userbits[4];
+  };
+  RR_VERIFY_TYPE(v4l2_timecode);
+
+  struct v4l2_buffer {
+    uint32_t index;
+    uint32_t type;
+    uint32_t bytesused;
+    uint32_t flags;
+    uint32_t field;
+    char __pad[sizeof(__kernel_ulong_t) - sizeof(uint32_t)];
+    struct timeval timestamp;
+    struct v4l2_timecode timecode;
+    uint32_t sequence;
+    uint32_t memory;
+    union {
+      uint32_t offset;
+      unsigned_long userptr;
+      ptr<void> planes;
+      int32_t fd;
+    } m;
+    uint32_t length;
+    uint32_t reserved2;
+    uint32_t reserved;
+  };
+  RR_VERIFY_TYPE(v4l2_buffer);
+
+  struct sock_filter {
+    uint16_t code;
+    uint8_t jt;
+    uint8_t jf;
+    uint32_t k;
+  };
+  RR_VERIFY_TYPE(sock_filter);
+
+  struct sock_fprog {
+    uint16_t len;
+    char _padding[sizeof(ptr<void>) - sizeof(uint16_t)];
+    ptr<sock_filter> filter;
+  };
+  RR_VERIFY_TYPE(sock_fprog);
 };
 
 struct X86Arch : public BaseArch<SupportedArch::x86, WordSize32Defs> {
@@ -1013,7 +1156,17 @@ struct X86Arch : public BaseArch<SupportedArch::x86, WordSize32Defs> {
   static const CloneParameterOrdering clone_parameter_ordering =
       FlagsStackParentTLSChild;
   static const SelectCallingSemantics select_semantics = SelectStructArguments;
-  static const UserAndGroupIDSizesSupported uid_gid_sizes = Mixed16And32Bit;
+
+  // The getgroups syscall (as well as several others) differs between
+  // architectures depending on whether they ever supported 16-bit
+  // {U,G}IDs or not.  Architectures such as x86, which did support
+  // 16-bit {U,G}IDs, have a getgroups syscall for the 16-bit GID case
+  // and a getgroups32 syscall for the 32-bit GID case.  Architectures
+  // such as as x86-64, which support 32-bit GIDs exclusively, have only
+  // a getgroups syscall.  We need to know which one we're dealing with
+  // when recording and replaying getgroups and related syscalls.
+  typedef uint16_t legacy_uid_t;
+  typedef uint16_t legacy_gid_t;
 
 #include "SyscallEnumsX86.generated"
 
@@ -1071,6 +1224,25 @@ struct X86Arch : public BaseArch<SupportedArch::x86, WordSize32Defs> {
                       user_fpxregs_struct);
 #endif
 
+  struct user {
+    user_regs_struct regs;
+    int u_fpvalid;
+    user_fpregs_struct i387;
+    uint32_t u_tsize;
+    uint32_t u_dsize;
+    uint32_t u_ssize;
+    uint32_t start_code;
+    uint32_t start_stack;
+    int32_t signal;
+    int reserved;
+    ptr<user_regs_struct> u_ar0;
+    ptr<user_fpregs_struct> u_fpstate;
+    uint32_t magic;
+    char u_comm[32];
+    int u_debugreg[8];
+  };
+  RR_VERIFY_TYPE_ARCH(SupportedArch::x86, ::user, user);
+
   struct stat {
     dev_t st_dev;
     unsigned_short __pad1;
@@ -1084,9 +1256,9 @@ struct X86Arch : public BaseArch<SupportedArch::x86, WordSize32Defs> {
     off_t st_size;
     blksize_t st_blksize;
     blkcnt_t st_blocks;
-    struct timespec st_atim;
-    struct timespec st_mtim;
-    struct timespec st_ctim;
+    timespec st_atim;
+    timespec st_mtim;
+    timespec st_ctim;
     unsigned_long __unused4;
     unsigned_long __unused5;
   };
@@ -1105,9 +1277,9 @@ struct X86Arch : public BaseArch<SupportedArch::x86, WordSize32Defs> {
     off64_t st_size;
     blksize_t st_blksize;
     blkcnt64_t st_blocks;
-    struct timespec st_atim;
-    struct timespec st_mtim;
-    struct timespec st_ctim;
+    timespec st_atim;
+    timespec st_mtim;
+    timespec st_ctim;
     ino64_t st_ino;
   };
   RR_VERIFY_TYPE_ARCH(SupportedArch::x86, struct ::stat64, struct stat64);
@@ -1123,7 +1295,9 @@ struct X64Arch : public BaseArch<SupportedArch::x86_64, WordSize64Defs> {
       FlagsStackParentChildTLS;
   static const SelectCallingSemantics select_semantics =
       SelectRegisterArguments;
-  static const UserAndGroupIDSizesSupported uid_gid_sizes = Only32Bit;
+
+  typedef uint32_t legacy_uid_t;
+  typedef uint32_t legacy_gid_t;
 
 #include "SyscallEnumsX64.generated"
 
@@ -1190,6 +1364,31 @@ struct X64Arch : public BaseArch<SupportedArch::x86_64, WordSize64Defs> {
   RR_VERIFY_TYPE_ARCH(SupportedArch::x86_64, ::user_fpregs_struct,
                       user_fpregs_struct);
 
+  struct user {
+    struct user_regs_struct regs;
+    int u_fpvalid;
+    struct user_fpregs_struct i387;
+    uint64_t u_tsize;
+    uint64_t u_dsize;
+    uint64_t u_ssize;
+    uint64_t start_code;
+    uint64_t start_stack;
+    int64_t signal;
+    int reserved;
+    union {
+      struct user_regs_struct* u_ar0;
+      uint64_t __u_ar0_word;
+    };
+    union {
+      struct user_fpregs_struct* u_fpstate;
+      uint64_t __u_fpstate_word;
+    };
+    uint64_t magic;
+    char u_comm[32];
+    uint64_t u_debugreg[8];
+  };
+  RR_VERIFY_TYPE_ARCH(SupportedArch::x86_64, ::user, user);
+
   struct stat {
     dev_t st_dev;
     ino_t st_ino;
@@ -1239,12 +1438,6 @@ struct X64Arch : public BaseArch<SupportedArch::x86_64, WordSize64Defs> {
       return f<rr::X64Arch>(args);                                             \
   }
 
-// Helper structure for determining the proper outparam sizes for syscalls
-// like getgroups.
-template <typename Arch> struct LegacyUIDSyscall {
-  static const size_t size = (Arch::uid_gid_sizes == Arch::Only32Bit ? 4 : 2);
-};
-
 #include "SyscallHelperFunctions.generated"
 
 /**
@@ -1262,7 +1455,7 @@ std::vector<uint8_t> syscall_instruction(SupportedArch arch);
  * Return the length of all invoke-syscall instructions. Currently,
  * they must all have the same length!
  */
-size_t syscall_instruction_length(SupportedArch arch);
+ssize_t syscall_instruction_length(SupportedArch arch);
 
 #if defined(__i386__)
 typedef X86Arch NativeArch;
