@@ -133,6 +133,11 @@ bool is_always_emulated_syscall(int syscall, SupportedArch arch) {
   RR_ARCH_FUNCTION(is_always_emulated_syscall_arch, arch, syscall);
 }
 
+bool is_sigreturn(int syscallno, SupportedArch arch) {
+  return is_sigreturn_syscall(syscallno, arch) ||
+         is_rt_sigreturn_syscall(syscallno, arch);
+}
+
 const char* errno_name(int err) {
   switch (err) {
     case 0:
@@ -317,3 +322,8 @@ std::ostream& operator<<(std::ostream& stream, const siginfo_t& siginfo) {
   stream << "}";
   return stream;
 }
+
+int shm_flags_to_mmap_prot(int flags) {
+  return PROT_READ | ((flags & SHM_RDONLY) ? 0 : PROT_WRITE) |
+         ((flags & SHM_EXEC) ? PROT_EXEC : 0);
+}
