@@ -41,6 +41,11 @@ const char* signal_name(int sig);
  */
 bool is_always_emulated_syscall(int syscall, SupportedArch arch);
 
+/**
+ * Return true if this is some kind of sigreturn syscall.
+ */
+bool is_sigreturn(int syscall, SupportedArch arch);
+
 /**
  * Return the symbolic error name (e.g. "EINVAL") for errno.
  */
@@ -56,4 +61,6 @@ const char* sicode_name(int code, int sig);
  */
 std::ostream& operator<<(std::ostream& stream, const siginfo_t& siginfo);
 
+int shm_flags_to_mmap_prot(int flags);
+
 #endif
