@@ -27,6 +27,13 @@
 #define PTRACE_O_EXITKILL (1 << 20)
 #endif
 
+#ifndef SECCOMP_SET_MODE_STRICT
+#define SECCOMP_SET_MODE_STRICT 0
+#endif
+#ifndef SECCOMP_SET_MODE_FILTER
+#define SECCOMP_SET_MODE_FILTER 1
+#endif
+
 // These are defined by the include/linux/errno.h in the kernel tree.
 // Since userspace doesn't see these errnos in normal operation, that
 // header apparently isn't distributed with libc.
@@ -35,6 +42,13 @@
 #define ERESTARTNOHAND 514
 #define ERESTART_RESTARTBLOCK 516
 
+// These definitions haven't made it out to current libc-dev packages
+// yet.
+#ifndef GRND_NONBLOCK
+#define GRND_NONBLOCK 0x0001
+#define GRND_RANDOM 0x0002
+#endif
+
 /* We need to complement sigsets in order to update the Task blocked
  * set, but POSIX doesn't appear to define a convenient helper.  So we
  * define our own linux-compatible sig_set_t and use bit operators to
