@@ -16,6 +16,9 @@
 
 using namespace std;
 
+// Show version and quit.
+static bool show_version = false;
+
 void assert_prerequisites(bool use_syscall_buffer) {
   struct utsname uname_buf;
   memset(&uname_buf, 0, sizeof(uname_buf));
@@ -88,9 +91,12 @@ void check_performance_settings() {
   }
 }
 
-int print_usage(void) {
-  fputs("Usage:\n", stderr);
-  Command::print_help_all(stderr);
+void print_version(FILE* out) { fprintf(out, "rr version %s\n", RR_VERSION); }
+
+void print_usage(FILE* out) {
+  print_version(out);
+  fputs("Usage:\n", out);
+  Command::print_help_all(out);
   fputs(
       "\n"
       "Common options:\n"
@@ -129,6 +135,7 @@ int print_usage(void) {
       "                             where EVENT-NO is the global trace time "
       "at\n"
       "                             which the write occures.\n"
+      "  -N, --version              print the version number and exit\n"
       "  -S, --suppress-environment-warnings\n"
       "                             suppress warnings about issues in the\n"
       "                             environment that rr has no control over\n"
@@ -137,8 +144,7 @@ int print_usage(void) {
       "                             critical to the user\n"
       "  -W, --wait-secs=<NUM_SECS> wait NUM_SECS seconds just after startup,\n"
       "                             before initiating recording or replaying\n",
-      stderr);
-  return 1;
+      out);
 }
 
 static void init_random() {
@@ -154,11 +160,12 @@ bool parse_global_option(std::vector<std::string>& args) {
     { 'T', "dump-at", HAS_PARAMETER },
     { 'D', "dump-on", HAS_PARAMETER },
     { 'F', "force-things", NO_PARAMETER },
-    { 'A', "force-microarch", HAS_PARAMETER },
+    { 'A', "microarch", HAS_PARAMETER },
     { 'M', "mark-stdio", NO_PARAMETER },
     { 'S', "suppress-environment-warnings", NO_PARAMETER },
     { 'E', "fatal-errors", NO_PARAMETER },
-    { 'V', "verbose", NO_PARAMETER }
+    { 'V', "verbose", NO_PARAMETER },
+    { 'N', "version", NO_PARAMETER }
   };
 
   ParsedOption opt;
@@ -207,6 +214,9 @@ bool parse_global_option(std::vector<std::string>& args) {
     case 'V':
       flags.verbose = true;
       break;
+    case 'N':
+      show_version = true;
+      break;
     default:
       assert(0 && "Invalid flag");
   }
@@ -224,8 +234,14 @@ int main(int argc, char* argv[]) {
   while (parse_global_option(args)) {
   }
 
+  if (show_version) {
+    print_version(stdout);
+    return 0;
+  }
+
   if (args.size() == 0) {
-    return print_usage();
+    print_usage(stderr);
+    return 1;
   }
 
   auto command = Command::command_for_name(args[0]);
@@ -233,7 +249,8 @@ int main(int argc, char* argv[]) {
     args.erase(args.begin());
   } else {
     if (!Command::verify_not_option(args)) {
-      return print_usage();
+      print_usage(stderr);
+      return 1;
     }
     command = RecordCommand::get();
   }
