@@ -74,7 +74,6 @@
 #include <unistd.h>
 
 /* NB: don't include any other local headers here. */
-#include "rr/rr.h"
 
 #ifdef memcpy
 #undef memcpy
@@ -111,6 +110,15 @@ static int buffer_enabled;
 /* Nonzero after process-global state has been initialized. */
 static int process_inited;
 
+/**
+ * If syscallbuf_fds_disabled[fd] is nonzero, then operations on that fd
+ * must be performed through traced syscalls, not the syscallbuf.
+ * The rr supervisor modifies this array directly to dynamically turn
+ * syscallbuf on and off for particular fds. fds outside the array range must
+ * never use the syscallbuf.
+ */
+static volatile char syscallbuf_fds_disabled[SYSCALLBUF_FDS_DISABLED_SIZE];
+
 /**
  * Because this library is always loaded via LD_PRELOAD, we can use the
  * initial-exec TLS model (see http://www.akkadia.org/drepper/tls.pdf) which
@@ -581,6 +589,8 @@ static void __attribute__((constructor)) init_process(void) {
   pthread_atfork(NULL, NULL, post_fork_child);
 
   params.syscallbuf_enabled = buffer_enabled;
+  params.syscallbuf_fds_disabled =
+      buffer_enabled ? syscallbuf_fds_disabled : NULL;
   params.syscall_hook_trampoline = (void*)_syscall_hook_trampoline;
 #if defined(__i386__)
   extern RR_HIDDEN void _syscall_hook_trampoline_3d_01_f0_ff_ff(void);
@@ -806,6 +816,19 @@ static void* prep_syscall(void) {
   return buffer_last() + sizeof(struct syscallbuf_record);
 }
 
+/**
+ * Like prep_syscall, but preps a syscall to operate on a particular fd. If
+ * syscallbuf is disabled for this fd, returns NULL (in which case
+ * start_commit_syscall will abort cleanly and a traced syscall will be used).
+ */
+static void* prep_syscall_for_fd(int fd) {
+  if (fd < 0 || fd >= SYSCALLBUF_FDS_DISABLED_SIZE ||
+      syscallbuf_fds_disabled[fd]) {
+    return NULL;
+  }
+  return prep_syscall();
+}
+
 static void arm_desched_event(void) {
   /* Don't trace the ioctl; doing so would trigger a flushing
    * ptrace trap, which is exactly what this code is trying to
@@ -1014,7 +1037,7 @@ static long sys_close(const struct syscall_info* call) {
   const int syscallno = SYS_close;
   int fd = call->args[0];
 
-  void* ptr = prep_syscall();
+  void* ptr = prep_syscall_for_fd(fd);
   long ret;
 
   if (!start_commit_buffered_syscall(syscallno, ptr, WONT_BLOCK)) {
@@ -1048,7 +1071,7 @@ static int sys_fcntl64_no_outparams(const struct syscall_info* call) {
 
   /* None of the no-outparam fcntl's are known to be
    * may-block. */
-  void* ptr = prep_syscall();
+  void* ptr = prep_syscall_for_fd(fd);
   long ret;
 
   assert(syscallno == call->no);
@@ -1067,7 +1090,7 @@ static int sys_fcntl64_own_ex(const struct syscall_info* call) {
   struct f_owner_ex* owner = (struct f_owner_ex*)call->args[2];
 
   /* The OWN_EX fcntl's aren't may-block. */
-  void* ptr = prep_syscall();
+  void* ptr = prep_syscall_for_fd(fd);
   struct f_owner_ex* owner2 = NULL;
   long ret;
 
@@ -1096,7 +1119,7 @@ static int sys_fcntl64_xlk64(const struct syscall_info* call) {
   int cmd = call->args[1];
   struct flock64* lock = (struct flock64*)call->args[2];
 
-  void* ptr = prep_syscall();
+  void* ptr = prep_syscall_for_fd(fd);
   struct flock64* lock2 = NULL;
   long ret;
 
@@ -1291,7 +1314,7 @@ static long sys__llseek(const struct syscall_info* call) {
   loff_t* result = (loff_t*)call->args[3];
   unsigned int whence = call->args[4];
 
-  void* ptr = prep_syscall();
+  void* ptr = prep_syscall_for_fd(fd);
   loff_t* result2 = NULL;
   long ret;
 
@@ -1322,7 +1345,7 @@ static long sys_lseek(const struct syscall_info* call) {
   off_t off = call->args[1];
   int whence = call->args[2];
 
-  void* ptr = prep_syscall();
+  void* ptr = prep_syscall_for_fd(fd);
   off_t ret = 0;
 
   assert(syscallno == call->no);
@@ -1434,7 +1457,7 @@ static long sys_read(const struct syscall_info* call) {
   void* buf = (void*)call->args[1];
   size_t count = call->args[2];
 
-  void* ptr = prep_syscall();
+  void* ptr = prep_syscall_for_fd(fd);
   void* buf2 = NULL;
   long ret;
 
@@ -1496,7 +1519,7 @@ static long sys_recv(const struct syscall_info* call) {
   size_t len = args[2];
   unsigned int flags = args[3];
 
-  void* ptr = prep_syscall();
+  void* ptr = prep_syscall_for_fd(sockfd);
   void* buf2 = NULL;
   long ret;
 
@@ -1595,17 +1618,12 @@ static long sys_write(const struct syscall_info* call) {
   const void* buf = (const void*)call->args[1];
   size_t count = call->args[2];
 
-  void* ptr = prep_syscall();
+  void* ptr = prep_syscall_for_fd(fd);
   long ret;
 
   assert(syscallno == call->no);
 
-  /* We always have to trap for writes to the magic save-data
-   * fd, because the rr tracer processes them specially.
-   *
-   * TODO: buffer them normally here. */
-  if (RR_MAGIC_SAVE_DATA_FD == fd ||
-      !start_commit_buffered_syscall(syscallno, ptr, MAY_BLOCK)) {
+  if (!start_commit_buffered_syscall(syscallno, ptr, MAY_BLOCK)) {
     return traced_raw_syscall(call);
   }
 
@@ -1620,7 +1638,7 @@ static long sys_writev(const struct syscall_info* call) {
   const struct iovec* iov = (const struct iovec*)call->args[1];
   unsigned long iovcnt = call->args[2];
 
-  void* ptr = prep_syscall();
+  void* ptr = prep_syscall_for_fd(fd);
   long ret;
 
   assert(syscallno == call->no);
@@ -1634,10 +1652,7 @@ static long sys_writev(const struct syscall_info* call) {
   return commit_raw_syscall(syscallno, ptr, ret);
 }
 
-/* Explicitly declare this as hidden so we can call it from
- * _syscall_hook_trampoline without doing all sorts of special PIC handling.
- */
-RR_HIDDEN long syscall_hook(const struct syscall_info* call) {
+static long syscall_hook_internal(const struct syscall_info* call) {
   switch (call->no) {
 #define CASE(syscallname)                                                      \
   case SYS_##syscallname:                                                      \
@@ -1691,6 +1706,46 @@ RR_HIDDEN long syscall_hook(const struct syscall_info* call) {
   }
 }
 
+/* Explicitly declare this as hidden so we can call it from
+ * _syscall_hook_trampoline without doing all sorts of special PIC handling.
+ */
+RR_HIDDEN long syscall_hook(const struct syscall_info* call) {
+  long result = syscall_hook_internal(call);
+  if (buffer_hdr() && buffer_hdr()->notify_on_syscall_hook_exit) {
+    // This syscall will clear notify_on_syscall_hook_exit. Clearing it
+    // ourselves is tricky to get right without races.
+    //
+    // During recording, this flag is set when the recorder needs to delay
+    // delivery of a signal until we've stopped using the syscallbuf.
+    // During replay, this flag is set when the next event is entering a
+    // SYS_rrcall_notify_syscall_hook_exit.
+    //
+    // The correctness argument is as follows:
+    // Correctness requires that a) replay's setting of the flag happens before
+    // we read the flag in the call to syscall_hook that triggered the
+    // SYS_rrcall_notify_syscall_hook_exit and b) replay's setting of the flag
+    // must happen after we read the flag in the previous execution of
+    // syscall_hook.
+    // Condition a) holds as long as no events are recorded between the
+    // checking of the flag above and the execution of this syscall. This
+    // should be the case; no synchronous signals or syscalls are
+    // triggerable, all async signals other than SYSCALLBUF_DESCHED_SIGNAL
+    // are delayed, and SYSCALLBUF_DESCHED_SIGNAL shouldn't fire since we've
+    // disarmed the desched fd at this point. SYSCALLBUF_FLUSH events may be
+    // emitted when we process the SYS_rrcall_notify_syscall_hook_exit event,
+    // but replay of those events ends at the last flushed syscall, before
+    // we exit syscall_hook_internal.
+    // Condition b) failing would mean no new events were generated between
+    // testing the flag in the previous syscall_hook and the execution of this
+    // SYS_rrcall_notify_syscall_hook_exit. However, every invocation of
+    // syscall_hook_internal generates either a traced syscall or a syscallbuf
+    // record that would be flushed by SYSCALLBUF_FLUSH, so that can't
+    // happen.
+    traced_syscall0(SYS_rrcall_notify_syscall_hook_exit);
+  }
+  return result;
+}
+
 /**
  * Exported glibc synonym for |sysconf()|.  We can't use |dlsym()| to
  * resolve the next "sysconf" symbol, because
