@@ -35,9 +35,14 @@
 /* Set this env var to enable syscall buffering. */
 #define SYSCALLBUF_ENABLED_ENV_VAR "_RR_USE_SYSCALLBUF"
 
+/* Size of table mapping fd numbers to syscallbuf-disabled flag.
+ * Most Linux kernels limit fds to 1024 so it probably doesn't make sense
+ * to raise this value... */
+#define SYSCALLBUF_FDS_DISABLED_SIZE 1024
+
 #define RR_PAGE_ADDR 0x70000000
 #define RR_PAGE_IN_UNTRACED_SYSCALL_ADDR (RR_PAGE_ADDR + 4)
-#define RR_PAGE_IN_TRACED_SYSCALL_ADDR (RR_PAGE_ADDR + 16)
+#define RR_PAGE_IN_TRACED_SYSCALL_ADDR (RR_PAGE_ADDR + 24)
 
 /* "Magic" (rr-implemented) syscalls that we use to initialize the
  * syscallbuf.
@@ -51,12 +56,17 @@
  * The preload library calls SYS_rrcall_init_preload during its
  * initialization.
  */
-#define SYS_rrcall_init_preload 443
+#define SYS_rrcall_init_preload 442
 /**
  * The preload library calls SYS_rrcall_init_buffers in each thread that
  * gets created (including the initial main thread).
  */
-#define SYS_rrcall_init_buffers 442
+#define SYS_rrcall_init_buffers 443
+/**
+ * The preload library calls SYS_rrcall_notify_syscall_hook_exit when
+ * unlocking the syscallbuf and notify_after_syscall_hook_exit has been set.
+ */
+#define SYS_rrcall_notify_syscall_hook_exit 444
 
 /* Define macros that let us compile a struct definition either "natively"
  * (when included by preload.c) or as a template over Arch for use by rr.
@@ -97,9 +107,11 @@ struct rrcall_init_preload_params {
    * We let the syscallbuf code decide in order to more simply
    * replay the same decision that was recorded. */
   int syscallbuf_enabled;
-  PTR(void) syscall_hook_trampoline;
   int syscall_patch_hook_count;
   PTR(struct syscall_patch_hook) syscall_patch_hooks;
+  PTR(void) syscall_hook_trampoline;
+  /* Array of size SYSCALLBUF_FDS_DISABLED_SIZE */
+  PTR(volatile char) syscallbuf_fds_disabled;
 };
 
 /**
@@ -111,6 +123,10 @@ TEMPLATE_ARCH
 struct rrcall_init_buffers_params {
   /* The fd we're using to track desched events. */
   int desched_counter_fd;
+  /* padding for 64-bit archs. Structs written to tracee memory must not have
+   * holes!
+   */
+  int padding;
 
   /* "Out" params. */
   /* Returned pointer to and size of the shared syscallbuf
@@ -163,8 +179,13 @@ struct syscallbuf_hdr {
   uint32_t num_rec_bytes;
   /* True if the current syscall should not be committed to the
    * buffer, for whatever reason; likely interrupted by
-   * desched. */
+   * desched. Set by rr. */
   uint8_t abort_commit;
+  /* True if, next time we exit the syscall buffer hook, libpreload should
+   * execute SYS_rrcall_notify_syscall_hook_exit to give rr the opportunity to
+   * deliver a signal.
+   */
+  uint8_t notify_on_syscall_hook_exit;
   /* This tracks whether the buffer is currently in use for a
    * system call. This is helpful when a signal handler runs
    * during a wrapped system call; we don't want it to use the
@@ -174,7 +195,6 @@ struct syscallbuf_hdr {
    * When it's zero, the desched signal can safely be
    * discarded. */
   uint8_t desched_signal_may_be_relevant;
-  uint8_t _padding;
 
   struct syscallbuf_record recs[0];
 } __attribute__((__packed__));
