@@ -56,6 +56,15 @@ public:
     ptr += arith_size();
     return p;
   }
+  remote_ptr<T>& operator--() {
+    ptr -= arith_size();
+    return *this;
+  }
+  remote_ptr<T> operator--(int) {
+    uintptr_t p = ptr;
+    ptr -= arith_size();
+    return p;
+  }
 
   template <typename U> remote_ptr<U> cast() const {
     return remote_ptr<U>(ptr);
@@ -85,6 +94,10 @@ private:
   uintptr_t ptr;
 };
 
+/**
+ * returns a remote_ptr pointing to field f of the struct pointed to by
+ * remote_ptr p
+ */
 #define REMOTE_PTR_FIELD(p, f) (p).field((p).dummy()->f)
 
 template <typename T>
