@@ -1,13 +1,3 @@
-class NullTerminatedString(object):
-    """A class for indicating an argument is a null-terminated string."""
-    def __init__(self):
-        pass
-
-class DynamicSize(object):
-    """A class for indicating the size of a syscall argument."""
-    def __init__(self, size_expr):
-        self.size_expr = size_expr
-
 class BaseSyscall(object):
     """A base class for syscalls.
 
@@ -27,10 +17,9 @@ class ReplaySemantics(object):
 
     EMU = "EMU"                 # Syscall is fully emulated.
     EXEC = "EXEC"               # Syscall is fully executed.
-    EXEC_RET_EMU = "EXEC_RET_EMU" # Syscall is executed, but retval is emulated.
-    MAY_EXEC = "MAY_EXEC"         # Syscall may be fully executed
+    MAY_EXEC = "MAY_EXEC"       # Syscall may be fully executed
 
-    ALL_SEMANTICS = ["EMU", "EXEC", "EXEC_RET_EMU", "MAY_EXEC"]
+    ALL_SEMANTICS = ["EMU", "EXEC", "MAY_EXEC"]
 
     def __init__(self, semantics):
         assert semantics in self.ALL_SEMANTICS
@@ -53,33 +42,36 @@ class UnsupportedSyscall(BaseSyscall, ReplaySemantics):
         BaseSyscall.__init__(self, x86=x86, x64=x64)
         ReplaySemantics.__init__(self, ReplaySemantics.EXEC)
 
+class InvalidSyscall(UnsupportedSyscall):
+    """A syscall that is unsupported by rr and unimplemented by Linux.
+
+    We distinguish syscalls unimplemented by any version of Linux supported
+    by rr from other UnsupportedSyscalls, to help us track the completeness
+    of rr's syscall support.
+    """
+    def __init__(self, x86=None, x64=None):
+        UnsupportedSyscall.__init__(self, x86=x86, x64=x64)
+
 class RegularSyscall(BaseSyscall, ReplaySemantics):
     """A syscall for which replay information may be recorded automatically.
 
     The arguments required for rr to record may be specified directly
     through the arg1...arg6 keyword arguments.  The values for these
     arguments determine the size of the associated arguments to the syscall.
-    The allowed types for a given argument are:
-
-    * A Python string, in which case the size of the argument is sizeof(arg);
-    * A DynamicSize object, in which case the size of the argument is the
-      C expression stored in the DynamicSize object;
-    * A NullTerminatedString object, in which case the size is determined at
-      runtime in the expected fashion.
+    The only allowed type for a given argument is a Python string, in which
+    case the size of the argument is sizeof(arg).
 
     To ensure correct handling for mixed-arch process groups (e.g. a mix of 32
     and 64-bit processes), types should be specified using Arch instead of
     referring directly to the host system types.
     """
-    ARGUMENT_SLOTS = ['arg1', 'arg2', 'arg3', 'arg4', 'arg5', 'arg6']
-
     def __init__(self, semantics=None, **kwargs):
         assert semantics in [ReplaySemantics.EMU,
-                             ReplaySemantics.EXEC,
-                             ReplaySemantics.EXEC_RET_EMU]
+                             ReplaySemantics.EXEC]
         ReplaySemantics.__init__(self, semantics)
         BaseSyscall.__init__(self, **kwargs)
-        for arg in self.ARGUMENT_SLOTS:
+        for a in range(1,6):
+            arg = 'arg' + str(a)
             if arg in kwargs:
                 self.__setattr__(arg, kwargs[arg])
 
@@ -103,16 +95,6 @@ class ExecutedSyscall(RegularSyscall):
     def __init__(self, **kwargs):
         RegularSyscall.__init__(self, semantics=ReplaySemantics.EXEC, **kwargs)
 
-class ExecRetEmuSyscall(RegularSyscall):
-    """A wrapper for regular syscalls.
-
-    This class should be used in preference to manually passing the semantics
-    keyword argument.  Its constructor ensures the correct value is passed for
-    the semantics argument.
-    """
-    def __init__(self, **kwargs):
-        RegularSyscall.__init__(self, semantics=ReplaySemantics.EXEC_RET_EMU, **kwargs)
-
 class IrregularSyscall(BaseSyscall, ReplaySemantics):
     """A base class for irregular syscalls.  Not to be manually instantiated."""
     def __init__(self, semantics=None, **kwargs):
@@ -145,10 +127,11 @@ class IrregularMayExecSyscall(IrregularSyscall):
 #
 # The exit() function causes normal process termination and the value
 # of status & 0377 is returned to the parent (see wait(2)).
-exit = IrregularMayExecSyscall(x86=1, x64=60)
+exit = IrregularEmulatedSyscall(x86=1, x64=60)
 
 # Obsolete, glibc calls clone() instead.
-fork = UnsupportedSyscall(x86=2, x64=57)
+# But Google Breakpad uses it!
+fork = EmulatedSyscall(x86=2, x64=57)
 
 #  ssize_t read(int fd, void *buf, size_t count);
 #
@@ -178,7 +161,7 @@ write = IrregularEmulatedSyscall(x86=4, x64=1)
 # (read(2), write(2), lseek(2), fcntl(2), etc.).  The file descriptor
 # returned by a successful call will be the lowest-numbered file
 # descriptor not currently open for the process.
-open = IrregularEmulatedSyscall(x86=5, x64=2)
+open = EmulatedSyscall(x86=5, x64=2)
 
 #  int close(int fd)
 #
@@ -187,7 +170,7 @@ open = IrregularEmulatedSyscall(x86=5, x64=2)
 # on the file it was associated with, and owned by the process, are
 # removed (regardless of the file descriptor that was used to obtain
 # the lock).
-close = EmulatedSyscall(x86=6, x64=3)
+close = IrregularEmulatedSyscall(x86=6, x64=3)
 
 #  pid_t waitpid(pid_t pid, int *status, int options);
 #
@@ -245,8 +228,8 @@ mknod = UnsupportedSyscall(x86=14, x64=133)
 # The mode of the file given by path or referenced by fildes is
 # changed.
 chmod = EmulatedSyscall(x86=15, x64=90)
-lchown = UnsupportedSyscall(x86=16, x64=94)
-_break = UnsupportedSyscall(x86=17)
+lchown = EmulatedSyscall(x86=16, x64=94)
+_break = InvalidSyscall(x86=17)
 oldstat = UnsupportedSyscall(x86=18)
 
 #  off_t lseek(int fd, off_t offset, int whence)
@@ -291,7 +274,7 @@ oldfstat = UnsupportedSyscall(x86=28)
 # pause() causes the calling process (or thread) to sleep until a
 # signal is delivered that either terminates the process or causes
 # the invocation of a signal-catching function.
-pause = EmulatedSyscall(x86=29, x64=34)
+pause = IrregularEmulatedSyscall(x86=29, x64=34)
 
 #  int utime(const char *filename, const struct utimbuf *times)
 #
@@ -310,8 +293,8 @@ pause = EmulatedSyscall(x86=29, x64=34)
 # FIXME: is mod_time set by the kernel?
 utime = EmulatedSyscall(x86=30, x64=132)
 
-stty = UnsupportedSyscall(x86=31)
-gtty = UnsupportedSyscall(x86=32)
+stty = InvalidSyscall(x86=31)
+gtty = InvalidSyscall(x86=32)
 
 #  int access(const char *pathname, int mode);
 #
@@ -320,7 +303,7 @@ gtty = UnsupportedSyscall(x86=32)
 access = EmulatedSyscall(x86=33, x64=21)
 
 nice = UnsupportedSyscall(x86=34)
-ftime = UnsupportedSyscall(x86=35)
+ftime = InvalidSyscall(x86=35)
 sync = UnsupportedSyscall(x86=36, x64=162)
 
 #  int kill(pid_t pid, int sig)
@@ -367,7 +350,7 @@ pipe = EmulatedSyscall(x86=42, x64=22, arg1="int[2]")
 #  points to.  The struct tms is as defined in <sys/times.h>:
 times = EmulatedSyscall(x86=43, x64=100, arg1="typename Arch::tms")
 
-prof = UnsupportedSyscall(x86=44)
+prof = InvalidSyscall(x86=44)
 
 #  int brk(void *addr)
 #
@@ -391,7 +374,7 @@ geteuid = EmulatedSyscall(x86=49, x64=107)
 getegid = EmulatedSyscall(x86=50, x64=108)
 acct = UnsupportedSyscall(x86=51, x64=163)
 umount2 = UnsupportedSyscall(x86=52, x64=166)
-lock = UnsupportedSyscall(x86=53)
+lock = InvalidSyscall(x86=53)
 
 #  int ioctl(int d, int request, ...)
 #
@@ -403,7 +386,7 @@ lock = UnsupportedSyscall(x86=53)
 ioctl = IrregularEmulatedSyscall(x86=54, x64=16)
 
 fcntl = IrregularEmulatedSyscall(x86=55, x64=72)
-mpx = UnsupportedSyscall(x86=56)
+mpx = InvalidSyscall(x86=56)
 
 #  int setpgid(pid_t pid, pid_t pgid);
 #
@@ -419,7 +402,7 @@ mpx = UnsupportedSyscall(x86=56)
 # joining process.
 setpgid = EmulatedSyscall(x86=57, x64=109)
 
-ulimit = UnsupportedSyscall(x86=58)
+ulimit = InvalidSyscall(x86=58)
 oldolduname = UnsupportedSyscall(x86=59)
 
 #  mode_t umask(mode_t mask);
@@ -429,14 +412,14 @@ oldolduname = UnsupportedSyscall(x86=59)
 # used), and returns the previous value of the mask.
 umask = EmulatedSyscall(x86=60, x64=95)
 
-chroot = UnsupportedSyscall(x86=61, x64=161)
+chroot = EmulatedSyscall(x86=61, x64=161)
 ustat = UnsupportedSyscall(x86=62, x64=136)
 
 #  int dup2(int oldfd, int newfd)
 #
 # dup2() makes newfd be the copy of oldfd, closing newfd first if
 #  necessary, but note the following..
-dup2 = EmulatedSyscall(x86=63, x64=33)
+dup2 = IrregularEmulatedSyscall(x86=63, x64=33)
 
 #  pid_t getppid(void);
 #
@@ -474,7 +457,7 @@ sgetmask = UnsupportedSyscall(x86=68)
 ssetmask = UnsupportedSyscall(x86=69)
 setreuid = UnsupportedSyscall(x86=70, x64=113)
 setregid = UnsupportedSyscall(x86=71, x64=114)
-sigsuspend = UnsupportedSyscall(x86=72)
+sigsuspend = IrregularEmulatedSyscall(x86=72)
 sigpending = UnsupportedSyscall(x86=73)
 sethostname = UnsupportedSyscall(x86=74, x64=170)
 
@@ -507,7 +490,7 @@ getrusage = EmulatedSyscall(x86=77, x64=98, arg2="typename Arch::rusage")
 gettimeofday = EmulatedSyscall(x86=78, x64=96, arg1="typename Arch::timeval", arg2="typename Arch::timezone")
 
 settimeofday = UnsupportedSyscall(x86=79, x64=164)
-getgroups = EmulatedSyscall(x86=80, x64=115, arg2=DynamicSize("(int)t->regs().syscall_result_signed() * LegacyUIDSyscall<Arch>::size"))
+getgroups = IrregularEmulatedSyscall(x86=80, x64=115)
 setgroups = UnsupportedSyscall(x86=81, x64=116)
 select = IrregularEmulatedSyscall(x86=82, x64=23)
 
@@ -515,8 +498,6 @@ select = IrregularEmulatedSyscall(x86=82, x64=23)
 #
 # symlink() creates a symbolic link named newpath which contains the
 # string oldpath.
-#
-# FIXME: Why was this disabled?
 symlink = EmulatedSyscall(x86=83, x64=88)
 
 oldlstat = UnsupportedSyscall(x86=84)
@@ -528,7 +509,7 @@ oldlstat = UnsupportedSyscall(x86=84)
 # null byte to buf.  It will truncate the contents (to a length of
 # bufsiz characters), in case the buffer is too small to hold all of
 # the contents.
-readlink = EmulatedSyscall(x86=85, x64=89, arg2=DynamicSize("(size_t)t->regs().arg3()"))
+readlink = IrregularEmulatedSyscall(x86=85, x64=89)
 
 uselib = UnsupportedSyscall(x86=86, x64=134)
 swapon = UnsupportedSyscall(x86=87, x64=167)
@@ -569,7 +550,7 @@ ftruncate = EmulatedSyscall(x86=93, x64=77)
 # open file descriptor fd
 fchmod = EmulatedSyscall(x86=94, x64=91)
 
-fchown = UnsupportedSyscall(x86=95, x64=93)
+fchown = EmulatedSyscall(x86=95, x64=93)
 
 #  int getpriority(int which, int who);
 #
@@ -582,9 +563,9 @@ getpriority = EmulatedSyscall(x86=96, x64=140)
 # The scheduling priority of the process, process group, or user, as
 # indicated by which and who is obtained with the getpriority() call
 # and set with the setpriority() call.
-setpriority = EmulatedSyscall(x86=97, x64=141)
+setpriority = IrregularEmulatedSyscall(x86=97, x64=141)
 
-profil = UnsupportedSyscall(x86=98)
+profil = InvalidSyscall(x86=98)
 
 #  int statfs(const char *path, struct statfs *buf)
 #
@@ -669,7 +650,7 @@ fsync = EmulatedSyscall(x86=118, x64=74)
 # When the Linux kernel creates the stack frame for a signal handler,
 # a call to sigreturn() is inserted into the stack frame so that upon
 # return from the signal handler, sigreturn() will be called.
-sigreturn = IrregularEmulatedSyscall(x86=119)
+sigreturn = EmulatedSyscall(x86=119)
 
 #  int clone(int (*fn)(void *), void *child_stack, int flags, void *arg, (pid_t
 #*ptid, struct user_desc *tls, pid_t *ctid));
@@ -748,7 +729,7 @@ fchdir = EmulatedSyscall(x86=133, x64=81)
 bdflush = UnsupportedSyscall(x86=134)
 sysfs = UnsupportedSyscall(x86=135, x64=139)
 personality = UnsupportedSyscall(x86=136, x64=135)
-afs_syscall = UnsupportedSyscall(x86=137, x64=183)
+afs_syscall = InvalidSyscall(x86=137, x64=183)
 setfsuid = UnsupportedSyscall(x86=138, x64=122)
 setfsgid = UnsupportedSyscall(x86=139, x64=123)
 
@@ -770,7 +751,7 @@ _llseek = EmulatedSyscall(x86=140, arg4="typename Arch::__kernel_loff_t")
 # from the directory referred to by the open file descriptor fd into
 # the buffer pointed to by dirp.  The argument count specifies the
 # size of that buffer.
-getdents = EmulatedSyscall(x86=141, x64=78, arg2=DynamicSize("(int)t->regs().syscall_result_signed()"))
+getdents = IrregularEmulatedSyscall(x86=141, x64=78)
 
 #  int select(int nfds, fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
 #struct timeval *timeout);
@@ -781,7 +762,7 @@ getdents = EmulatedSyscall(x86=141, x64=78, arg2=DynamicSize("(int)t->regs().sys
 # possible).  A file descriptor is considered ready if it is possible
 # to perform the corresponding I/O operation (e.g., read(2)) without
 # blocking.
-_newselect = EmulatedSyscall(x86=142, arg2="typename Arch::fd_set", arg3="typename Arch::fd_set", arg4="typename Arch::fd_set", arg5="typename Arch::timeval")
+_newselect = IrregularEmulatedSyscall(x86=142)
 
 flock = UnsupportedSyscall(x86=143, x64=73)
 
@@ -872,7 +853,7 @@ sched_getscheduler = EmulatedSyscall(x86=157, x64=145)
 # sched_yield() causes the calling thread to relinquish the CPU.  The
 # thread is moved to the end of the queue for its static priority and
 # a new thread gets to run.
-sched_yield = EmulatedSyscall(x86=158, x64=24)
+sched_yield = IrregularEmulatedSyscall(x86=158, x64=24)
 
 #  int sched_get_priority_max(int policy)
 #
@@ -912,7 +893,7 @@ mremap = ExecutedSyscall(x86=163, x64=25)
 # saved set-user-ID of the calling process.
 setresuid = EmulatedSyscall(x86=164, x64=117)
 
-getresuid = EmulatedSyscall(x86=165, x64=118, arg1=DynamicSize("LegacyUIDSyscall<Arch>::size"), arg2=DynamicSize("LegacyUIDSyscall<Arch>::size"), arg3=DynamicSize("LegacyUIDSyscall<Arch>::size"))
+getresuid = EmulatedSyscall(x86=165, x64=118, arg1="typename Arch::legacy_uid_t", arg2="typename Arch::legacy_uid_t", arg3="typename Arch::legacy_uid_t")
 vm86 = UnsupportedSyscall(x86=166)
 query_module = UnsupportedSyscall(x86=167, x64=178)
 
@@ -940,7 +921,7 @@ nfsservctl = UnsupportedSyscall(x86=169, x64=180)
 # set-group-ID of the calling process.
 setresgid = EmulatedSyscall(x86=170, x64=119)
 
-getresgid = EmulatedSyscall(x86=171, x64=120, arg1=DynamicSize("LegacyUIDSyscall<Arch>::size"), arg2=DynamicSize("LegacyUIDSyscall<Arch>::size"), arg3=DynamicSize("LegacyUIDSyscall<Arch>::size"))
+getresgid = EmulatedSyscall(x86=171, x64=120, arg1="typename Arch::legacy_gid_t", arg2="typename Arch::legacy_gid_t", arg3="typename Arch::legacy_gid_t")
 
 #  int prctl(int option, unsigned long arg2, unsigned long arg3, unsigned long
 #arg4, unsigned long arg5);
@@ -951,7 +932,7 @@ getresgid = EmulatedSyscall(x86=171, x64=120, arg1=DynamicSize("LegacyUIDSyscall
 #
 prctl = IrregularMayExecSyscall(x86=172, x64=157)
 
-rt_sigreturn = IrregularEmulatedSyscall(x86=173, x64=15)
+rt_sigreturn = EmulatedSyscall(x86=173, x64=15)
 rt_sigaction = EmulatedSyscall(x86=174, x64=13, arg3="typename Arch::kernel_sigaction")
 rt_sigprocmask = EmulatedSyscall(x86=175, x64=14, arg3="typename Arch::sigset_t")
 
@@ -961,7 +942,7 @@ rt_sigprocmask = EmulatedSyscall(x86=175, x64=14, arg3="typename Arch::sigset_t"
 # delivery to the calling thread (i.e., the signals which have been
 # raised while blocked).  The mask of pending signals is returned in
 # set.
-rt_sigpending = EmulatedSyscall(x86=176, x64=127, arg1=DynamicSize("t->regs().arg2()"))
+rt_sigpending = IrregularEmulatedSyscall(x86=176, x64=127)
 
 #  int sigtimedwait(const sigset_t *set, siginfo_t *info,
 #                   const struct timespec *timeout);
@@ -977,30 +958,22 @@ rt_sigpending = EmulatedSyscall(x86=176, x64=127, arg1=DynamicSize("t->regs().ar
 # signal.
 rt_sigtimedwait = IrregularEmulatedSyscall(x86=177, x64=128)
 
-#  int sigpending(sigset_t *set);
-#
-# sigpending() returns the set of signals that are pending for
-# delivery to the calling thread (i.e., the signals which have been
-# raised while blocked).  The mask of pending signals is returned in
-# set.
-rt_sigqueueinfo = EmulatedSyscall(x86=178, x64=129, arg1=DynamicSize("t->regs().arg2()"))
-
 #  int sigsuspend(const sigset_t *mask);
 #
 # sigsuspend() temporarily replaces the signal mask of the calling
 # process with the mask given by mask and then suspends the process
 # until delivery of a signal whose action is to invoke a signal
 # handler or to terminate a process.
-rt_sigsuspend = EmulatedSyscall(x86=179, x64=130)
+rt_sigsuspend = IrregularEmulatedSyscall(x86=179, x64=130)
 
 #  ssize_t pread(int fd, void *buf, size_t count, off_t offset);
 #
 # pread, pwrite - read from or write to a file descriptor at a given
 # offset
-pread64 = EmulatedSyscall(x86=180, x64=17, arg2=DynamicSize("(ssize_t)t->regs().syscall_result_signed()"))
+pread64 = IrregularEmulatedSyscall(x86=180, x64=17)
 pwrite64 = EmulatedSyscall(x86=181, x64=18)
 
-chown = UnsupportedSyscall(x86=182, x64=92)
+chown = EmulatedSyscall(x86=182, x64=92)
 
 #  char *getwd(char *buf);
 #
@@ -1008,7 +981,7 @@ chown = UnsupportedSyscall(x86=182, x64=92)
 # absolute pathname that is the current working directory of the
 # calling process.  The pathname is returned as the function result
 # and via the argument buf, if present.
-getcwd = EmulatedSyscall(x86=183, x64=79, arg1=NullTerminatedString())
+getcwd = IrregularEmulatedSyscall(x86=183, x64=79)
 
 capget = UnsupportedSyscall(x86=184, x64=125)
 capset = UnsupportedSyscall(x86=185, x64=126)
@@ -1020,12 +993,12 @@ capset = UnsupportedSyscall(x86=185, x64=126)
 # stack.  An alternate signal stack is used during the execution of a
 # signal handler if the establishment of that handler (see
 # sigaction(2)) requested it.
-sigaltstack = EmulatedSyscall(x86=186, x64=131, arg2=DynamicSize("t->regs().arg2() ? sizeof(typename Arch::stack_t) : 0"))
+sigaltstack = EmulatedSyscall(x86=186, x64=131, arg2="typename Arch::stack_t")
 
 sendfile = IrregularEmulatedSyscall(x86=187, x64=40)
-getpmsg = UnsupportedSyscall(x86=188, x64=181)
-putpmsg = UnsupportedSyscall(x86=189, x64=182)
-vfork = UnsupportedSyscall(x86=190, x64=58)
+getpmsg = InvalidSyscall(x86=188, x64=181)
+putpmsg = InvalidSyscall(x86=189, x64=182)
+vfork = IrregularEmulatedSyscall(x86=190, x64=58)
 
 #  int getrlimit(int resource, struct rlimit *rlim)
 #
@@ -1058,7 +1031,7 @@ lstat64 = EmulatedSyscall(x86=196, arg2="typename Arch::stat64")
 # is specified by the file descriptor fd.
 fstat64 = EmulatedSyscall(x86=197, arg2="typename Arch::stat64")
 
-lchown32 = UnsupportedSyscall(x86=198)
+lchown32 = EmulatedSyscall(x86=198)
 
 #  uid_t getuid(void);
 #
@@ -1102,10 +1075,10 @@ setregid32 = EmulatedSyscall(x86=204)
 # supplementary group IDs for the process is returned.  This allows
 # the caller to determine the size of a dynamically allocated list to
 # be used in a further call to getgroups().
-getgroups32 = EmulatedSyscall(x86=205, arg2=DynamicSize("(int)t->regs().syscall_result_signed() * sizeof(typename Arch::gid_t)"))
+getgroups32 = IrregularEmulatedSyscall(x86=205)
 
 setgroups32 = UnsupportedSyscall(x86=206)
-fchown32 = UnsupportedSyscall(x86=207)
+fchown32 = EmulatedSyscall(x86=207)
 
 #  int setresuid32(uid_t ruid, uid_t euid, uid_t suid);
 #
@@ -1137,13 +1110,13 @@ setresgid32 = EmulatedSyscall(x86=210)
 # appropriately.
 getresgid32 = EmulatedSyscall(x86=211, arg1="typename Arch::gid_t", arg2="typename Arch::gid_t", arg3="typename Arch::gid_t")
 
-chown32 = UnsupportedSyscall(x86=212)
+chown32 = EmulatedSyscall(x86=212)
 setuid32 = UnsupportedSyscall(x86=213)
 setgid32 = UnsupportedSyscall(x86=214)
 setfsuid32 = UnsupportedSyscall(x86=215)
 setfsgid32 = UnsupportedSyscall(x86=216)
 pivot_root = UnsupportedSyscall(x86=217, x64=155)
-mincore = UnsupportedSyscall(x86=218, x64=27)
+mincore = IrregularEmulatedSyscall(x86=218, x64=27)
 
 #  int madvise(void *addr, size_t length, int advice);
 #
@@ -1158,7 +1131,7 @@ mincore = UnsupportedSyscall(x86=218, x64=27)
 # advice.
 madvise = ExecutedSyscall(x86=219, x64=28)
 
-getdents64 = EmulatedSyscall(x86=220, x64=217, arg2=DynamicSize("(int)t->regs().syscall_result_signed()"))
+getdents64 = IrregularEmulatedSyscall(x86=220, x64=217)
 
 #  int fcntl(int fd, int cmd, ... ( arg ));
 #
@@ -1191,9 +1164,9 @@ gettid = EmulatedSyscall(x86=224, x64=186)
 # offset of the open file referred to by fd is left unchanged.
 readahead = EmulatedSyscall(x86=225, x64=187)
 
-setxattr = UnsupportedSyscall(x86=226, x64=188)
-lsetxattr = UnsupportedSyscall(x86=227, x64=189)
-fsetxattr = UnsupportedSyscall(x86=228, x64=190)
+setxattr = EmulatedSyscall(x86=226, x64=188)
+lsetxattr = EmulatedSyscall(x86=227, x64=189)
+fsetxattr = EmulatedSyscall(x86=228, x64=190)
 
 #  ssize_t getxattr(const char *path, const char *name,
 #                   void *value, size_t size);
@@ -1290,7 +1263,7 @@ fadvise64 = EmulatedSyscall(x86=250, x64=221)
 # This system call is equivalent to exit(2) except that it terminates
 # not only the calling thread, but all threads in the calling
 # process's thread group.
-exit_group = IrregularMayExecSyscall(x86=252, x64=231)
+exit_group = IrregularEmulatedSyscall(x86=252, x64=231)
 
 lookup_dcookie = UnsupportedSyscall(x86=253, x64=212)
 
@@ -1336,7 +1309,7 @@ remap_file_pages = UnsupportedSyscall(x86=257, x64=216)
 #
 # When set_child_tid is set, the very first thing the new process
 # does is writing its PID at this address.
-set_tid_address = ExecRetEmuSyscall(x86=258, x64=218, arg1="typename Arch::pid_t")
+set_tid_address = ExecutedSyscall(x86=258, x64=218, arg1="typename Arch::pid_t")
 
 timer_create = UnsupportedSyscall(x86=259, x64=222)
 timer_settime = UnsupportedSyscall(x86=260, x64=223)
@@ -1391,12 +1364,12 @@ tgkill = EmulatedSyscall(x86=270, x64=234)
 # of the inode specified by filename to the actime and modtime fields
 # of times respectively.
 #
-utimes = EmulatedSyscall(x86=271, x64=235, arg2=DynamicSize("2 * sizeof(typename Arch::timeval)"))
+utimes = EmulatedSyscall(x86=271, x64=235)
 
 fadvise64_64 = EmulatedSyscall(x86=272)
 
-vserver = UnsupportedSyscall(x86=273, x64=236)
-mbind = UnsupportedSyscall(x86=274, x64=237)
+vserver = InvalidSyscall(x86=273, x64=236)
+mbind = EmulatedSyscall(x86=274, x64=237)
 get_mempolicy = UnsupportedSyscall(x86=275, x64=239)
 set_mempolicy = UnsupportedSyscall(x86=276, x64=238)
 mq_open = UnsupportedSyscall(x86=277, x64=240)
@@ -1495,11 +1468,11 @@ fchmodat = UnsupportedSyscall(x86=306, x64=268)
 # page....
 faccessat = EmulatedSyscall(x86=307, x64=269)
 
-pselect6 = UnsupportedSyscall(x86=308, x64=270)
+pselect6 = IrregularEmulatedSyscall(x86=308, x64=270)
 
 ppoll = IrregularEmulatedSyscall(x86=309, x64=271)
 
-unshare = UnsupportedSyscall(x86=310, x64=272)
+unshare = IrregularEmulatedSyscall(x86=310, x64=272)
 
 #  long set_robust_list(struct robust_list_head *head, size_t len)
 #
@@ -1612,7 +1585,7 @@ eventfd2 = EmulatedSyscall(x86=328, x64=290)
 # file descriptor.
 epoll_create1 = EmulatedSyscall(x86=329, x64=291)
 
-dup3 = UnsupportedSyscall(x86=330, x64=292)
+dup3 = IrregularEmulatedSyscall(x86=330, x64=292)
 
 #  int pipe2(int pipefd[2], int flags)
 #
@@ -1625,6 +1598,7 @@ inotify_init1 = EmulatedSyscall(x86=332, x64=294)
 preadv = IrregularEmulatedSyscall(x86=333, x64=295)
 pwritev = EmulatedSyscall(x86=334, x64=296)
 
+#  int rt_sigqueueinfo(pid_t tgid, int sig, siginfo_t *uinfo);
 #  int rt_tgsigqueueinfo(pid_t tgid, pid_t tid, int sig,
 #                        siginfo_t *uinfo);
 #
@@ -1633,6 +1607,7 @@ pwritev = EmulatedSyscall(x86=334, x64=296)
 # or thread.  The receiver of the signal can obtain the accompanying
 # data by establishing a signal handler with the sigaction(2)
 # SA_SIGINFO flag.
+rt_sigqueueinfo = EmulatedSyscall(x86=178, x64=129)
 rt_tgsigqueueinfo = EmulatedSyscall(x86=335, x64=297)
 
 #  int perf_event_open(struct perf_event_attr *attr,
@@ -1690,12 +1665,18 @@ process_vm_readv = UnsupportedSyscall(x86=347, x64=310)
 process_vm_writev = UnsupportedSyscall(x86=348, x64=311)
 kcmp = UnsupportedSyscall(x86=349, x64=312)
 finit_module = UnsupportedSyscall(x86=350, x64=313)
+sched_setattr = UnsupportedSyscall(x86=351, x64=314)
+sched_getattr = UnsupportedSyscall(x86=352, x64=315)
+renameat2 = UnsupportedSyscall(x86=353, x64=316)
+seccomp = IrregularEmulatedSyscall(x86=354, x64=317)
+getrandom = IrregularEmulatedSyscall(x86=355, x64=318)
 
 # restart_syscall is a little special.
 restart_syscall = RestartSyscall(x86=0, x64=219)
 
-rrcall_init_buffers = IrregularEmulatedSyscall(x86=442, x64=442)
-rrcall_init_preload = IrregularEmulatedSyscall(x86=443, x64=443)
+rrcall_init_preload = IrregularEmulatedSyscall(x86=442, x64=442)
+rrcall_init_buffers = IrregularEmulatedSyscall(x86=443, x64=443)
+rrcall_notify_syscall_hook_exit = IrregularEmulatedSyscall(x86=444, x64=444)
 
 # These syscalls are subsumed under socketcall on x86.
 socket = EmulatedSyscall(x64=41)
@@ -1703,7 +1684,7 @@ connect = EmulatedSyscall(x64=42)
 accept = IrregularEmulatedSyscall(x64=43)
 sendto = EmulatedSyscall(x64=44)
 recvfrom = IrregularEmulatedSyscall(x64=45)
-sendmsg = EmulatedSyscall(x64=46)
+sendmsg = IrregularEmulatedSyscall(x64=46)
 recvmsg = IrregularEmulatedSyscall(x64=47)
 shutdown = EmulatedSyscall(x64=48)
 bind = EmulatedSyscall(x64=49)
@@ -1716,12 +1697,25 @@ getsockopt = IrregularEmulatedSyscall(x64=55)
 accept4 = IrregularEmulatedSyscall(x64=288)
 
 # These syscalls are subsumed under ipc on x86.
+shmget = EmulatedSyscall(x64=29)
+shmat = IrregularEmulatedSyscall(x64=30)
+shmctl = IrregularEmulatedSyscall(x64=31)
+semget = EmulatedSyscall(x64=64)
+semop = IrregularEmulatedSyscall(x64=65)
+semctl = IrregularEmulatedSyscall(x64=66)
+shmdt = IrregularEmulatedSyscall(x64=67)
 msgget = EmulatedSyscall(x64=68)
-msgsnd = EmulatedSyscall(x64=69)
+msgsnd = IrregularEmulatedSyscall(x64=69)
 msgrcv = IrregularEmulatedSyscall(x64=70)
 msgctl = IrregularEmulatedSyscall(x64=71)
+semtimedop = IrregularEmulatedSyscall(x64=220)
 
+# These syscalls simply don't exist on x86.
 arch_prctl = ExecutedSyscall(x64=158)
+tuxcall = InvalidSyscall(x64=184)
+security = InvalidSyscall(x64=185)
+epoll_ctl_old = UnsupportedSyscall(x64=214)
+epoll_wait_old = UnsupportedSyscall(x64=215)
 
 def _syscalls():
     for name, obj in globals().iteritems():
