@@ -11,10 +11,13 @@
 #include "AddressSpace.h"
 #include "Event.h"
 #include "ExtraRegisters.h"
+#include "FdTable.h"
 #include "kernel_abi.h"
 #include "kernel_supplement.h"
 #include "PerfCounters.h"
+#include "PropertyTable.h"
 #include "Registers.h"
+#include "TaskishUid.h"
 #include "TraceStream.h"
 #include "util.h"
 
@@ -24,11 +27,36 @@ class RecordSession;
 class ReplaySession;
 struct Sighandlers;
 class Task;
-struct TaskGroup;
 
 struct syscallbuf_hdr;
 struct syscallbuf_record;
 
+/**
+ * A list of return addresses extracted from the stack. The tuple
+ * (perfcounter ticks, regs, return addresses) may be needed to disambiguate
+ * states that aren't unique in (perfcounter ticks, regs).
+ * When return addresses can't be extracted, some suffix of the list may be
+ * all zeroes.
+ */
+struct ReturnAddressList {
+  enum {
+    COUNT = 8
+  };
+  remote_ptr<void> addresses[COUNT];
+
+  bool operator==(const ReturnAddressList& other) const {
+    for (int i = 0; i < COUNT; ++i) {
+      if (addresses[i] != other.addresses[i]) {
+        return false;
+      }
+    }
+    return true;
+  }
+  bool operator!=(const ReturnAddressList& other) const {
+    return !(*this == other);
+  }
+};
+
 /**
  * Tracks a group of tasks with an associated ID, set from the
  * original "thread group leader", the child of |fork()| which became
@@ -37,7 +65,9 @@ struct syscallbuf_record;
  */
 class TaskGroup : public HasTaskSet {
 public:
-  TaskGroup(pid_t tgid, pid_t real_tgid);
+  TaskGroup(Session* session, TaskGroup* parent, pid_t tgid, pid_t real_tgid,
+            uint32_t serial);
+  ~TaskGroup();
 
   typedef std::shared_ptr<TaskGroup> shr_ptr;
 
@@ -49,9 +79,28 @@ public:
 
   int exit_code;
 
+  Session* session() const { return session_; }
+  void forget_session() { session_ = nullptr; }
+
+  TaskGroup* parent() { return parent_; }
+
+  TaskGroupUid tguid() const { return TaskGroupUid(tgid, serial); }
+
+  // We don't allow tasks to make themselves undumpable. If they try,
+  // record that here and lie about it if necessary.
+  bool dumpable;
+
 private:
   TaskGroup(const TaskGroup&) = delete;
   TaskGroup operator=(const TaskGroup&) = delete;
+
+  Session* session_;
+  /** Parent TaskGroup, or nullptr if it's not a tracee (rr or init). */
+  TaskGroup* parent_;
+
+  std::set<TaskGroup*> children;
+
+  uint32_t serial;
 };
 
 enum CloneFlags {
@@ -70,10 +119,12 @@ enum CloneFlags {
   CLONE_SHARE_TASK_GROUP = 1 << 1,
   /** Child will share its parent's address space. */
   CLONE_SHARE_VM = 1 << 2,
+  /** Child will share its parent's file descriptor table. */
+  CLONE_SHARE_FILES = 1 << 3,
   /** Kernel will clear and notify tid futex on task exit. */
-  CLONE_CLEARTID = 1 << 3,
+  CLONE_CLEARTID = 1 << 4,
   // Set the thread area to what's specified by the |tls| arg.
-  CLONE_SET_TLS = 1 << 4,
+  CLONE_SET_TLS = 1 << 5,
 };
 
 /**
@@ -105,14 +156,27 @@ enum ShareDeschedEventFd {
   DONT_SHARE_DESCHED_EVENT_FD = 0
 };
 
-enum DestroyBufferFlags {
-  DESTROY_SCRATCH = 1 << 0,
-  DESTROY_SYSCALLBUF = 1 << 1,
+/** Different kinds of waits a task can do.
+ */
+enum WaitType {
+  // Not waiting for anything
+  WAIT_TYPE_NONE,
+  // Waiting for any child process
+  WAIT_TYPE_ANY,
+  // Waiting for any child with the same process group ID
+  WAIT_TYPE_SAME_PGID,
+  // Waiting for any child with a specific process group ID
+  WAIT_TYPE_PGID,
+  // Waiting for a specific process ID
+  WAIT_TYPE_PID
 };
 
-enum Switchable {
-  PREVENT_SWITCH,
-  ALLOW_SWITCH
+/** Reasons why we simulate stopping of a task (see ptrace(2) man page).
+ */
+enum EmulatedStopType {
+  NOT_STOPPED,
+  GROUP_STOP,          // stopped by a signal. This applies to non-ptracees too.
+  SIGNAL_DELIVERY_STOP // Stopped before delivering a signal. ptracees only.
 };
 
 /**
@@ -181,12 +245,6 @@ public:
    * |cont_sysemu()| or |cont_sysemu_singlestep()|, but that's
    * not checked.  If so, step over the system call instruction
    * to "exit" the emulated syscall.
-   *
-   * This operation is (assumed to be) idempotent:
-   * |finish_emulated_syscall()| can be called any number of
-   * times without changing state other than
-   * emulated-syscall-exited-ness.  Checkpointing relies on this
-   * assumption.
    */
   void finish_emulated_syscall();
 
@@ -198,6 +256,16 @@ public:
    */
   const struct syscallbuf_record* desched_rec() const;
 
+  size_t syscallbuf_data_size() const {
+    if (syscallbuf_hdr->locked) {
+      // There may be an incomplete syscall record after num_rec_bytes that
+      // we need to record. We don't know how big that record is,
+      // so just record the entire buffer. This should not be common.
+      return num_syscallbuf_bytes;
+    }
+    return syscallbuf_hdr->num_rec_bytes + sizeof(*syscallbuf_hdr);
+  }
+
   /**
    * Mark the members of this task's group as "unstable",
    * meaning that even though a task may look runnable, it
@@ -264,6 +332,57 @@ public:
    */
   void destabilize_task_group();
 
+  /**
+   * Print out a warning message when we're in the process of delivering a
+   * fatal signal, if appropriate.
+   */
+  void notify_fatal_signal();
+
+  /**
+   * Emulate 'tracer' ptracing this task.
+   */
+  void set_emulated_ptracer(Task* tracer);
+
+  /**
+   * Call this when an event occurs that should stop a ptraced task.
+   * If we're emulating ptrace of the task, stop the task and wake the ptracer
+   * if it's waiting, and queue "code" as an status code to be reported to the
+   * ptracer.
+   * Returns true if the task is stopped-for-emulated-ptrace, false otherwise.
+   */
+  bool emulate_ptrace_stop(int code, EmulatedStopType stop_type);
+  /**
+   * Force the ptrace-stop state no matter what state the task is currently in.
+   */
+  void force_emulate_ptrace_stop(int code, EmulatedStopType stop_type);
+
+  /**
+   * Called when this task is able to receive a SIGCHLD (e.g. because
+   * we completed delivery of a signal already). Sends a new synthetic
+   * SIGCHLD to the task if there are still ptraced tasks that need a SIGCHLD
+   * sent for them.
+   */
+  void send_synthetic_SIGCHLD_if_necessary();
+  /**
+   * Called when we're about to deliver a signal to this task. If it's a
+   * synthetic SIGCHLD and there's a ptraced task that needs to SIGCHLD,
+   * update the siginfo to reflect the status and note that that
+   * ptraced task has had its SIGCHLD sent.
+   */
+  void set_siginfo_for_synthetic_SIGCHLD(siginfo_t* si);
+
+  /**
+   * Returns true if this task is in a waitpid or similar that would return
+   * when t's status changes due to a ptrace event.
+   */
+  bool is_waiting_for_ptrace(Task* t);
+
+  /**
+   * Returns true if this task is in a waitpid or similar that would return
+   * when t's status changes due to a regular event (exit).
+   */
+  bool is_waiting_for(Task* t);
+
   /**
    * Dump attributes of this process, including pending events,
    * to |out|, which defaults to LOG_FILE.
@@ -290,12 +409,6 @@ public:
    */
   void set_tick_count(Ticks count);
 
-  /**
-   * Return the exe path passed to the most recent (successful)
-   * execve call.
-   */
-  const std::string& exec_file() const { return execve_file; }
-
   /**
    * Return true if this exited because of a SYS_exit/exit_group
    * call.
@@ -312,9 +425,7 @@ public:
   };
   /**
    * Stat |fd| in the context of this task's fd table, returning
-   * the result in |buf|.  The name of the referent file is
-   * returned in |buf|, of max size |buf_num_bytes|.  Return
-   * true on success, false on error.
+   * the result and the file name in FStatResult.
    */
   FStatResult fstat(int fd);
 
@@ -322,8 +433,17 @@ public:
    * Force the wait status of this to |status|, as if
    * |wait()/try_wait()| had returned it. Call this whenever a waitpid
    * returned activity for this past.
+   * If override_siginfo is non-null and status indicates a pending signal,
+   * use *override_siginfo as the siginfo instead of reading it from the kernel.
    */
-  void did_waitpid(int status);
+  void did_waitpid(int status, siginfo_t* override_siginfo = nullptr);
+
+  /**
+   * If we're at the entry to (or exit from) a syscall, we may need to
+   * fix up registers that are set nondeterministically so they're
+   * deterministic.
+   */
+  void fixup_syscall_regs();
 
   /**
    * Wait for |futex| in this address space to have the value
@@ -341,6 +461,8 @@ public:
    */
   pid_t get_ptrace_eventmsg_pid();
 
+  uint16_t get_ptrace_eventmsg_seccomp_data();
+
   /**
    * Return the siginfo at the signal-stop of this.
    * Not meaningful unless this is actually at a signal stop.
@@ -369,16 +491,16 @@ public:
    *
    * Pass SHARE_DESCHED_EVENT_FD to additionally share that fd.
    */
-  remote_ptr<void> init_buffers(remote_ptr<void> map_hint,
-                                ShareDeschedEventFd share_desched_fd);
+  void init_buffers(remote_ptr<void> map_hint,
+                    ShareDeschedEventFd share_desched_fd);
 
   /**
-   * Destroy in the tracee task the buffer(s) specified by the
-   * DestroyBufferFlags mask |which|.  This task must already be
-   * at a state in which remote syscalls can be executed; if
-   * it's not, results are undefined.
+   * Destroy in the tracee task the scratch buffer and syscallbuf (if
+   * syscallbuf_child is non-null).
+   * This task must already be at a state in which remote syscalls can be
+   * executed; if it's not, results are undefined.
    */
-  void destroy_buffers(int which);
+  void destroy_buffers();
 
   /** Return the current $ip of this. */
   remote_ptr<uint8_t> ip() { return regs().ip(); }
@@ -489,12 +611,12 @@ public:
   bool may_be_blocked() const;
 
   /**
-   * If |syscallno| at |state| changes our VM mapping, then
-   * update the cache for that change.  The exception is mmap()
-   * calls: they're complicated enough to be handled separately.
-   * Client code should call |t->vm()->map(...)| directly.
+   * Call this hook just before exiting a syscall.  Often Task
+   * attributes need to be updated based on the finishing syscall.
+   * Use 'regs' instead of this->regs() because some registers may not be
+   * set properly in the task yet.
    */
-  void maybe_update_vm(int syscallno, SyscallEntryOrExit state);
+  void on_syscall_exit(int syscallno, const Registers& regs);
 
   /**
    * Assuming ip() is just past a breakpoint instruction, adjust
@@ -508,12 +630,6 @@ public:
    */
   const std::string& name() const { return prname; }
 
-  /**
-   * Call this method when this task has entered an |execve()|
-   * call, but before the exec has actually taken place.
-   */
-  void pre_exec();
-
   /**
    * Call this method when this task has just performed an |execve()|
    * (so we're in the new address space), but before the system call has
@@ -521,13 +637,16 @@ public:
    * During replay replay_regs is non-null and contains the register values
    * recorded immediately after the exec.
    */
-  void post_exec(const Registers* replay_regs = nullptr);
+  void post_exec(const Registers* replay_regs = nullptr,
+                 const ExtraRegisters* replay_extra_regs = nullptr);
 
   /**
    * Call this method when this task has exited a successful execve() syscall.
    * At this point it is safe to make remote syscalls.
+   * |event| is the TraceTaskEvent (EXEC) that will be recorded or is being
+   * replayed.
    */
-  void post_exec_syscall();
+  void post_exec_syscall(TraceTaskEvent& event);
 
   /**
    * Manage pending events.  |push_event()| pushes the given
@@ -536,10 +655,7 @@ public:
    * the specified type.
    */
   void push_event(const Event& ev) { pending_events.push_back(ev); }
-  void pop_event(EventType expected_type) {
-    assert(pending_events.back().type() == expected_type);
-    pending_events.pop_back();
-  }
+  void pop_event(EventType expected_type);
   void pop_noop() { pop_event(EV_NOOP); }
   void pop_desched() { pop_event(EV_DESCHED); }
   void pop_signal_delivery() { pop_event(EV_SIGNAL_DELIVERY); }
@@ -573,7 +689,7 @@ public:
    * variants record data that's already been read from this,
    * and the |record_remote*()| variants read the data and then
    * record it.
-   * If 'addr' is null then num_bytes is treated as zero.
+   * If 'addr' is null then no record is written.
    */
   void record_local(remote_ptr<void> addr, ssize_t num_bytes, const void* buf);
   template <typename T> void record_local(remote_ptr<T> addr, const T* buf) {
@@ -585,24 +701,17 @@ public:
     record_remote(addr, sizeof(T));
   }
 
-  void record_remote_str(remote_ptr<void> str);
-
   /**
-   * Attempt to find the value of |regname| (a DebuggerRegister
-   * name) in this task, and if so (i) write it to |buf|; (ii)
-   * set |*defined = true|; (iii) return the size of written
-   * data.  If |*defined == false|, the value of |buf| is
-   * meaningless.
-   *
-   * This helper can fetch the values of both general-purpose
-   * and "extra" registers.
-   *
-   * NB: |buf| must be large enough to hold the largest register
-   * value that can be named by |regname|.
-   *
-   * TODO: nicer API.
+   * Save tracee data to the trace.  |addr| is the address in
+   * the address space of this task.
+   * If 'addr' is null then a zero-length record is written.
    */
-  size_t get_reg(uint8_t* buf, GdbRegister regname, bool* defined);
+  void record_remote_even_if_null(remote_ptr<void> addr, ssize_t num_bytes);
+  template <typename T> void record_remote_even_if_null(remote_ptr<T> addr) {
+    record_remote_even_if_null(addr, sizeof(T));
+  }
+
+  void record_remote_str(remote_ptr<void> str);
 
   /** Return the current regs of this. */
   const Registers& regs() const;
@@ -611,7 +720,11 @@ public:
   const ExtraRegisters& extra_regs();
 
   /** Return the current arch of this. This can change due to exec(). */
-  SupportedArch arch() const { return regs().arch(); }
+  SupportedArch arch() const {
+    // Use 'registers' directly instead of calling regs(), since this can
+    // be called while the task is not stopped.
+    return registers.arch();
+  }
 
   enum {
     /* The x86 linux 3.5.0-36 kernel packaged with Ubuntu
@@ -628,11 +741,25 @@ public:
    */
   void validate_regs(uint32_t flags = 0);
 
+  /**
+   * Capture return addresses from this task's stack. The returned
+   * address list may not be actual return addresses (in optimized code,
+   * will probably not be), but they will be a function of the task's current
+   * state, so may be useful for distinguishing this state from other states.
+   */
+  ReturnAddressList return_addresses();
+
   /**
    * Return the debug status, which is a bitfield comprising
    * |DebugStatus| bits (see above).
    */
   uintptr_t debug_status();
+  /**
+   * Return the debug status, which is a bitfield comprising
+   * |DebugStatus| bits (see above), and clear the kernel state.
+   */
+  uintptr_t consume_debug_status();
+  void replace_debug_status(uintptr_t status);
 
   /**
    * Return the address of the watchpoint programmed at slot
@@ -645,10 +772,13 @@ public:
 
   /**
    * Read |val| from |child_addr|.
+   * If the data can't all be read, then if |ok| is non-null
+   * sets *ok to false, otherwise asserts.
    */
-  template <typename T> T read_mem(remote_ptr<T> child_addr) {
+  template <typename T>
+  T read_mem(remote_ptr<T> child_addr, bool* ok = nullptr) {
     T val;
-    read_bytes_helper(child_addr, sizeof(val), &val);
+    read_bytes_helper(child_addr, sizeof(val), &val, ok);
     return val;
   }
 
@@ -656,10 +786,11 @@ public:
    * Read |count| values from |child_addr|.
    */
   template <typename T>
-  std::vector<T> read_mem(remote_ptr<T> child_addr, size_t count) {
+  std::vector<T> read_mem(remote_ptr<T> child_addr, size_t count,
+                          bool* ok = nullptr) {
     std::vector<T> v;
     v.resize(count);
-    read_bytes_helper(child_addr, sizeof(T) * count, v.data());
+    read_bytes_helper(child_addr, sizeof(T) * count, v.data(), ok);
     return v;
   }
 
@@ -703,6 +834,9 @@ public:
   /** Restore the next chunk of saved data from the trace to this. */
   ssize_t set_data_from_trace();
 
+  /** Restore all remaining chunks of saved data for the current trace frame. */
+  void apply_all_data_records_from_trace();
+
   /**
    * Set the syscall-return-value register of this to what was
    * saved in the current trace frame.
@@ -726,6 +860,11 @@ public:
    */
   bool set_debug_regs(const DebugRegs& regs);
 
+  /**
+   * Reads the value of the given debug register.
+   */
+  uintptr_t get_debug_reg(size_t regno);
+
   /**
    * Update the futex robust list head pointer to |list| (which
    * is of size |len|).
@@ -743,7 +882,6 @@ public:
 
   /** Update the clear-tid futex to |tid_addr|. */
   void set_tid_addr(remote_ptr<int> tid_addr);
-  remote_ptr<int> tid_addr() const { return tid_futex; }
 
   /**
    * Call this after |sig| is delivered to this task.  Emulate
@@ -760,6 +898,11 @@ public:
    * invoked when |sig| is received.
    */
   bool signal_has_user_handler(int sig) const;
+  /**
+   * If signal_has_user_handler(sig) is true, return the address of the
+   * user handler, otherwise return null.
+   */
+  remote_ptr<uint8_t> get_signal_user_handler(int sig) const;
 
   /**
    * Return |sig|'s current sigaction. Returned as raw bytes since the
@@ -767,15 +910,6 @@ public:
    */
   const std::vector<uint8_t>& signal_action(int sig) const;
 
-  /**
-   * Return the |stack| argument passed to |clone()|, i.e. the
-   * top of this thread's stack if it's a thread.
-   *
-   * This IS NOT the stack pointer.  Call |sp()| if you want
-   * that.
-   */
-  remote_ptr<void> stack() const { return top_of_stack; }
-
   /**
    * Stashed-signal API: if a signal becomes pending at an
    * awkward time, but could be handled "soon", call
@@ -798,8 +932,14 @@ public:
    */
   void stash_sig();
   bool has_stashed_sig() const { return !stashed_signals.empty(); }
+  siginfo_t peek_stash_sig();
   siginfo_t pop_stash_sig();
 
+  /**
+   * Return true when the task is running, false if it's stopped.
+   */
+  bool is_running() const { return !is_stopped; }
+
   /**
    * Return the status of this as of the last successful
    * wait()/try_wait() call.
@@ -826,10 +966,7 @@ public:
    */
   int pending_sig() const { return pending_sig_from_status(wait_status); }
 
-  /**
-   * Return true if we stopped due to a syscall event.
-   */
-  bool syscall_stop() const { return stop_sig() == (SIGTRAP | 0x80); }
+  void clear_wait_status() { wait_status = 0; }
 
   /** Return the task group this belongs to. */
   TaskGroup::shr_ptr task_group() { return tg; }
@@ -839,6 +976,8 @@ public:
   /** Return id of real OS task group. */
   pid_t real_tgid() const { return tg->real_tgid; }
 
+  TaskUid tuid() const { return TaskUid(rec_tid, serial); }
+
   /** Return the dir of the trace we're using. */
   const std::string& trace_dir() const;
 
@@ -881,6 +1020,8 @@ public:
    */
   AddressSpace::shr_ptr vm() { return as; }
 
+  FdTable::shr_ptr fd_table() { return fds; }
+
   enum AllowInterrupt {
     ALLOW_INTERRUPT,
     // Pass this when the caller has already triggered a ptrace stop
@@ -899,8 +1040,13 @@ public:
   bool try_wait();
 
   /**
-   * Write |N| bytes from |buf| to |child_addr|, or don't
-   * return.
+   * Currently we don't allow recording across uid changes, so we can just
+   * return rr's uid.
+   */
+  uid_t getuid() { return ::getuid(); }
+
+  /**
+   * Write |N| bytes from |buf| to |child_addr|, or don't return.
    */
   template <size_t N>
   void write_bytes(remote_ptr<void> child_addr, const uint8_t (&buf)[N]) {
@@ -909,11 +1055,9 @@ public:
 
   /**
    * Write |val| to |child_addr|.
-   *
-   * NB: doesn't use the ptrace API, so safe to use even when
-   * the tracee isn't at a trace-stop.
    */
   template <typename T> void write_mem(remote_ptr<T> child_addr, const T& val) {
+    assert(type_has_no_holes<T>());
     write_bytes_helper(child_addr, sizeof(val), static_cast<const void*>(&val));
   }
   /**
@@ -926,6 +1070,7 @@ public:
 
   template <typename T>
   void write_mem(remote_ptr<T> child_addr, const T* val, int count) {
+    assert(type_has_no_holes<T>());
     write_bytes_helper(child_addr, sizeof(*val) * count,
                        static_cast<const void*>(val));
   }
@@ -939,7 +1084,12 @@ public:
    */
   ssize_t read_bytes_fallible(remote_ptr<void> addr, ssize_t buf_size,
                               void* buf);
-  void read_bytes_helper(remote_ptr<void> addr, ssize_t buf_size, void* buf);
+  /**
+   * If the data can't all be read, then if |ok| is non-null, sets *ok to
+   * false, otherwise asserts.
+   */
+  void read_bytes_helper(remote_ptr<void> addr, ssize_t buf_size, void* buf,
+                         bool* ok = nullptr);
   void write_bytes_helper(remote_ptr<void> addr, ssize_t buf_size,
                           const void* buf);
 
@@ -956,10 +1106,28 @@ public:
   /**
    * Call this when performing a clone syscall in this task. Returns
    * true if the call completed, false if it was interrupted and
-   * needs to be resumed.
+   * needs to be resumed. When the call returns true, the task is
+   * stopped at a PTRACE_EVENT_CLONE or PTRACE_EVENT_FORK.
    */
   bool clone_syscall_is_complete();
 
+  /**
+   * Return the pid of the newborn thread created by this task.
+   * Called when this task has a PTRACE_CLONE_EVENT with CLONE_THREAD.
+   */
+  pid_t find_newborn_thread();
+  /**
+   * Return the pid of the newborn process created by this task.
+   * Called when this task has a PTRACE_CLONE_EVENT without CLONE_THREAD,
+   * or PTRACE_FORK_EVENT.
+   */
+  pid_t find_newborn_child_process();
+
+  /**
+   * Called when SYS_rrcall_init_preload has happened.
+   */
+  void at_preload_init();
+
   /**
    * Open /proc/[tid]/mem fd for our AddressSpace, closing the old one
    * first.
@@ -973,6 +1141,11 @@ public:
    */
   void open_mem_fd_if_needed();
 
+  /**
+   * Do a tgkill to send a specific signal to this task.
+   */
+  void tgkill(int sig);
+
   /**
    * Return the name of the given syscall.
    */
@@ -980,10 +1153,6 @@ public:
 
   /* State only used during recording. */
 
-  /* Whether switching away from this task is allowed in its
-   * current state.  Some operations must be completed
-   * atomically and aren't switchable. */
-  Switchable switchable;
   /* True when this is switchable for semantic purposes, but
    * definitely isn't blocked on ony resource.  In that case,
    * it's safe for the scheduler to do a blocking waitpid on
@@ -996,22 +1165,39 @@ public:
   uint32_t succ_event_counter;
   /* True when any assumptions made about the status of this
    * process have been invalidated, and must be re-established
-   * with a waitpid() call. */
+   * with a waitpid() call. Only applies to tasks which are dying, usually
+   * due to a signal sent to the entire task group. */
   bool unstable;
   /* exit(), or exit_group() with one task, has been called, so
    * the exit can be treated as stable. */
   bool stable_exit;
-
   /* Task 'nice' value set by setpriority(2).
      We use this to drive scheduling decisions. rr's scheduler is
      deliberately simple and unfair; a task never runs as long as there's
      another runnable task with a lower nice value. */
   int priority;
-
   /* Tasks with in_round_robin_queue set are in the session's
    * in_round_robin_queue instead of its task_priority_set.
    */
   bool in_round_robin_queue;
+  // The set of signals that were blocked during a sigsuspend. Only present
+  // during the first EV_SIGNAL during an interrupted sigsuspend.
+  std::unique_ptr<sig_set_t> sigsuspend_blocked_sigs;
+  // If not NOT_STOPPED, then the task is logically stopped and this is the type
+  // of stop.
+  EmulatedStopType emulated_stop_type;
+
+  // Task for which we're emulating ptrace of this task, or null
+  Task* emulated_ptracer;
+  // true if this task needs to send a SIGCHLD to its ptracer for its
+  // emulated ptrace stop
+  bool emulated_ptrace_SIGCHLD_pending;
+  // if nonzero, code to deliver to ptracer when it waits
+  int emulated_ptrace_stop_code;
+  std::set<Task*> emulated_ptrace_tracees;
+
+  WaitType in_wait_type;
+  pid_t in_wait_pid;
 
   /* Imagine that task A passes buffer |b| to the read()
    * syscall.  Imagine that, after A is switched out for task B,
@@ -1112,27 +1298,50 @@ public:
   struct syscallbuf_hdr* syscallbuf_hdr;
   size_t num_syscallbuf_bytes;
   /* Points at the tracee's mapping of the buffer. */
-  remote_ptr<void> syscallbuf_child;
+  remote_ptr<struct syscallbuf_hdr> syscallbuf_child;
+  remote_ptr<char> syscallbuf_fds_disabled_child;
 
-  /* The value of arg1 passed to the last execve syscall in this task. */
-  uintptr_t exec_saved_arg1;
-  std::unique_ptr<TraceTaskEvent> exec_saved_event;
+  PropertyTable& properties() { return properties_; }
+
+  struct CapturedState {
+    pid_t rec_tid;
+    uint32_t serial;
+    Registers regs;
+    ExtraRegisters extra_regs;
+    std::string prname;
+    remote_ptr<void> robust_futex_list;
+    size_t robust_futex_list_len;
+    struct user_desc thread_area;
+    bool thread_area_valid;
+    size_t num_syscallbuf_bytes;
+    int desched_fd_child;
+    remote_ptr<struct syscallbuf_hdr> syscallbuf_child;
+    std::vector<uint8_t> syscallbuf_hdr;
+    remote_ptr<char> syscallbuf_fds_disabled_child;
+    remote_ptr<void> scratch_ptr;
+    ssize_t scratch_size;
+    int wait_status;
+    sig_set_t blocked_sigs;
+    std::deque<Event> pending_events;
+    Ticks ticks;
+    remote_ptr<int> tid_futex;
+    remote_ptr<void> top_of_stack;
+  };
 
 private:
-  Task(Session& session, pid_t tid, pid_t rec_tid, int priority,
-       SupportedArch a);
+  Task(Session& session, pid_t tid, pid_t rec_tid, uint32_t serial,
+       int priority, SupportedArch a);
 
-  /** Helper function for maybe_update_vm. */
   template <typename Arch>
-  void maybe_update_vm_arch(int syscallno, SyscallEntryOrExit state);
+  void on_syscall_exit_arch(int syscallno, const Registers& regs);
 
   /** Helper function for update_sigaction. */
   template <typename Arch> void update_sigaction_arch(const Registers& regs);
 
   /** Helper function for init_buffers. */
   template <typename Arch>
-  remote_ptr<void> init_buffers_arch(remote_ptr<void> map_hint,
-                                     ShareDeschedEventFd share_desched_fd);
+  void init_buffers_arch(remote_ptr<void> map_hint,
+                         ShareDeschedEventFd share_desched_fd);
 
   /**
    * Return a new Task cloned from |p|.  |flags| are a set of
@@ -1143,20 +1352,27 @@ private:
    * to the task during recording.
    */
   Task* clone(int flags, remote_ptr<void> stack, remote_ptr<void> tls,
-              remote_ptr<int> cleartid_addr, pid_t new_tid,
-              pid_t new_rec_tid = -1, Session* other_session = nullptr);
+              remote_ptr<int> cleartid_addr, pid_t new_tid, pid_t new_rec_tid,
+              uint32_t new_serial, Session* other_session = nullptr);
 
   /**
-   * Make this task look like an identical copy of |from| in
+   * Grab state from this task into a structure that we can use to
+   * initialize a new task via os_clone_into/os_fork_into and copy_state.
+   */
+  CapturedState capture_state();
+
+  /**
+   * Make this task look like an identical copy of the task whose state
+   * was captured by capture_task_state(), in
    * every way relevant to replay.  This task should have been
-   * created by calling |from->os_clone()| or |from->os_fork()|,
+   * created by calling os_clone_into() or os_fork_into(),
    * and if it wasn't results are undefined.
    *
    * Some task state must be copied into this by injecting and
    * running syscalls in this task.  Other state is metadata
    * that can simply be copied over in local memory.
    */
-  void copy_state(Task* from);
+  void copy_state(const CapturedState& state);
 
   /**
    * Destroy tracer-side state of this (as opposed to remote,
@@ -1164,15 +1380,6 @@ private:
    */
   void destroy_local_buffers();
 
-  /**
-   * Detach this from rr and try hard to ensure any operations
-   * related to it have completed by the time this function
-   * returns.
-   *
-   * Warning: called by destructor.
-   */
-  void detach_and_reap();
-
   /**
    * Make the ptrace |request| with |addr| and |data|, return
    * the ptrace return value.
@@ -1218,23 +1425,16 @@ private:
    * |map_hint| is the address where the syscallbuf is expected
    * to be mapped --- and this is asserted --- or nullptr if
    * there are no expectations.
+   * Initializes syscallbuf_child.
    */
-  remote_ptr<void> init_syscall_buffer(AutoRemoteSyscalls& remote,
-                                       remote_ptr<void> map_hint);
+  void init_syscall_buffer(AutoRemoteSyscalls& remote,
+                           remote_ptr<void> map_hint);
 
   /**
    * True if this has blocked delivery of the desched signal.
    */
   bool is_desched_sig_blocked();
 
-  /**
-   * Destroy the OS task backing this by sending it SIGKILL and
-   * ensuring it was delivered.  After |kill()|, the only
-   * meaningful thing that can be done with this task is to
-   * delete it.
-   */
-  void kill();
-
   /**
    * Make the OS-level calls to create a new fork or clone that
    * will eventually be a copy of this task and return that Task
@@ -1251,7 +1451,8 @@ private:
    * create the new child.
    */
   Task* os_fork_into(Session* session);
-  Task* os_clone_into(Task* task_leader, AutoRemoteSyscalls& remote);
+  static Task* os_clone_into(const CapturedState& state, Task* task_leader,
+                             AutoRemoteSyscalls& remote);
 
   /**
    * Return the TraceStream that we're using, if in recording or replay.
@@ -1270,7 +1471,8 @@ private:
    */
   static Task* os_clone(Task* parent, Session* session,
                         AutoRemoteSyscalls& remote, pid_t rec_child_tid,
-                        unsigned base_flags, remote_ptr<void> stack = nullptr,
+                        uint32_t new_serial, unsigned base_flags,
+                        remote_ptr<void> stack = nullptr,
                         remote_ptr<int> ptid = nullptr,
                         remote_ptr<void> tls = nullptr,
                         remote_ptr<int> ctid = nullptr);
@@ -1279,13 +1481,13 @@ private:
   static Task* spawn(Session& session, const TraceStream& trace,
                      pid_t rec_tid = -1);
 
+  uint32_t serial;
   // The address space of this task.
-  std::shared_ptr<AddressSpace> as;
+  AddressSpace::shr_ptr as;
+  // The file descriptor table of this task.
+  FdTable::shr_ptr fds;
   // The set of signals that are currently blocked.
   sig_set_t blocked_sigs;
-  // The exe-file argument passed to the most recent execve call
-  // made by this task.
-  std::string execve_file;
   // The current stack of events being processed.  (We use a
   // deque instead of a stack because we need to iterate the
   // events.)
@@ -1300,6 +1502,9 @@ private:
   // True when we know via waitpid() that the task is stopped and we haven't
   // resumed it.
   bool is_stopped;
+  // True when there was a breakpoint set at the location where we resumed
+  // execution
+  bool breakpoint_set_where_execution_resumed;
   // When |extra_registers_known|, we have saved our extra registers.
   ExtraRegisters extra_registers;
   bool extra_registers_known;
@@ -1322,13 +1527,10 @@ private:
   // of user sighandlers (see below). */
   std::shared_ptr<Sighandlers> sighandlers;
   // Stashed signal-delivery state, ready to be delivered at
-  // next opportunity.  |stashed_si| is only meaningful when
-  // |stashed_wait_status| is nonzero.
+  // next opportunity.
   struct StashedSignal {
-    StashedSignal(const siginfo_t& si, int wait_status)
-        : si(si), wait_status(wait_status) {}
+    StashedSignal(const siginfo_t& si) : si(si) {}
     siginfo_t si;
-    int wait_status;
   };
   std::deque<StashedSignal> stashed_signals;
   // The task group this belongs to.
@@ -1352,6 +1554,8 @@ private:
   // for this task.
   bool seen_ptrace_exit_event;
 
+  PropertyTable properties_;
+
   Task(Task&) = delete;
   Task operator=(Task&) = delete;
 };
