@@ -2,6 +2,8 @@
 
 #include "rrutil.h"
 
+static void breakpoint(void) {}
+
 int caught_sig = 0;
 
 void catcher(int signum, siginfo_t* siginfo_ptr, void* ucontext_ptr) {
@@ -19,12 +21,17 @@ int main(int argc, char** argv) {
 
   alarm(1); /* timer will pop in 1 second */
 
-  for (counter = 0; counter >= 0 && !caught_sig; counter++)
-    if (counter % 100000 == 0)
+  for (counter = 0; counter >= 0 && !caught_sig; counter++) {
+    if (counter % 100000 == 0) {
       write(STDOUT_FILENO, ".", 1);
+    }
+  }
 
   atomic_printf("\nSignal %d caught, Counter is %d\n", caught_sig, counter);
   test_assert(SIGALRM == caught_sig);
+
+  breakpoint();
+
   atomic_puts("EXIT-SUCCESS");
 
   return 0;
