@@ -10,7 +10,7 @@ expect_gdb('Breakpoint 1')
 send_gdb('c\n')
 observe_normal_parent_exit()
 
-restart_replay_at_end()
+restart_replay()
 observe_normal_parent_exit()
 
 ok()
