@@ -3,7 +3,10 @@ from rrutil import *
 send_gdb('c\n')
 expect_gdb('exited')
 
-send_gdb('run\n')
+restart_replay()
+expect_gdb('exited')
+
+restart_replay()
 expect_gdb('exited')
 
 ok()
