@@ -3,7 +3,7 @@ from rrutil import *
 send_gdb('c\n')
 expect_gdb('exited normally')
 
-restart_replay_at_end()
+restart_replay()
 
 expect_gdb('exited normally')
 
