@@ -51,12 +51,13 @@ function delay_kill { sig=$1; delay_secs=$2; proc=$3
 
     kill -s $sig $pid
     if [[ $? != 0 ]]; then
-        leave_data=y
-        echo FAILED: signal $sig not delivered to "'$proc'" >&2
-        exit 1
+        # Sometimes we fail to deliver a signal to a process because
+        # it finished first (due to scheduling races). That's a benign
+        # failure.
+        echo signal $sig not delivered to "'$proc'", letting test succeed anyway
+    else
+        echo Successfully delivered signal $sig to "'$proc'"
     fi
-
-    echo Successfully delivered signal $sig to "'$proc'"
 }
 
 function fatal { #...
@@ -98,13 +99,17 @@ GLOBAL_OPTIONS="$DEFAULT_FLAGS"
 GLOBAL_OPTIONS_BIND_CPU="$DEFAULT_FLAGS"
 
 LIB_ARG=$1
-OBJDIR=$2
+SRCDIR=$2
+if [[ ! -d "$SRCDIR" ]]; then
+    fatal "FAILED: srcdir missing"
+fi
+OBJDIR=$3
 if [[ "$OBJDIR" == "" ]]; then
     # Default to assuming that the user's working directory is the
     # test/ directory within the rr clone.
     OBJDIR=`realpath ../../../obj`
 fi
-TESTNAME=$3
+TESTNAME=$4
 if [[ "$TESTNAME" == "" ]]; then
     [[ $0 =~ ([A-Za-z0-9_]+)\.run$ ]] || fatal "FAILED: bad test script name"
     TESTNAME=${BASH_REMATCH[1]}
@@ -125,7 +130,6 @@ leave_data=n
 nonce=
 
 # Set up the environment and working directory.
-SRCDIR="${OBJDIR}/../rr"
 TESTDIR="${SRCDIR}/src/test"
 
 export PATH="${OBJDIR}/bin:${PATH}"
@@ -136,6 +140,14 @@ if [[ "$?" != "0" ]]; then
     fatal FAILED: rr not found in PATH "($PATH)"
 fi
 
+if [[ ! -d $SRCDIR ]]; then
+    fatal FAILED: SRCDIR "($SRCDIR)" not found. objdir and srcdir must share the same parent.
+fi
+
+if [[ ! -d $TESTDIR ]]; then
+    fatal FAILED: TESTDIR "($TESTDIR)" not found.
+fi
+
 # NB: must set up the trap handler *before* mktemp
 trap onexit EXIT
 workdir=`mktemp -dt rr-test-$TESTNAME-XXXXXXXXX`
@@ -144,7 +156,8 @@ cd $workdir
 # XXX technically the trailing -XXXXXXXXXX isn't unique, since there
 # could be "foo-123456789" and "bar-123456789", but if that happens,
 # buy me a lottery ticket.
-nonce=${workdir#/tmp/rr-test-$TESTNAME-}
+baseworkdir=$(basename ${workdir})
+nonce=${baseworkdir#rr-test-$TESTNAME-}
 
 ##--------------------------------------------------
 ## Now we come to the helpers available to test runners.  This is the
@@ -176,7 +189,7 @@ function skip_if_syscall_buf {
 
 function just_record { exe=$1; exeargs=$2;
     _RR_TRACE_DIR="$workdir" \
-        rr $GLOBAL_OPTIONS record $LIB_ARG $RECORD_ARGS $exe $exeargs 1> record.out
+        rr $GLOBAL_OPTIONS record $LIB_ARG $RECORD_ARGS $exe $exeargs 1> record.out 2> record.err
 }
 
 function save_exe { exe=$1;
@@ -192,6 +205,8 @@ function record { exe=$1;
 #  record_async_signal <signal> <delay-secs> <test>
 #
 # Record $test, delivering $signal to it after $delay-secs.
+# If for some reason delay_kill doesn't run in time, the signal
+# will not be delivered but the test will not be aborted.
 function record_async_signal { sig=$1; delay_secs=$2; exe=$3; exeargs=$4;
     delay_kill $sig $delay_secs $exe-$nonce &
     record $exe $exeargs
@@ -203,12 +218,12 @@ function replay { replayflags=$1
         rr $GLOBAL_OPTIONS replay -a $replayflags 1> replay.out 2> replay.err
 }
 
-#  debug <exe> <expect-script-name> [replay-args]
+#  debug <expect-script-name> [replay-args]
 #
 # Load the "expect" script to drive replay of the recording of |exe|.
-function debug { exe=$1; expectscript=$2; replayargs=$3
+function debug { expectscript=$1; replayargs=$2
     _RR_TRACE_DIR="$workdir" \
-        python $TESTDIR/$expectscript.py $exe-$nonce \
+        python2 $TESTDIR/$expectscript.py \
         rr $GLOBAL_OPTIONS replay -x $TESTDIR/test_setup.gdb $replayargs
     if [[ $? == 0 ]]; then
         passed
@@ -230,19 +245,32 @@ function passed {
 # output match; (iii) the supplied token was found in the output.
 # Otherwise the test fails.
 function check { token=$1;
-    if [ ! -f record.out -o ! -f replay.err -o ! -f replay.out ]; then
+    if [ ! -f record.out -o ! -f replay.err -o ! -f replay.out -o ! -f record.err ]; then
         failed "output files not found."
+    elif [[ $(cat record.err) != "" ]]; then
+        failed ": error during recording:"
+        echo "--------------------------------------------------"
+        cat record.err
+        echo "--------------------------------------------------"
+        echo "record.out:"
+        echo "--------------------------------------------------"
+        cat record.out
+        echo "--------------------------------------------------"
     elif [[ $(cat replay.err) != "" ]]; then
         failed ": error during replay:"
         echo "--------------------------------------------------"
         cat replay.err
         echo "--------------------------------------------------"
+        echo "replay.out:"
+        echo "--------------------------------------------------"
+        cat replay.out
+        echo "--------------------------------------------------"
     elif [[ $(diff record.out replay.out) != "" ]]; then
         failed ": output from recording different than replay"
         echo "diff -U8 $workdir/record.out $workdir/replay.out"
         diff -U8 record.out replay.out
     elif [[ "$token" != "" && "record.out" != $(grep -l "$token" record.out) ]]; then
-        failed ": token '$token' not in output:"
+        failed ": token '$token' not in record.out:"
         echo "--------------------------------------------------"
         cat record.out
         echo "--------------------------------------------------"
@@ -277,7 +305,7 @@ function compare_test { token=$1; replayflags=$2;
 # computing test pass/fail.
 function debug_test {
     record $TESTNAME
-    debug $TESTNAME $TESTNAME_NO_BITNESS
+    debug $TESTNAME_NO_BITNESS
 }
 
 # Return the number of events in the most recent local recording.
@@ -324,7 +352,7 @@ function checkpoint_test { exe=$1; min=$2; max=$3;
     stride=$(rand_range $min $max)
     for i in $(seq 1 $stride $num_events); do
         echo Checkpointing at event $i ...
-        debug $exe restart_finish "-g $i"
+        debug restart_finish "-g $i"
         if [[ "$leave_data" == "y" ]]; then
             break
         fi
