@@ -7,11 +7,16 @@
 
 #include "util.h"
 
+#include <algorithm>
+
 #include <assert.h>
 #include <elf.h>
 #include <fcntl.h>
 #include <inttypes.h>
+#include <linux/filter.h>
 #include <linux/magic.h>
+#include <linux/prctl.h>
+#include <linux/seccomp.h>
 #include <string.h>
 #include <stdlib.h>
 #include <sys/vfs.h>
@@ -19,6 +24,7 @@
 
 #include "preload/preload_interface.h"
 
+#include "AutoRemoteSyscalls.h"
 #include "Flags.h"
 #include "kernel_abi.h"
 #include "kernel_metadata.h"
@@ -51,21 +57,6 @@ bool probably_not_interactive(int fd) {
   return !isatty(fd);
 }
 
-void maybe_mark_stdio_write(Task* t, int fd) {
-  char buf[256];
-  ssize_t len;
-
-  if (!Flags::get().mark_stdio ||
-      !(STDOUT_FILENO == fd || STDERR_FILENO == fd)) {
-    return;
-  }
-  snprintf(buf, sizeof(buf) - 1, "[rr %d %d]", t->tgid(), t->trace_time());
-  len = strlen(buf);
-  if (write(fd, buf, len) != len) {
-    FATAL() << "Couldn't write to " << fd;
-  }
-}
-
 int clone_flags_to_task_flags(int flags_arg) {
   int flags = CLONE_SHARE_NOTHING;
   // See task.h for description of the flags.
@@ -74,6 +65,7 @@ int clone_flags_to_task_flags(int flags_arg) {
   flags |= (CLONE_SIGHAND & flags_arg) ? CLONE_SHARE_SIGHANDLERS : 0;
   flags |= (CLONE_THREAD & flags_arg) ? CLONE_SHARE_TASK_GROUP : 0;
   flags |= (CLONE_VM & flags_arg) ? CLONE_SHARE_VM : 0;
+  flags |= (CLONE_FILES & flags_arg) ? CLONE_SHARE_FILES : 0;
   return flags;
 }
 
@@ -84,10 +76,19 @@ size_t ceil_page_size(size_t sz) {
   return (sz + page_size() - 1) & page_mask;
 }
 
+size_t floor_page_size(size_t sz) {
+  size_t page_mask = ~(page_size() - 1);
+  return sz & page_mask;
+}
+
 remote_ptr<void> ceil_page_size(remote_ptr<void> addr) {
   return remote_ptr<void>(ceil_page_size(addr.as_int()));
 }
 
+remote_ptr<void> floor_page_size(remote_ptr<void> addr) {
+  return remote_ptr<void>(floor_page_size(addr.as_int()));
+}
+
 /**
  * Dump |buf_len| words in |buf| to |out|, starting with a line
  * containing |label|.  See |dump_binary_data()| for a description of
@@ -116,8 +117,9 @@ void dump_binary_data(const char* filename, const char* label,
   fclose(out);
 }
 
-void format_dump_filename(Task* t, int global_time, const char* tag,
-                          char* filename, size_t filename_size) {
+void format_dump_filename(Task* t, TraceFrame::Time global_time,
+                          const char* tag, char* filename,
+                          size_t filename_size) {
   snprintf(filename, filename_size - 1, "%s/%d_%d_%s", t->trace_dir().c_str(),
            t->rec_tid, global_time, tag);
 }
@@ -140,7 +142,8 @@ bool should_dump_memory(Task* t, const TraceFrame& f) {
          flags->dump_at == int(f.time());
 }
 
-void dump_process_memory(Task* t, int global_time, const char* tag) {
+void dump_process_memory(Task* t, TraceFrame::Time global_time,
+                         const char* tag) {
   char filename[PATH_MAX];
   FILE* dump_file;
 
@@ -168,8 +171,8 @@ void dump_process_memory(Task* t, int global_time, const char* tag) {
   fclose(dump_file);
 }
 
-static void notify_checksum_error(Task* t, int global_time, unsigned checksum,
-                                  unsigned rec_checksum,
+static void notify_checksum_error(Task* t, TraceFrame::Time global_time,
+                                  unsigned checksum, unsigned rec_checksum,
                                   const string& raw_map_line) {
   char cur_dump[PATH_MAX];
   char rec_dump[PATH_MAX];
@@ -185,7 +188,7 @@ static void notify_checksum_error(Task* t, int global_time, unsigned checksum,
                        sizeof(cur_dump));
   format_dump_filename(t, global_time, "rec", rec_dump, sizeof(rec_dump));
 
-  Event ev(t->current_trace_frame().event());
+  const Event& ev = t->current_trace_frame().event();
   ASSERT(t, checksum == rec_checksum)
       << "Divergence in contents of memory segment after '" << ev << "':\n"
                                                                      "\n"
@@ -218,7 +221,7 @@ enum ChecksumMode {
 struct checksum_iterator_data {
   ChecksumMode mode;
   FILE* checksums_file;
-  int global_time;
+  TraceFrame::Time global_time;
 };
 
 static bool checksum_segment_filter(const Mapping& m,
@@ -248,7 +251,8 @@ static bool checksum_segment_filter(const Mapping& m,
  * address space, or validate an existing computed checksum.  Behavior
  * is selected by |mode|.
  */
-static void iterate_checksums(Task* t, ChecksumMode mode, int global_time) {
+static void iterate_checksums(Task* t, ChecksumMode mode,
+                              TraceFrame::Time global_time) {
   struct checksum_iterator_data c;
   memset(&c, 0, sizeof(c));
   char filename[PATH_MAX];
@@ -349,8 +353,8 @@ static void iterate_checksums(Task* t, ChecksumMode mode, int global_time) {
 
 bool should_checksum(Task* t, const TraceFrame& f) {
   int checksum = Flags::get().checksum;
-  bool is_syscall_exit =
-      EV_SYSCALL == f.event().type && SYSCALL_EXIT == f.event().state;
+  bool is_syscall_exit = EV_SYSCALL == f.event().type() &&
+                         EXITING_SYSCALL == f.event().Syscall().state;
 
 #if defined(FIRST_INTERESTING_EVENT)
   if (is_syscall_exit && FIRST_INTERESTING_EVENT <= global_time &&
@@ -374,11 +378,11 @@ bool should_checksum(Task* t, const TraceFrame& f) {
   return checksum <= int(f.time());
 }
 
-void checksum_process_memory(Task* t, int global_time) {
+void checksum_process_memory(Task* t, TraceFrame::Time global_time) {
   iterate_checksums(t, STORE_CHECKSUMS, global_time);
 }
 
-void validate_process_memory(Task* t, int global_time) {
+void validate_process_memory(Task* t, TraceFrame::Time global_time) {
   iterate_checksums(t, VALIDATE_CHECKSUMS, global_time);
 }
 
@@ -433,6 +437,32 @@ signal_action default_action(int sig) {
   }
 }
 
+SignalDeterministic is_deterministic_signal(const siginfo_t& si) {
+  switch (si.si_signo) {
+    /* These signals may be delivered deterministically;
+     * we'll check for sure below. */
+    case SIGILL:
+    case SIGTRAP:
+    case SIGBUS:
+    case SIGFPE:
+    case SIGSEGV:
+      /* As bits/siginfo.h documents,
+       *
+       *   Values for `si_code'.  Positive values are
+       *   reserved for kernel-generated signals.
+       *
+       * So if the signal is maybe-synchronous, and the
+       * kernel delivered it, then it must have been
+       * delivered deterministically. */
+      return si.si_code > 0 ? DETERMINISTIC_SIG : NONDETERMINISTIC_SIG;
+    default:
+      /* All other signals can never be delivered
+       * deterministically (to the approximation required by
+       * rr). */
+      return NONDETERMINISTIC_SIG;
+  }
+}
+
 bool possibly_destabilizing_signal(Task* t, int sig,
                                    SignalDeterministic deterministic) {
   signal_action action = default_action(sig);
@@ -557,9 +587,9 @@ bool should_copy_mmap_region(const string& file_name, const struct stat* stat,
   return true;
 }
 
-ScopedFd create_shmem_segment(const char* name, size_t num_bytes) {
+ScopedFd create_shmem_segment(const string& name, size_t num_bytes) {
   char path[PATH_MAX];
-  snprintf(path, sizeof(path) - 1, "%s/%s", SHMEM_FS, name);
+  snprintf(path, sizeof(path) - 1, "%s/%s", SHMEM_FS, name.c_str());
 
   ScopedFd fd = open(path, O_CREAT | O_EXCL | O_RDWR | O_CLOEXEC, 0600);
   if (0 > fd) {
@@ -580,60 +610,6 @@ void resize_shmem_segment(ScopedFd& fd, size_t num_bytes) {
   }
 }
 
-// TODO de-dup
-static void advance_syscall(Task* t) {
-  do {
-    t->cont_syscall();
-  } while (t->is_ptrace_seccomp_event() ||
-           ReplaySession::is_ignored_signal(t->pending_sig()));
-  assert(t->ptrace_event() == 0);
-}
-
-void destroy_buffers(Task* t) {
-  // NB: we have to pay all this complexity here because glibc
-  // makes its SYS_exit call through an inline int $0x80 insn,
-  // instead of going through the vdso.  There may be a deep
-  // reason for why it does that, but if it starts going through
-  // the vdso in the future, this code can be eliminated in
-  // favor of a *much* simpler vsyscall SYS_exit hook in the
-  // preload lib.
-  Registers exit_regs = t->regs();
-  ASSERT(t, is_exit_syscall(exit_regs.original_syscallno(), t->arch()))
-      << "Tracee should have been at exit, but instead at "
-      << t->syscall_name(exit_regs.original_syscallno());
-
-  // The tracee is at the entry to SYS_exit, but hasn't started
-  // the call yet.  We can't directly start injecting syscalls
-  // because the tracee is still in the kernel.  And obviously,
-  // if we finish the SYS_exit syscall, the tracee isn't around
-  // anymore.
-  //
-  // So hijack this SYS_exit call and rewrite it into a harmless
-  // one that we can exit successfully, SYS_gettid here (though
-  // that choice is arbitrary).
-  exit_regs.set_original_syscallno(syscall_number_for_gettid(t->arch()));
-  t->set_regs(exit_regs);
-  // This exits the hijacked SYS_gettid.  Now the tracee is
-  // ready to do our bidding.
-  advance_syscall(t);
-
-  // Restore these regs to what they would have been just before
-  // the tracee trapped at SYS_exit.  When we've finished
-  // cleanup, we'll restart the SYS_exit call.
-  exit_regs.set_original_syscallno(-1);
-  exit_regs.set_syscallno(syscall_number_for_exit(t->arch()));
-  exit_regs.set_ip(exit_regs.ip() - syscall_instruction_length(t->arch()));
-  ASSERT(t, is_at_syscall_instruction(t, exit_regs.ip()))
-      << "Tracee should have entered through int $0x80.";
-
-  // Do the actual buffer and fd cleanup.
-  t->destroy_buffers(DESTROY_SCRATCH | DESTROY_SYSCALLBUF);
-
-  // Restart the SYS_exit call.
-  t->set_regs(exit_regs);
-  advance_syscall(t);
-}
-
 void cpuid(int code, int subrequest, unsigned int* a, unsigned int* c,
            unsigned int* d) {
   asm volatile("cpuid"
@@ -728,3 +704,86 @@ bool trace_instructions_up_to_event(TraceFrame::Time event) {
   return event > instruction_trace_at_event_start &&
          event <= instruction_trace_at_event_last;
 }
+
+void dump_task_set(const set<Task*>& tasks) {
+  printf("[");
+  for (auto& t : tasks) {
+    printf("%p (pid=%d, rec=%d),", t, t->tid, t->rec_tid);
+  }
+  printf("]\n");
+}
+
+void dump_task_map(const map<pid_t, Task*>& tasks) {
+  printf("[");
+  for (auto& t : tasks) {
+    printf("%p (pid=%d, rec=%d),", t.second, t.second->tid, t.second->rec_tid);
+  }
+  printf("]\n");
+}
+
+static void set_syscall_result(Task* t, long ret) {
+  Registers r = t->regs();
+  r.set_syscall_result(ret);
+  t->set_regs(r);
+}
+
+template <typename Arch>
+static void install_patched_seccomp_filter_arch(Task* t) {
+  // Take advantage of the fact that the filter program is arg3() in both
+  // prctl and seccomp syscalls.
+  bool ok = true;
+  auto prog =
+      t->read_mem(remote_ptr<typename Arch::sock_fprog>(t->regs().arg3()), &ok);
+  if (!ok) {
+    set_syscall_result(t, -EFAULT);
+    return;
+  }
+  auto code = t->read_mem(prog.filter.rptr(), prog.len, &ok);
+  if (!ok) {
+    set_syscall_result(t, -EFAULT);
+    return;
+  }
+  for (auto& u : code) {
+    if (BPF_CLASS(u.code) == BPF_RET) {
+      // XXX If we need to support RET with A/X registers, we should
+      // extend the filter patch to dynamically cap the errno (if any) to
+      // max_errno and switch to SECCOMP_RET_TRACE.
+      ASSERT(t, BPF_RVAL(u.code) == BPF_K)
+          << "seccomp-bpf program uses BPF_RET with A/X register, not "
+             "supported";
+      if ((u.k & SECCOMP_RET_ACTION) == SECCOMP_RET_ERRNO) {
+        static int max_errno = 4095;
+        // The kernel caps the max errno to 4095, so we may as well do that
+        // here. This means filter data values > 4095 cannot be generated
+        // by this filter, which lets us disambiguate seccomp errno filter
+        // returns from our filter returns.
+        int filter_errno = min<int>(max_errno, u.k & SECCOMP_RET_DATA);
+        // Instead of forcing an errno return directly, trigger a ptrace
+        // trap so we can detect and handle it.
+        u.k = filter_errno | SECCOMP_RET_TRACE;
+      }
+    }
+  }
+
+  long ret;
+  {
+    AutoRemoteSyscalls remote(t);
+    AutoRestoreMem mem(remote, nullptr,
+                       sizeof(prog) +
+                           code.size() * sizeof(typename Arch::sock_filter));
+    auto code_ptr = mem.get().cast<typename Arch::sock_filter>();
+    t->write_mem(code_ptr, code.data(), code.size());
+    prog.filter = code_ptr;
+    auto prog_ptr = remote_ptr<void>(code_ptr + code.size())
+                        .cast<typename Arch::sock_fprog>();
+    t->write_mem(prog_ptr, prog);
+
+    ret = remote.syscall(t->regs().original_syscallno(), t->regs().arg1(),
+                         t->regs().arg2(), prog_ptr);
+  }
+  set_syscall_result(t, ret);
+}
+
+void install_patched_seccomp_filter(Task* t) {
+  RR_ARCH_FUNCTION(install_patched_seccomp_filter_arch, t->arch(), t);
+}
