@@ -4,6 +4,8 @@
 #define RR_UTIL_H_
 
 #include <array>
+#include <map>
+#include <set>
 #include <string>
 
 #include "Event.h"
@@ -23,6 +25,28 @@ constexpr size_t array_length(std::array<T, N>& array) {
   return N;
 }
 
+template <typename T> T return_dummy_value() {
+  T v;
+  memset(&v, 1, sizeof(T));
+  return v;
+}
+template <typename T> bool check_type_has_no_holes() {
+  T v;
+  memset(&v, 2, sizeof(T));
+  v = return_dummy_value<T>();
+  return memchr(&v, 2, sizeof(T)) == NULL;
+}
+/**
+ * Returns true when type T has no holes. Preferably should not be defined
+ * at all otherwise.
+ * This is not 100% reliable since the check_type_has_no_holes may be
+ * compiled to copy holes. However, it has detected at least two bugs.
+ */
+template <typename T> bool type_has_no_holes() {
+  static bool check = check_type_has_no_holes<T>();
+  return check;
+}
+
 #define SHMEM_FS "/dev/shm"
 #define SHMEM_FS2 "/run/shm"
 
@@ -39,6 +63,11 @@ enum Completion {
   INCOMPLETE
 };
 
+enum Switchable {
+  PREVENT_SWITCH,
+  ALLOW_SWITCH
+};
+
 /**
  * Create a file named |filename| and dump |buf_len| words in |buf| to
  * that file, starting with a line containing |label|.  |start_addr|
@@ -54,13 +83,14 @@ void dump_binary_data(const char* filename, const char* label,
  * Format a suitable filename within the trace directory for dumping
  * information about |t| at the current global time, to a file that
  * contains |tag|.  The constructed filename is returned through
- * |filename|.  For example, a filename for a task with tid 12345 at
+ * |filename|.  For example, a filengit logame for a task with tid 12345 at
  * time 111, for a file tagged "foo", would be something like
  * "trace_0/12345_111_foo".  The returned name is not guaranteed to be
  * unique, caveat emptor.
  */
-void format_dump_filename(Task* t, int global_time, const char* tag,
-                          char* filename, size_t filename_size);
+void format_dump_filename(Task* t, TraceFrame::Time global_time,
+                          const char* tag, char* filename,
+                          size_t filename_size);
 
 /**
  * Return true if the user requested memory be dumped for |t| at |f|.
@@ -70,7 +100,8 @@ bool should_dump_memory(Task* t, const TraceFrame& f);
  * Dump all of the memory in |t|'s address to the file
  * "[trace_dir]/[t->tid]_[global_time]_[tag]".
  */
-void dump_process_memory(Task* t, int global_time, const char* tag);
+void dump_process_memory(Task* t, TraceFrame::Time global_time,
+                         const char* tag);
 
 /**
  * Return true if the user has requested |t|'s memory be
@@ -82,12 +113,12 @@ bool should_checksum(Task* t, const TraceFrame& f);
  * special log, where it can be read by |validate_process_memory()|
  * during replay.
  */
-void checksum_process_memory(Task* t, int global_time);
+void checksum_process_memory(Task* t, TraceFrame::Time global_time);
 /**
  * Validate the checksum of |t|'s address space that was written
  * during recording.
  */
-void validate_process_memory(Task* t, int global_time);
+void validate_process_memory(Task* t, TraceFrame::Time global_time);
 
 /**
  * Return nonzero if the rr session is probably not interactive (that
@@ -96,13 +127,6 @@ void validate_process_memory(Task* t, int global_time);
  */
 bool probably_not_interactive(int fd = STDERR_FILENO);
 
-/**
- * If |child_fd| is a stdio fd and stdio-marking is enabled, prepend
- * the stdio write with "[rr.<global-time>]".  This allows users to
- * more easily correlate stdio with trace event numbers.
- */
-void maybe_mark_stdio_write(Task* t, int child_fd);
-
 /**
  * Convert the flags passed to the clone() syscall, |flags_arg|, into
  * the format understood by Task::clone().
@@ -116,6 +140,13 @@ int clone_flags_to_task_flags(int flags_arg);
 size_t ceil_page_size(size_t sz);
 remote_ptr<void> ceil_page_size(remote_ptr<void> addr);
 
+/**
+ * Return the argument rounded down to the nearest multiple of the
+ * system |page_size()|.
+ */
+size_t floor_page_size(size_t sz);
+remote_ptr<void> floor_page_size(remote_ptr<void> addr);
+
 /** Return the system page size. */
 size_t page_size();
 
@@ -139,6 +170,8 @@ signal_action default_action(int sig);
 bool possibly_destabilizing_signal(Task* t, int sig,
                                    SignalDeterministic deterministic);
 
+SignalDeterministic is_deterministic_signal(const siginfo_t& si);
+
 /**
  * Return nonzero if a mapping of |filename| with metadata |stat|,
  * using |flags| and |prot|, should almost certainly be copied to
@@ -154,7 +187,7 @@ bool should_copy_mmap_region(const std::string& filename,
  * Return an fd referring to a new shmem segment with descriptive
  * |name| of size |num_bytes|.
  */
-ScopedFd create_shmem_segment(const char* name, size_t num_bytes);
+ScopedFd create_shmem_segment(const std::string& name, size_t num_bytes);
 
 /**
  * Ensure that the shmem segment referred to by |fd| has exactly the
@@ -162,15 +195,6 @@ ScopedFd create_shmem_segment(const char* name, size_t num_bytes);
  */
 void resize_shmem_segment(ScopedFd& fd, size_t num_bytes);
 
-/**
- * At thread exit time, undo the work that init_buffers() did.
- *
- * Call this when the tracee has already entered SYS_exit. The
- * tracee will be returned at a state in which it has entered (or
- * re-entered) SYS_exit.
- */
-void destroy_buffers(Task* t);
-
 enum cpuid_requests {
   CPUID_GETVENDORSTRING,
   CPUID_GETFEATURES,
@@ -221,4 +245,17 @@ int read_elf_class(const std::string& filename);
 
 bool trace_instructions_up_to_event(TraceFrame::Time event);
 
+/* Helpful for broken debuggers */
+
+void dump_task_set(const std::set<Task*>& tasks);
+
+void dump_task_map(const std::map<pid_t, Task*>& tasks);
+
+/**
+ * Assuming |t| is set up for a prctl or seccomp syscall that
+ * installs a seccomp-bpf filter, patch the filter to signal the tracer instead
+ * of silently delivering an errno, and install it.
+ */
+void install_patched_seccomp_filter(Task* t);
+
 #endif /* RR_UTIL_H_ */
